@@ -1,7 +1,7 @@
 /*
- * Copyright (C) 2004, 2005, 2007, 2008  Internet Systems Consortium, Inc. ("ISC")
- * Copyright (C) 1999-2003  Internet Software Consortium.
- * Copyright 2015 by the NTPsec project contributors
+ * Copyright Internet Systems Consortium, Inc. ("ISC")
+ * Copyright Internet Software Consortium.
+ * Copyright the NTPsec project contributors
  * SPDX-License-Identifier: ISC
  */
 
@@ -382,14 +382,6 @@ internal_current(isc_interfaceiter_t *iter) {
         if ((ifa->ifa_flags & IFF_LOOPBACK) != 0)
                 iter->current.flags |= INTERFACE_F_LOOPBACK;
 
-        if ((ifa->ifa_flags & IFF_BROADCAST) != 0)
-                iter->current.flags |= INTERFACE_F_BROADCAST;
-
-#ifdef IFF_MULTICAST
-        if ((ifa->ifa_flags & IFF_MULTICAST) != 0)
-                iter->current.flags |= INTERFACE_F_MULTICAST;
-#endif
-
         iter->current.af = (unsigned int)family;
 
         get_addr((unsigned int)family, &iter->current.address,
@@ -405,12 +397,6 @@ internal_current(isc_interfaceiter_t *iter) {
                          ifa->ifa_dstaddr,
                          ifa->ifa_name);
 
-        if (ifa->ifa_broadaddr != NULL &&
-            (iter->current.flags & INTERFACE_F_BROADCAST) != 0)
-                get_addr((unsigned int)family, &iter->current.broadcast,
-                         ifa->ifa_broadaddr,
-                         ifa->ifa_name);
-
         iter->current.ifindex = if_nametoindex(iter->current.name);
         return (ISC_R_SUCCESS);
 }
@@ -616,14 +602,6 @@ internal_current(isc_interfaceiter_t *iter) {
                 if ((ifam->ifam_flags & IFF_LOOPBACK) != 0)
                         iter->current.flags |= INTERFACE_F_LOOPBACK;
 
-                if ((ifam->ifam_flags & IFF_BROADCAST) != 0)
-                        iter->current.flags |= INTERFACE_F_BROADCAST;
-
-#ifdef IFF_MULTICAST
-                if ((ifam->ifam_flags & IFF_MULTICAST) != 0)
-                        iter->current.flags |= INTERFACE_F_MULTICAST;
-#endif
-
                 /*
                  * This is not an interface address.
                  * Force another iteration.
@@ -688,11 +666,6 @@ internal_current(isc_interfaceiter_t *iter) {
                         get_addr(family, &iter->current.dstaddress, dst_sa,
                                  iter->current.name);
 
-                if (dst_sa != NULL &&
-                    (iter->current.flags & INTERFACE_F_BROADCAST) != 0)
-                        get_addr(family, &iter->current.broadcast, dst_sa,
-                                 iter->current.name);
-
                 return (ISC_R_SUCCESS);
         } else {
                 printf("warning: unexpected interface list message type\n");
@@ -1156,14 +1129,6 @@ internal_current4(isc_interfaceiter_t *iter) {
         if ((ifreq.ifr_flags & IFF_LOOPBACK) != 0)
                 iter->current.flags |= INTERFACE_F_LOOPBACK;
 
-        if ((ifreq.ifr_flags & IFF_BROADCAST) != 0)
-                iter->current.flags |= INTERFACE_F_BROADCAST;
-
-#ifdef IFF_MULTICAST
-        if ((ifreq.ifr_flags & IFF_MULTICAST) != 0)
-                iter->current.flags |= INTERFACE_F_MULTICAST;
-#endif
-
         if (family == AF_INET) {
                 goto inet;
 	}
@@ -1223,23 +1188,6 @@ internal_current4(isc_interfaceiter_t *iter) {
         }
 #endif
 
-        if ((iter->current.flags & INTERFACE_F_BROADCAST) != 0) {
-                /*
-                 * Ignore the HP/UX warning about "integer overflow during
-                 * conversion.  It comes from its own macro definition,
-                 * and is really hard to shut up.
-                 */
-                if (isc_ioctl(iter->socket, SIOCGIFBRDADDR, (char *)&ifreq)
-                    < 0) {
-                        ntp_strerror_r(errno, strbuf, sizeof(strbuf));
-                        msyslog(LOG_ERR, "%s: getting broadcast address: %s",
-                                         ifreq.ifr_name, strbuf);
-                        return (ISC_R_IGNORE);
-                }
-                get_addr(family, &iter->current.broadcast,
-                         (struct sockaddr *)&ifreq.ifr_broadaddr, ifreq.ifr_name);
-        }
-
         /*
          * Get the network mask.
          */
@@ -1348,16 +1296,8 @@ internal_current6(isc_interfaceiter_t *iter) {
         if ((lifreq.lifr_flags & IFF_LOOPBACK) != 0)
                 iter->current.flags |= INTERFACE_F_LOOPBACK;
 
-        if ((lifreq.lifr_flags & IFF_BROADCAST) != 0) {
-                iter->current.flags |= INTERFACE_F_BROADCAST;
         }
 
-#ifdef IFF_MULTICAST
-        if ((lifreq.lifr_flags & IFF_MULTICAST) != 0) {
-                iter->current.flags |= INTERFACE_F_MULTICAST;
-        }
-#endif
-
 #ifdef IFF_POINTOPOINT
         /*
          * If the interface is point-to-point, get the destination address.
@@ -1381,26 +1321,6 @@ internal_current6(isc_interfaceiter_t *iter) {
         }
 #endif
 
-#ifdef SIOCGLIFBRDADDR
-        if ((iter->current.flags & INTERFACE_F_BROADCAST) != 0) {
-                /*
-                 * Ignore the HP/UX warning about "integer overflow during
-                 * conversion.  It comes from its own macro definition,
-                 * and is really hard to shut up.
-                 */
-                if (isc_ioctl(iter->socket, SIOCGLIFBRDADDR, (char *)&lifreq)
-                    < 0) {
-                        ntp_strerror_r(errno, strbuf, sizeof(strbuf));
-                        msyslog(LOG_ERR, "%s: getting broadcast address: %s",
-                                         lifreq.lifr_name, strbuf);
-                        return (ISC_R_IGNORE);
-                }
-                get_addr(family, &iter->current.broadcast,
-                         (struct sockaddr *)&lifreq.lifr_broadaddr,
-                         lifreq.lifr_name);
-        }
-#endif  /* SIOCGLIFBRDADDR */
-
         /*
          * Get the network mask.  Netmask already zeroed.
          */
