@@ -1,6 +1,47 @@
 /*
  *	CMAC and digest support for NTP
  */
+
+/*  Notes:
+ *
+ * This module covers the simple shared-key authentication.
+ * This is the working code that has to go fast.
+ * The setup code in authreadkeys is not time critical.
+ *
+ * There are 3 main options: MD5, SHA1, and AES.
+ * MD5 and SHA1 are digests.
+ *   https://en.wikipedia.org/wiki/Message_digest
+ * AES is a MAC.
+ *   https://en.wikipedia.org/wiki/Message_authentication_code
+ * OpenSSL has different APIs for them.
+ *
+ * Before OpenSSL 3, we (and many others) used the undocumented
+ * CMAC interface via openssl/cmac.h which is now (loudly) deprecated.
+ *
+ * The per packet operations all have this form:
+ *   EVP_MAC_init()
+ *   EVP_MAC_update()
+ *   EVP_MAC_final()
+ *
+ * The init step involves setting things up for the desired algorithm
+ * and key.  This can be an expensive step.  Some or much of the work
+ * can be pushed back to the one-time setup routines at the cost of
+ * more memory.  I call that preloading.
+ *
+ * This code now expects both the cipher and key to be preloaded.
+ * Just preloading the cipher will save a lot of memory if you
+ * are using a lot of keys.  The edit in this code is simple.
+ *
+ * Play with attic/cmac-timing for numbers.
+ *
+ *
+ * Modern CPUs come with support to speed up AES operations.
+ * On Intel, it's the aes capabilitiy.  You can see them
+ * under flags in /proc/cpuinfo
+ *
+ * Maybe more info in attic/cmac-timing and friends.
+ */
+
 #include "config.h"
 
 #include <string.h>
@@ -8,45 +49,26 @@
 #include <stdbool.h>
 #include <stdint.h>
 
+#include <openssl/err.h>
 #include <openssl/evp.h>	/* provides OpenSSL digest API */
 #include <openssl/md5.h>
+#include "hack-ancient-openssl.h"
 
 #include "ntp_fp.h"
 #include "ntp_stdlib.h"
 #include "ntp_auth.h"
 #include "ntp.h"
 
-#ifndef EVP_MD_CTX_reset
-/* Slightly older version of OpenSSL */
-/* Similar hack in ssl_init.c and attic/digest-timing.c */
-#define EVP_MD_CTX_reset(ctx) EVP_MD_CTX_init(ctx)
-#endif
-
 /* Need one per thread. */
 extern EVP_MD_CTX *digest_ctx;
+
+#if OPENSSL_VERSION_NUMBER > 0x20000000L
+#include <openssl/params.h>
+#else
+#include <openssl/cmac.h>
 extern CMAC_CTX *cmac_ctx;
+#endif
 
-/* ctmemeq - test two blocks memory for equality without leaking
- * timing information.
- *
- * Return value: true if the two blocks of memory are equal, false
- * otherwise.
- *
- * TODO: find out if this is useful elsewhere and if so move
- * it to a more appropriate place and give it a prototype in a
- * header file.
- */
-static bool ctmemeq(const void *s1, const void *s2, size_t n) {
-	const uint8_t *a = s1;
-	const uint8_t *b = s2;
-	uint8_t accum = 0;
-
-	for(size_t i = 0; i < n; i++) {
-		accum |= a[i] ^ b[i];
-	}
-
-	return accum == 0;
-}
 
 /*
  * cmac_encrypt - generate CMAC authenticator
@@ -62,18 +84,45 @@ cmac_encrypt(
 {
 	uint8_t	mac[CMAC_MAX_MAC_LENGTH];
 	size_t	len;
-	CMAC_CTX *ctx = cmac_ctx;
+#if OPENSSL_VERSION_NUMBER > 0x20000000L
+        EVP_MAC_CTX *ctx = auth->mac_ctx;
 
-	CMAC_resume(ctx);
+#if OPENSSL_VERSION_NUMBER > 0x30000020L
+        if (0 == EVP_MAC_init(ctx, NULL, 0, NULL)) {
+#else
+// Bug in OpenSSL 3.0.2
+// Need to reload key which is slow.  See attic/cmac-timing
+        if (0 == EVP_MAC_init(ctx, auth->key, auth->key_size, NULL)) {
+#endif
+                unsigned long err = ERR_get_error();
+                char * str = ERR_error_string(err, NULL);
+                msyslog(LOG_ERR, "encrypt: EVP_MAC_init() failed: %s.", str);
+                exit(1);
+        }
+        if (0 == EVP_MAC_update(ctx, (unsigned char *)pkt, length)) {
+                unsigned long err = ERR_get_error();
+                char * str = ERR_error_string(err, NULL);
+                msyslog(LOG_ERR, "encrypt: EVP_MAC_update() failed: %s.", str);
+                exit(1);
+        }
+        if (0 == EVP_MAC_final(ctx, mac, &len, sizeof(mac))) {
+                unsigned long err = ERR_get_error();
+                char * str = ERR_error_string(err, NULL);
+                msyslog(LOG_ERR, "encrypt: EVP_MAC_final() failed: %s.", str);
+                exit(1);
+        }
+#else
+	CMAC_CTX *ctx = cmac_ctx;
 	if (!CMAC_Init(ctx, auth->key, auth->key_size, auth->cipher, NULL)) {
 		/* Shouldn't happen.  Does if wrong key_size. */
 		msyslog(LOG_ERR,
-		    "CMAC: encrypt: CMAC init failed, %u, %u",
+		    "encrypt: CMAC init failed, %u, %u",
 			auth->keyid, auth->key_size);
 		return (0);
 	}
 	CMAC_Update(ctx, (uint8_t *)pkt, (unsigned int)length);
 	CMAC_Final(ctx, mac, &len);
+#endif
 	if (MAX_BARE_MAC_LENGTH < len)
 		len = MAX_BARE_MAC_LENGTH;
 	memmove((uint8_t *)pkt + length + 4, mac, len);
@@ -96,20 +145,48 @@ cmac_decrypt(
 {
 	uint8_t	mac[CMAC_MAX_MAC_LENGTH];
 	size_t	len;
-	CMAC_CTX *ctx = cmac_ctx;
+#if OPENSSL_VERSION_NUMBER > 0x20000000L
+        EVP_MAC_CTX *ctx = auth->mac_ctx;
 
-	CMAC_resume(ctx);
+#if OPENSSL_VERSION_NUMBER > 0x30000020L
+        if (0 == EVP_MAC_init(ctx, NULL, 0, NULL)) {
+#else
+// Bug in OpenSSL 3.0.2
+// Need to reload key which is slow.  See attic/cmac-timing
+        if (0 == EVP_MAC_init(ctx, auth->key, auth->key_size, NULL)) {
+#endif
+                unsigned long err = ERR_get_error();
+                char * str = ERR_error_string(err, NULL);
+                msyslog(LOG_ERR, "decrypt: EVP_MAC_init() failed: %s.", str);
+                return false;
+        }
+        if (0 == EVP_MAC_update(ctx, (unsigned char *)pkt, length)) {
+                unsigned long err = ERR_get_error();
+                char * str = ERR_error_string(err, NULL);
+                msyslog(LOG_ERR, "decrypt: EVP_MAC_update() failed: %s.", str);
+                return false;
+        }
+        if (0 == EVP_MAC_final(ctx, mac, &len, sizeof(mac))) {
+                unsigned long err = ERR_get_error();
+                char * str = ERR_error_string(err, NULL);
+                msyslog(LOG_ERR, "decrypt: EVP_MAC_final() failed: %s.", str);
+                return false;
+        }
+#else
+	CMAC_CTX *ctx = cmac_ctx;
 	if (!CMAC_Init(ctx, auth->key, auth->key_size, auth->cipher, NULL)) {
 		/* Shouldn't happen.  Does if wrong key_size. */
 		msyslog(LOG_ERR,
-		    "CMAC: decrypt: CMAC init failed, %u, %u",
+		    "decrypt: CMAC init failed, %u, %u",
 			auth->keyid, auth->key_size);
 		return false;
 	}
 	CMAC_Update(ctx, (uint8_t *)pkt, (unsigned int)length);
 	CMAC_Final(ctx, mac, &len);
+#endif
 	if (MAX_BARE_MAC_LENGTH < len)
 		len = MAX_BARE_MAC_LENGTH;
+
 	if ((unsigned int)size != len + 4) {
 		/* Beware of DoS attack.
 		 * This indicates either the sender is broken
@@ -117,10 +194,10 @@ cmac_decrypt(
 		 * Similar code at digest_decrypt.
 		 */
 		if (0) msyslog(LOG_ERR,
-		    "CMAC: decrypt: MAC length error");
+		    "MAC: decrypt: MAC length error");
 		return false;
 	}
-	return ctmemeq(mac, (char *)pkt + length + 4, len);
+	return !CRYPTO_memcmp(mac, (char *)pkt + length + 4, len);
 }
 
 /*
@@ -144,7 +221,6 @@ digest_encrypt(
 	 * key type and digest type have been verified when the key
 	 * was created.
 	 */
-	EVP_MD_CTX_reset(ctx);
 	if (!EVP_DigestInit_ex(ctx, auth->digest, NULL)) {
 		msyslog(LOG_ERR,
 		    "MAC: encrypt: digest init failed");
@@ -182,7 +258,6 @@ digest_decrypt(
 	 * key type and digest type have been verified when the key
 	 * was created.
 	 */
-	EVP_MD_CTX_reset(ctx);
 	if (!EVP_DigestInit_ex(ctx, auth->digest, NULL)) {
 		msyslog(LOG_ERR,
 		    "MAC: decrypt: digest init failed");
@@ -203,7 +278,7 @@ digest_decrypt(
 		    "ERR: decrypt: digest length error");
 		return false;
 	}
-	return ctmemeq(digest, (char *)pkt + length + 4, len);
+	return !CRYPTO_memcmp(digest, (char *)pkt + length + 4, len);
 }
 
 /*
@@ -218,13 +293,43 @@ addr2refid(sockaddr_u *addr)
 {
 	uint8_t		digest[MD5_DIGEST_LENGTH];
 	uint32_t	addr_refid;
-	EVP_MD_CTX	*ctx;
+	static EVP_MD_CTX	*ctx;
 	unsigned int	len;
 
 	if (IS_IPV4(addr))
 		return (NSRCADR(addr));
 
-	ctx = EVP_MD_CTX_create();
+#if OPENSSL_VERSION_NUMBER >= 0x30000000L
+	static OSSL_LIB_CTX *libctx;
+	EVP_MD *md5 = 0;
+	if(libctx == NULL) {
+		libctx = OSSL_LIB_CTX_new();
+		if(libctx == NULL) {
+			msyslog(LOG_ERR, "MAC: MD5 init failed");
+			exit(1);
+		}
+	}
+	if(ctx == NULL) {
+		ctx = EVP_MD_CTX_new();
+		if(ctx == NULL) {
+			msyslog(LOG_ERR, "MAC: MD5 init failed");
+			exit(1);
+		}
+	}
+	/* See section FIPS Provider:
+	 * https://www.openssl.org/docs/man3.0/man7/crypto.html
+	 * for property query strings
+	 */
+	md5 = EVP_MD_fetch(libctx, "MD5", "fips=no");
+	if(!EVP_DigestInit_ex(ctx, md5, NULL)) {
+		msyslog(LOG_ERR, "MAC: MD5 init failed");
+		exit(1);
+	}
+	EVP_MD_free(md5);
+#else
+	if(ctx == NULL) {
+		ctx = EVP_MD_CTX_new();
+	}
 #ifdef EVP_MD_CTX_FLAG_NON_FIPS_ALLOW
 	/* MD5 is not used as a crypto hash here. */
 	EVP_MD_CTX_set_flags(ctx, EVP_MD_CTX_FLAG_NON_FIPS_ALLOW);
@@ -233,11 +338,10 @@ addr2refid(sockaddr_u *addr)
 		msyslog(LOG_ERR, "MAC: MD5 init failed");
 		exit(1);
 	}
-
+#endif
 	EVP_DigestUpdate(ctx, (uint8_t *)PSOCK_ADDR6(addr),
 	    sizeof(struct in6_addr));
 	EVP_DigestFinal_ex(ctx, digest, &len);
-	EVP_MD_CTX_destroy(ctx);
 	memcpy(&addr_refid, digest, sizeof(addr_refid));
 	return (addr_refid);
 }
