@@ -1,5 +1,10 @@
-#!/usr/bin/env python
+#! @PYSHEBANG@
 # -*- coding: utf-8 -*-
+
+# Copyright the NTPsec project contributors
+#
+# SPDX-License-Identifier: BSD-2-Clause
+
 '''
 ntpkeygen - generate cryptographic keys for NTP clients and servers
 
@@ -23,11 +28,34 @@ from __future__ import print_function
 import os
 import sys
 import socket
-import random
 import time
 import getopt
 import stat
 
+try:
+    import secrets
+    def gen_key(bytes, asciified=True):
+        if asciified:
+            result = ''
+            for index in range(bytes):
+                # Start ASCII characters with 0x24 so as not to include comment-beginning #
+                result += chr(0x24 + secrets.randbelow(0x5a))
+            return result
+        else:
+            return secrets.token_hex(bytes)
+except ImportError:
+    import random
+    def gen_key(bytes, asciified=True):
+        result = ''
+        if asciified:
+            for index in range(bytes):
+                # Start ASCII characters with 0x24 so as not to include comment-beginning #
+                result += chr(random.randint(0x24, 0x7e))
+        else:
+            for index in range(bytes):
+                result += "%02x" % random.randint(0x0, 0xff)
+        return result
+
 #
 # Cryptodefines
 #
@@ -39,18 +67,10 @@ def gen_keys(ident, groupname):
     "Generate semi-random AES keys for versions of ntpd with CMAC support."
     with fheader("AES", ident, groupname) as wp:
         for i in range(1, NUMKEYS+1):
-            key = ""
-            for j in range(KEYSIZE):
-                while True:
-                    r = randomizer.randint(0x21, 0x7e)
-                    if r != ord('#'):
-                        break
-                key += chr(r)
+            key = gen_key(KEYSIZE, True)
             wp.write("%2d AES %s\n" % (i, key))
         for i in range(1, NUMKEYS+1):
-            key = ""
-            for j in range(KEYSIZE):
-                key += "%02x" % randomizer.randint(0x00, 0xff)
+            key = gen_key(KEYSIZE, False)
             wp.write("%2d AES %s\n" % (i + NUMKEYS, key))
 
 
@@ -84,17 +104,14 @@ def fheader(fileid,     # file name id
 
 if __name__ == '__main__':
     try:
-        (options, arguments) = getopt.getopt(sys.argv[1:], "hMV",
+        (options, arguments) = getopt.getopt(sys.argv[1:], "hV",
                                              ["help", "version"])
     except getopt.GetoptError as e:
         print(e)
         raise SystemExit(1)
 
     for (switch, val) in options:
-        if switch == '-M':
-            # dummy MD5 option for backwards compatibility, ignored
-            pass
-        elif switch in ("-h", "--help"):
+        if switch in ("-h", "--help"):
             print("usage: ntpkeygen")
             raise SystemExit(0)
         elif switch in ("-V", "--version"):
@@ -103,7 +120,6 @@ if __name__ == '__main__':
 
     # The seed is ignored by random.SystemRandom,
     # even though the docs do not say so.
-    randomizer = random.SystemRandom()
     gen_keys("AES", socket.gethostname())
     raise SystemExit(0)
 
