@@ -1,7 +1,10 @@
-#!/usr/bin/env python
+#! @PYSHEBANG@
 # -*- coding: utf-8 -*-
 
+# Copyright the NTPsec project contributors
+#
 # SPDX-License-Identifier: BSD-2-Clause
+
 '''\
 Any keystroke causes a poll and update. Keystroke commands:
 
@@ -19,7 +22,7 @@ Any keystroke causes a poll and update. Keystroke commands:
 'u': Toggle display of units.
 'w': Toggle wide mode.
 'x': Cleanly terminate the program.
-' ': Rotate through a/n/o/p display modes.
+' ': Rotate through a/n/o/p/r display modes.
 '+': Increase debugging level.  Output goes to ntpmon.log
 '-': Decrease debugging level.
 '?': Display helpscreen.
@@ -212,23 +215,21 @@ USAGE: ntpmon [-dhnuV] [-D lvl] [-l logfile] [host]
 
 if __name__ == '__main__':
     bin_ver = "ntpsec-@NTPSEC_VERSION_EXTENDED@"
-    if ntp.util.stdversion() != bin_ver:
-        sys.stderr.write("Module/Binary version mismatch\n")
-        sys.stderr.write("Binary: %s\n" % bin_ver)
-        sys.stderr.write("Module: %s\n" % ntp.util.stdversion())
+    ntp.util.stdversioncheck(bin_ver)
     try:
         (options, arguments) = getopt.getopt(sys.argv[1:],
-                                             "dD:hl:nuV",
+                                             "dD:hl:nsSuV",
                                              ["debug", "help", "logfile=",
                                               "numeric", "units",
-                                              "set-debug-level=", "version"])
+                                              "set-debug-level=", "version",
+                                              "srcname", "srcnumber"])
     except getopt.GetoptError as e:
         sys.stderr.write("%s\n" % e)
         sys.stderr.write(usage)
         raise SystemExit(1)
     progname = sys.argv[0]
 
-    showhostnames = True
+    showhostnames = 1
     wideremote = False
     showall = True
     showpeers = True
@@ -252,7 +253,11 @@ if __name__ == '__main__':
                 logfp.close()
             logfp = open(val, "a", 1)  # 1 => line buffered
         elif switch in ("-n", "--numeric"):
-            showhostnames = False
+            showhostnames = 0
+        elif switch in ("-s", "--srcname"):
+            showhostnames = 3
+        elif switch in ("-S", "--srcnumber"):
+            showhostnames = 2
         elif switch in ("-u", "--units"):
             showunits = True
         elif switch in ("-V", "--version"):
@@ -419,13 +424,22 @@ if __name__ == '__main__':
                         showpeers = not showpeers
                         detailmode = False  # detail + hide peers == crash
                     elif key == 'n':
-                        peer_report.showhostnames = \
-                            not peer_report.showhostnames
-                        mru_report.showhostnames = not mru_report.showhostnames
+                        if peer_report.showhostnames == 2:
+                            peer_report.showhostnames = 0
+                        elif peer_report.showhostnames == 0:
+                            peer_report.showhostnames = 1
+                        elif peer_report.showhostnames == 1:
+                            peer_report.showhostnames = 3
+                        else:
+                            peer_report.showhostnames = 2
+                        mru_report.showhostnames = \
+                            peer_report.showhostnames
                     elif key == 'o':
                         peer_report.displaymode = 'opeers'
                     elif key == 'p':
                         peer_report.displaymode = 'peers'
+                    elif key == 'r':
+                        peer_report.displaymode = 'rpeers'
                     elif key == 's':
                         showall = not showall
                     elif key == 'u':
@@ -440,6 +454,8 @@ if __name__ == '__main__':
                             peer_report.displaymode = 'apeers'
                         elif peer_report.displaymode == 'apeers':
                             peer_report.displaymode = 'opeers'
+                        elif peer_report.displaymode == 'opeers':
+                            peer_report.displaymode = 'rpeers'
                         else:
                             peer_report.displaymode = 'peers'
                     elif key == 'j' or key == "KEY_DOWN":
