@@ -1,4 +1,4 @@
-#!/usr/bin/env python
+#! @PYSHEBANG@
 # -*- coding: utf-8 -*-
 #
 # ntpq - query an NTP server using mode 6 commands
@@ -8,13 +8,15 @@
 # mainly command interpretation and a back-end that presents the take
 # from ntpd as objects that can be re-used by other front
 # ends. Reusable pieces live in pylib.
+
+# Copyright the NTPsec project contributors
 #
 # SPDX-License-Identifier: BSD-2-Clause
+
 from __future__ import print_function, division
 
 import cmd
 import getopt
-import hashlib
 import os
 import re
 import resource
@@ -54,6 +56,8 @@ NTP_LFP = 0x7     # NTP timestamp
 NTP_MODE = 0x8    # peer mode
 NTP_2BIT = 0x9    # leap bits
 NTP_FLOAT = 0xa   # Float value
+NTP_UPTIME = 0xb  # uptime in daysD H:M:S (usually no frac)
+NTP_PACKETS = 0xc # packet counts
 
 
 class Ntpq(cmd.Cmd):
@@ -71,7 +75,7 @@ class Ntpq(cmd.Cmd):
         #  so I am leaving them, and possibly duplicating them.
         self.rawmode = False            # Flag which indicates raw mode output.
         self.directmode = False         # Flag for direct MRU output.
-        self.showhostnames = True       # If false, display numeric IPs
+        self.showhostnames = 1          # If & 1, display names
         self.showunits = False          # If False, show old style float
         self.auth_delay = 20            # delay time (default 20msec)
         self.wideremote = False         # show wide remote names?
@@ -257,7 +261,8 @@ usage: help [ command ]
                     continue
                 if len(self.chosts) > 1:
                     self.say(ntp.util.PeerSummary.high_truncate(
-                             self.session.hostname, maxhostlen) + " ")
+                             self.session.hostname, maxhostlen) + " " *
+                             (maxhostlen + 1 - len(self.session.hostname)))
                 self.say(report.summary(self.session.rstatus,
                                         variables, peer.associd))
         except KeyboardInterrupt:
@@ -325,8 +330,7 @@ usage: help [ command ]
             # files a bug, Mode 6 never seems to generate those in
             # variable fetches.
             text = ntp.poly.polystr(session.response.replace(
-                ntp.poly.polybytes(",\r\n"),
-                ntp.poly.polybytes(",\n")))
+                ",\r\n", ",\n"))
         else:
             if not quiet:
                 self.say("status=%04x %s,\n"
@@ -394,6 +398,83 @@ function: set the primary receive time out
 usage: timeout [ msec ]
 """)
 
+    def collect_display2(self, variables):
+        "Query and display a collection of variables from the system."
+        try:
+            queried = self.session.readvar(0,
+                                           [v[0] for v in variables] +
+                                           [v[0] + '_r' for v in variables],
+                                           raw=True)
+        except ntp.packet.ControlException as e:
+            if ntp.control.CERR_UNKNOWNVAR == e.errorcode:
+                self.warn("Unknown variable.  Trying one at a time.")
+                varlist = [v[0] for v in variables] + \
+                          [v[0] + '_r' for v in variables]
+                items = []
+                for var in varlist:
+                    try:
+                        queried = self.session.readvar(0, [var],
+                                                       raw=True)
+                        for (name, (value, rawvalue)) in queried.items():
+                            items.append((name, (value, rawvalue)))
+                    except ntp.packet.ControlException as e:
+                        if ntp.control.CERR_UNKNOWNVAR == e.errorcode:
+                            items.append((var, ("???", None)))
+                            continue
+                        raise e
+                queried = ntp.util.OrderedDict(items)
+            else:
+                self.warn(e.message)
+                return
+        except IOError as e:
+            self.warn(e.strerror)
+            return
+        if self.rawmode:
+            self.say(self.session.response)
+            return
+        try:
+            runs, runl = None, None
+            for (name, legend, fmt) in variables:
+                if name not in queried:
+                    continue
+                value = queried[name][0]
+                rawvalue = queried[name][1]
+                value2 = queried[name + '_r'][0]
+                rawvalue2 = queried[name + '_r'][1]
+                if fmt in (NTP_UINT, NTP_INT, NTP_FLOAT):
+                    if self.showunits and isinstance(rawvalue, (int, float)):
+                        display = ntp.util.unitifyvar(rawvalue, name)
+                    else:
+                        display = value
+                    if self.showunits and isinstance(rawvalue2, (int, float)):
+                        display2 = ntp.util.unitifyvar(rawvalue2, name)
+                    else:
+                        display2 = value2
+                        self.say(
+                            "{0:<13} {1:>13} {2:>13}\n".format(
+                                legend, display, display2)
+                        )
+                elif fmt == NTP_PACKETS:
+                    self.say(
+                        "{0:<13} {1[0]:>13} {2[0]:>13} {1[1]:>10}{1[2]:<3} {2[1]:>10}{2[2]:<3}\n".format(
+                                legend,
+                                ntp.util.packetize(value, runs),
+                                ntp.util.packetize(value2, runl),
+                        )
+                    )
+                elif fmt == NTP_UPTIME:
+                    runs, display = ntp.util.periodize(value)
+                    runl, display2 = ntp.util.periodize(value2)
+                    self.say(
+                        "{0:<13} {1:>13} {2:>13}\n".format(
+                            legend, display, display2)
+                    )
+                else:
+                    self.warn("unexpected vc type %s for %s, value %s %s    "
+                              % (fmt, name, value, value2))
+        except KeyboardInterrupt:
+            self.warn("display interrupted")
+
     def collect_display(self, associd, variables, decodestatus):
         "Query and display a collection of variables from the system."
         try:
@@ -435,33 +516,35 @@ usage: timeout [ msec ]
                      % (associd, self.session.rstatus,
                         ntp.ntpc.statustoa(statype, self.session.rstatus)))
         try:
+            run = 1
             for (name, legend, fmt) in variables:
                 if name not in queried:
                     continue
                 value = queried[name][0]
                 rawvalue = queried[name][1]
                 if fmt in (NTP_ADD, NTP_ADP):
-                    if self.showhostnames:
+                    if self.showhostnames & 1:  # if & 1, display names
                         if self.debug:
                             self.say("DNS lookup begins...")
                         value = ntp.util.canonicalize_dns(
                             value, family=self.ai_family)
                         if self.debug:
                             self.say("DNS lookup complete.")
-                    self.say("%s  %s\n" % (legend, value))
+                    self.say("%13s %13s\n" % (legend, value))
                 elif fmt == NTP_STR:
                     if value:
-                        self.say("%s  %s\n" % (legend, value))
+                        self.say("%13s %13s\n" % (legend, value))
                 elif fmt in (NTP_UINT, NTP_INT, NTP_FLOAT):
                     if self.showunits:
                         displayvalue = ntp.util.unitifyvar(rawvalue, name)
                     else:
                         displayvalue = value
-                    self.say("%s  %s\n" % (legend, displayvalue))
+                    self.say("%13s %13s\n" % (legend, displayvalue))
                 elif fmt == NTP_LFP:
-                    self.say("%s  %s\n" % (legend, ntp.ntpc.prettydate(value)))
+                    self.say("%13s %13s\n" % (
+                        legend, ntp.ntpc.prettydate(value)))
                 elif fmt == NTP_2BIT:
-                    self.say("%s  %s\n"
+                    self.say("%13s %13s\n"
                              % (legend, ("00", "01", "10", "11")[value]))
                 elif fmt == NTP_MODE:
                     modes = (
@@ -473,6 +556,17 @@ usage: timeout [ msec ]
                         self.say("%s  %s\n" % (legend, modes[value]))
                     except IndexError:
                         self.say("%s  %s%d\n" % (legend, "mode#", value))
+                    except TypeError:
+                        self.say("%s  mode: %s\n" % (legend, repr(value)))
+                elif fmt == NTP_PACKETS:
+                    self.say(
+                        "{0:<13} {1[0]:>13} {1[1]:>10}{1[2]:<3}\n".format(
+                            legend, ntp.util.packetize(value, run)
+                        )
+                    )
+                elif fmt == NTP_UPTIME:
+                    run, display = ntp.util.periodize(value)
+                    self.say("{0:<13} {1:>13}\n".format(legend, display))
                 else:
                     self.warn("unexpected vc type %s for %s, value %s"
                               % (fmt, name, value))
@@ -562,20 +656,29 @@ usage: passwd []
         if not line:
             pass
         elif line == "yes":
-            self.showhostnames = True
+            self.showhostnames = 1
         elif line == "no":
-            self.showhostnames = False
+            self.showhostnames = 0
+        elif line == 'hostnum':
+            self.showhostnames = 2
+        elif line == 'hostname':
+            self.showhostnames = 3
         else:
             self.say("What?\n")
-        if self.showhostnames:
-            self.say("hostnames being shown\n")
+            pass
+        if self.showhostnames & 1:
+            self.say('resolved hostnames being shown\n')
         else:
-            self.say("hostnames not being shown\n")
+            self.say('resolved hostnames not being shown\n')
+        if self.showhostnames & 2:
+            self.say('supplied hostnames being shown\n')
+        else:
+            self.say('supplied hostnames not being shown\n')
 
     def help_hostnames(self):
         self.say("""\
 function: specify whether hostnames or net numbers are printed
-usage: hostnames [yes|no]
+usage: hostnames [yes|no|hostname|hostnum]
 """)
 
     def do_debug(self, line):
@@ -751,15 +854,14 @@ usage: ntpversion [version number]
         "set key type to use for authenticated requests"
         if not line:
             self.say("Keytype: %s\n" % self.session.keytype)
-        elif line not in "DSA, MD4, MD5, MDC2, RIPEMD160, SHA-1, AES-CMAC":
-            # Above list is somewhat bogus. All but oldest versions of NTPsec
-            # will cheerfully use any 16- or 20-bit MAC supported by libcrypto;
-            # NTP Classic will probably barf on AES-CMAC.
-            self.warn("Keytype %s is not supported by ntpd.\n" % line)
-        elif line not in hashlib.algorithms_available:
-            self.warn("Keytype %s is not supported by ntpq.\n" % line)
+        elif line.upper() in ['AES', 'AES128CMAC']:
+            self.session.keytype = 'AES-128'
+        elif line.upper() in ['SHA-1']:
+            self.session.keytype = 'SHA1'
+        elif not ntp.ntpc.checkname(line.upper()):
+            self.warn("Keytype %s is not supported by openSSL or ntpq.\n" % line)
         else:
-            self.session.keytype = line
+            self.session.keytype = line.upper()
 
     def help_keytype(self):
         self.say("""\
@@ -915,8 +1017,10 @@ usage: writelist [ assocID ]
         associd = self.__assoc_valid(line)
         if associd >= 0:
             qtype = ntp.ntpc.TYPE_SYS if associd == 0 else ntp.ntpc.TYPE_PEER
+            silence = bool(len(line.split()) >= 2)
+            # Some scripts written for C ntpq need associd printed here
             self.__dolist(line.split()[1:], associd,
-                          ntp.control.CTL_OP_READVAR, qtype, quiet=True)
+                          ntp.control.CTL_OP_READVAR, qtype, quiet=silence)
 
     def help_readvar(self):
         self.say("""\
@@ -1071,8 +1175,8 @@ usage: cv [ assocID ] [ name=value[,...] ]
         pstats = (
             ("srcadr", "remote host:          ", NTP_ADD),
             ("dstadr", "local address:        ", NTP_ADD),
-            ("timerec", "time last received:   ", NTP_INT),
-            ("timer", "time until next send: ", NTP_INT),
+            ("timerec", "time last received:   ", NTP_UPTIME),
+            ("timer", "time until next send: ", NTP_UPTIME),
             ("timereach", "reachability change:  ", NTP_INT),
             ("sent", "packets sent:         ", NTP_INT),
             ("received", "packets received:     ", NTP_INT),
@@ -1106,6 +1210,16 @@ usage: pstats assocID
         self.say("""\
 function: obtain and print a list of the server's peers [IP version]
 usage: peers
+""")
+
+    def do_rpeers(self, line):
+        "obtain and print a list of the server's peers (dextral)"
+        self.__dopeers(showall=True, mode="rpeers")
+
+    def help_rpeers(self):
+        self.say("""\
+function: obtain and print a list of the server's peers (dextral)
+usage: rpeers
 """)
 
     def do_apeers(self, line):
@@ -1170,6 +1284,7 @@ usage: lopeers
             self.warn("In Config\nKeyword = :config\nCommand = %s" % line)
         try:
             self.session.config(line)
+            self.session.response = ntp.poly.polystr(self.session.response)
             m = re.match("column ([0-9]+) syntax error", self.session.response)
             if m:
                 col = int(m.group(1))
@@ -1208,6 +1323,42 @@ usage: config_from_file <configuration filename>
         for entry in entries:
             self.say(self.formatter.summary(entry) + "\n")
 
+    def do_noflake(self):
+        """Disables the dropping of control packets by ntpq for testing."""
+        self.session.flakey = False
+
+    def help_noflake(self):
+        """Print help for noflake."""
+        self.say("""\
+function: Disables the dropping of received packets by ntpq for testing.
+usage: noflake
+""")
+
+    def do_doflake(self, line):
+        """Drop some received packets for testing.
+
+        Probabilities 0 and 1 should be certainly accepted
+        and discarded respectively. No default, but 0.1
+        should be a one in ten loss rate.
+        """
+        try:
+            _ = float(line)
+            if not 0 < _ < 1:
+                raise ValueError
+            self.session.flakey = _
+        except ValueError:
+            self.say('Flakiness must be a (positive) float less than 1.')
+
+    def help_doflake(self):
+        """Print help for doflake."""
+        self.say("""\
+function: Enables the dropping of control packets by
+ntpq for testing. Probabilities 0 and 1 should be
+certainly accepted and discarded respectively. No
+default, but 0.1 should be a one in ten loss rate.
+usage: doflake <probability>
+""")
+
     def do_mrulist(self, line):
         """display the list of most recently seen source addresses,
            tags mincount=... resall=0x... resany=0x..."""
@@ -1409,21 +1560,28 @@ usage: kerninfo
     def do_sysstats(self, _line):
         "display system uptime and packet counts"
         sysstats = (
-            ("ss_uptime", "uptime:               ", NTP_INT),
-            ("ss_reset", "sysstats reset:       ", NTP_INT),
-            ("ss_received", "packets received:     ", NTP_INT),
-            ("ss_thisver", "current version:      ", NTP_INT),
-            ("ss_oldver", "older version:        ", NTP_INT),
+            ("ss_uptime",    "uptime:               ", NTP_UPTIME),
             ("ss_numctlreq", "control requests:     ", NTP_INT),
-            ("ss_badformat", "bad length or format: ", NTP_INT),
-            ("ss_badauth", "authentication failed:", NTP_INT),
-            ("ss_declined", "declined:             ", NTP_INT),
-            ("ss_restricted", "restricted:           ", NTP_INT),
-            ("ss_limited", "rate limited:         ", NTP_INT),
-            ("ss_kodsent", "KoD responses:        ", NTP_INT),
-            ("ss_processed", "processed for time:   ", NTP_INT),
+        )
+        sysstats2 = (
+            ("ss_reset",     "sysstats reset:       ", NTP_UPTIME),
+            ("ss_received",  "packets received:     ", NTP_PACKETS),
+            ("ss_thisver",   "current version:      ", NTP_PACKETS),
+            ("ss_oldver",    "older version:        ", NTP_PACKETS),
+            ("ss_ver1",      "NTPv1 total:          ", NTP_PACKETS),
+            ("ss_ver1client"," NTPv1 clients:       ", NTP_PACKETS),
+            ("ss_ver1zero",  " NTPv1 mode0:         ", NTP_PACKETS),
+            ("ss_ver1symm",  " NTPv1 symm act:      ", NTP_PACKETS),
+            ("ss_badformat", "bad length or format: ", NTP_PACKETS),
+            ("ss_badauth",   "authentication failed:", NTP_PACKETS),
+            ("ss_declined",  "declined:             ", NTP_PACKETS),
+            ("ss_restricted","restricted:           ", NTP_PACKETS),
+            ("ss_limited",   "rate limited:         ", NTP_PACKETS),
+            ("ss_kodsent",   "KoD responses:        ", NTP_PACKETS),
+            ("ss_processed", "processed for time:   ", NTP_PACKETS),
         )
         self.collect_display(associd=0, variables=sysstats, decodestatus=False)
+        self.collect_display2(variables=sysstats2)
 
     def help_sysstats(self):
         self.say("""\
@@ -1436,21 +1594,22 @@ usage: sysstats
     def do_monstats(self, _line):
         "display monitor (mrulist) counters and limits"
         monstats = (
-            ("mru_enabled", "enabled:              ", NTP_INT),
-            ("mru_depth", "addresses:            ", NTP_INT),
-            ("mru_deepest", "peak addresses:       ", NTP_INT),
-            ("mru_maxdepth", "maximum addresses:    ", NTP_INT),
-            ("mru_mindepth", "reclaim above count:  ", NTP_INT),
-            ("mru_maxage", "reclaim maxage:       ", NTP_INT),
-            ("mru_minage", "reclaim minage:       ", NTP_INT),
-            ("mru_mem", "kilobytes:            ", NTP_INT),
-            ("mru_maxmem", "maximum kilobytes:    ", NTP_INT),
-            ("mru_exists", "alloc: exists:        ", NTP_INT),
-            ("mru_new", "alloc: new:           ", NTP_INT),
-            ("mru_recycleold", "alloc: recycle old:   ", NTP_INT),
+            ("mru_enabled",     "enabled:              ", NTP_INT),
+            ("mru_hashslots",   "hash slots in use:    ", NTP_INT),
+            ("mru_depth",       "addresses in use:     ", NTP_INT),
+            ("mru_deepest",     "peak addresses:       ", NTP_INT),
+            ("mru_maxdepth",    "maximum addresses:    ", NTP_INT),
+            ("mru_mindepth",    "reclaim above count:  ", NTP_INT),
+            ("mru_maxage",      "reclaim maxage:       ", NTP_UPTIME),
+            ("mru_minage",      "reclaim minage:       ", NTP_UPTIME),
+            ("mru_mem",         "kilobytes:            ", NTP_INT),
+            ("mru_maxmem",      "maximum kilobytes:    ", NTP_INT),
+            ("mru_exists",      "alloc: exists:        ", NTP_INT),
+            ("mru_new",         "alloc: new:           ", NTP_INT),
+            ("mru_recycleold",  "alloc: recycle old:   ", NTP_INT),
             ("mru_recyclefull", "alloc: recycle full:  ", NTP_INT),
-            ("mru_none", "alloc: none:          ", NTP_INT),
-            ("mru_oldest_age", "age of oldest slot:   ", NTP_INT),
+            ("mru_none",        "alloc: none:          ", NTP_INT),
+            ("mru_oldest_age",  "age of oldest slot:   ", NTP_UPTIME),
         )
         self.collect_display(associd=0, variables=monstats, decodestatus=False)
 
@@ -1465,19 +1624,19 @@ usage: monstats
     def do_authinfo(self, _line):
         "display symmetric authentication counters"
         authinfo = (
-            ("authreset",          "time since reset:    ", NTP_INT),
+            ("authreset",          "time since reset:    ", NTP_UPTIME),
             ("authkeys",           "stored keys:         ", NTP_INT),
             ("authfreek",          "free keys:           ", NTP_INT),
             ("authklookups",       "key lookups:         ", NTP_INT),
             ("authknotfound",      "keys not found:      ", NTP_INT),
-            ("authencrypts",       "encryptions:         ", NTP_INT),
-            ("authdigestencrypts", "digest encryptions:  ", NTP_INT),
-            ("authcmacencrypts",   "CMAC encryptions:    ", NTP_INT),
-            ("authdecrypts",       "decryptions:         ", NTP_INT),
-            ("authdigestdecrypts", "digest decryptions:  ", NTP_INT),
-            ("authdigestfails",    "digest failures:     ", NTP_INT),
-            ("authcmacdecrypts",   "CMAC decryptions:    ", NTP_INT),
-            ("authcmacfails",      "CMAC failures:       ", NTP_INT),
+            ("authencrypts",       "encryptions:         ", NTP_PACKETS),
+            ("authdigestencrypts", "digest encryptions:  ", NTP_PACKETS),
+            ("authcmacencrypts",   "CMAC encryptions:    ", NTP_PACKETS),
+            ("authdecrypts",       "decryptions:         ", NTP_PACKETS),
+            ("authdigestdecrypts", "digest decryptions:  ", NTP_PACKETS),
+            ("authdigestfails",    "digest failures:     ", NTP_PACKETS),
+            ("authcmacdecrypts",   "CMAC decryptions:    ", NTP_PACKETS),
+            ("authcmacfails",      "CMAC failures:       ", NTP_PACKETS),
             # Old variables no longer supported.
             # Interesting if looking at an old system.
             ("authkuncached",      "uncached keys:       ", NTP_INT),
@@ -1496,23 +1655,37 @@ usage: authinfo
     def do_ntsinfo(self, _line):
         "display NTS authentication counters"
         ntsinfo = (
-   ("nts_client_send",           "NTS client sends:          ", NTP_INT),
-   ("nts_client_recv_good",      "NTS client recvs good:     ", NTP_INT),
-   ("nts_client_recv_bad",       "NTS client recvs w error:  ", NTP_INT),
-   ("nts_server_recv_good",      "NTS server recvs good:     ", NTP_INT),
-   ("nts_server_recv_bad",       "NTS server recvs w error:  ", NTP_INT),
-   ("nts_server_send",           "NTS server sends:          ", NTP_INT),
-   ("nts_cookie_make",           "NTS make cookies:          ", NTP_INT),
-   ("nts_cookie_decode",         "NTS decode cookies:        ", NTP_INT),
-   ("nts_cookie_decode_old",     "NTS decode cookies old:    ", NTP_INT),
-   ("nts_cookie_decode_too_old", "NTS decode cookies too old:", NTP_INT),
-   ("nts_cookie_decode_error",   "NTS decode cookies error:  ", NTP_INT),
-   ("nts_ke_probes_good",        "NTS KE probes good:        ", NTP_INT),
-   ("nts_ke_probes_bad",         "NTS KE probes_bad:         ", NTP_INT),
-   ("nts_ke_serves_good",        "NTS KE serves good:        ", NTP_INT),
-   ("nts_ke_serves_bad",         "NTS KE serves_bad:         ", NTP_INT),
+   ("nts_client_send",           "NTS client sends:           ", NTP_UINT),
+   ("nts_client_recv_good",      "NTS client recvs good:      ", NTP_UINT),
+   ("nts_client_recv_bad",       "NTS client recvs w error:   ", NTP_UINT),
+   ("nts_server_recv_good",      "NTS server recvs good:      ", NTP_UINT),
+   ("nts_server_recv_bad",       "NTS server recvs w error:   ", NTP_UINT),
+   ("nts_server_send",           "NTS server sends:           ", NTP_UINT),
+   ("nts_cookie_make",           "NTS make cookies:           ", NTP_UINT),
+   ("nts_cookie_not_server",     "NTS cookies not server:     ", NTP_UINT),
+   ("nts_cookie_decode_total",   "NTS decode cookies total:   ", NTP_UINT),
+   ("nts_cookie_decode_current", " NTS decode cookies current:", NTP_UINT),
+   ("nts_cookie_decode_old",     " NTS decode cookies old:    ", NTP_UINT),
+   ("nts_cookie_decode_old2",    " NTS decode cookies old2:   ", NTP_UINT),
+   ("nts_cookie_decode_older",   " NTS decode cookies older:  ", NTP_UINT),
+   ("nts_cookie_decode_too_old", " NTS decode cookies too old:", NTP_UINT),
+   ("nts_cookie_decode_error",   "NTS decode cookies error:   ", NTP_UINT),
+  )
+        ntskeinfo = (
+   ("nts_ke_serves_good",        "NTS KE serves good:         ", NTP_UINT),
+   ("nts_ke_serves_good_wall",   "NTS KE serves good wall:    ", NTP_FLOAT),
+   ("nts_ke_serves_good_cpu",    "NTS KE serves good CPU:     ", NTP_FLOAT),
+   ("nts_ke_serves_nossl",       "NTS KE serves no-TLS:       ", NTP_UINT),
+   ("nts_ke_serves_nossl_wall",  "NTS KE serves no-TLS wall:  ", NTP_FLOAT),
+   ("nts_ke_serves_nossl_cpu",   "NTS KE serves no-TLS CPU:   ", NTP_FLOAT),
+   ("nts_ke_serves_bad",         "NTS KE serves bad:          ", NTP_UINT),
+   ("nts_ke_serves_bad_wall",    "NTS KE serves bad wall:     ", NTP_FLOAT),
+   ("nts_ke_serves_bad_cpu",     "NTS KE serves bad CPU:      ", NTP_FLOAT),
+   ("nts_ke_probes_good",        "NTS KE client probes good:  ", NTP_UINT),
+   ("nts_ke_probes_bad",         "NTS KE client probes bad:   ", NTP_UINT),
   )
         self.collect_display(associd=0, variables=ntsinfo, decodestatus=False)
+        self.collect_display(associd=0, variables=ntskeinfo, decodestatus=False)
 
     def help_ntsinfo(self):
         self.say("""\
@@ -1521,21 +1694,43 @@ usage: ntsinfo
 """)
 
 
+    def do_mssntpinfo(self, _line):
+        "display MSSNTP authentication counters"
+        mssntpinfo = (
+   ("mssntp_serves",              "MSSNTP requests processed:     ", NTP_UINT),
+   ("mssntp_serves_no",           "MSSNTP rejected by samba:      ", NTP_UINT),
+   ("mssntp_serves_err",          "MSSNTP errors talking to samba:", NTP_UINT),
+   ("mssntp_serves_good",         "MSSNTP requests good:          ", NTP_UINT),
+   ("mssntp_serves_good_wall",    "MSSNTP requests good wall:     ", NTP_FLOAT),
+   ("mssntp_serves_good_slowest", "MSSNTP requests good slowest:  ", NTP_FLOAT),
+   ("mssntp_serves_bad",          "MSSNTP requests bad:           ", NTP_UINT),
+   ("mssntp_serves_bad_wall",     "MSSNTP requests bad wall:      ", NTP_FLOAT),
+   ("mssntp_serves_bad_slowest",  "MSSNTP requests bad slowest:   ", NTP_FLOAT),
+  )
+        self.collect_display(associd=0, variables=mssntpinfo, decodestatus=False)
+
+    def help_mssntpinfo(self):
+        self.say("""\
+function: display MSSNTP authentication counters
+usage: mssntpinfo
+""")
+
+
 # FIXME: This table should move to ntpd
 #          so the answers track when ntpd is updated
     def do_iostats(self, _line):
         "display network input and output counters"
         iostats = (
-            ("iostats_reset", "time since reset:     ", NTP_INT),
+            ("iostats_reset", "time since reset:     ", NTP_UPTIME),
             ("total_rbuf", "receive buffers:      ", NTP_INT),
             ("free_rbuf", "free receive buffers: ", NTP_INT),
             ("used_rbuf", "used receive buffers: ", NTP_INT),
             ("rbuf_lowater", "low water refills:    ", NTP_INT),
-            ("io_dropped", "dropped packets:      ", NTP_INT),
-            ("io_ignored", "ignored packets:      ", NTP_INT),
-            ("io_received", "received packets:     ", NTP_INT),
-            ("io_sent", "packets sent:         ", NTP_INT),
-            ("io_sendfailed", "packet send failures: ", NTP_INT),
+            ("io_dropped", "dropped packets:      ", NTP_PACKETS),
+            ("io_ignored", "ignored packets:      ", NTP_PACKETS),
+            ("io_received", "received packets:     ", NTP_PACKETS),
+            ("io_sent", "packets sent:         ", NTP_PACKETS),
+            ("io_sendfailed", "packet send failures: ", NTP_PACKETS),
             ("io_wakeups", "input wakeups:        ", NTP_INT),
             ("io_goodwakeups", "useful input wakeups: ", NTP_INT),
         )
@@ -1552,7 +1747,7 @@ usage: iostats
     def do_timerstats(self, line):
         "display interval timer counters"
         timerstats = (
-            ("timerstats_reset", "time since reset:  ", NTP_INT),
+            ("timerstats_reset", "time since reset:  ", NTP_UPTIME),
             ("timer_overruns", "timer overruns:    ", NTP_INT),
             ("timer_xmts", "calls to transmit: ", NTP_INT),
         )
@@ -1603,19 +1798,16 @@ USAGE: ntpq [-46dphinOV] [-c str] [-D lvl] [host ...]
 
 if __name__ == '__main__':
     bin_ver = "ntpsec-@NTPSEC_VERSION_EXTENDED@"
-    if ntp.util.stdversion() != bin_ver:
-        sys.stderr.write("Module/Binary version mismatch\n")
-        sys.stderr.write("Binary: %s\n" % bin_ver)
-        sys.stderr.write("Module: %s\n" % ntp.util.stdversion())
+    ntp.util.stdversioncheck(bin_ver)
     try:
         (options, arguments) = getopt.getopt(
             sys.argv[1:],
-            "46a:c:dD:hk:npVwW:ul:",
+            "46a:c:dD:hk:npsSVwW:ul:",
             ["ipv4", "ipv6", "authentication=",
              "command=", "debug", "set-debug-level=",
              "help", "keyfile", "numeric", "peers",
              "version", "wide", "width=", "units",
-             "logfile="])
+             "logfile=", "srcname", "srcnumber"])
     except getopt.GetoptError as e:
         sys.stderr.write("%s\n" % e)
         sys.stderr.write(usage)
@@ -1650,11 +1842,15 @@ if __name__ == '__main__':
             sys.stderr.write(usage)
             raise SystemExit(0)
         elif switch in ("-n", "--numeric"):
-            interpreter.showhostnames = False
+            interpreter.showhostnames = 0
         elif switch in ("-p", "--peers"):
             interpreter.ccmds.append("peers")
         elif switch in ("-k", "--keyfile"):
             keyfile = val
+        elif switch in ("-s", "--srcname"):
+            interpreter.showhostnames = 3
+        elif switch in ("-S", "--srcnumber"):
+            interpreter.showhostnames = 2
         elif switch in ("-V", "--version"):
             sys.stdout.write("ntpq %s\n" % version)
             raise SystemExit(0)
@@ -1720,12 +1916,21 @@ if __name__ == '__main__':
             session.openhost(*interpreter.chosts[0])
             interpreter.cmdloop()
         else:
+            latter = False
             for ihost in interpreter.chosts:
                 if session.openhost(*ihost):
                     for command in interpreter.ccmds:
+                        if latter:
+                            print()
+                        latter = True
                         interpreter.onecmd(interpreter.precmd(command))
                     session.close()
         raise SystemExit(0)
+    except BrokenPipeError:
+        # https://docs.python.org/3/library/signal.html#note-on-sigpipe
+        devnull = os.open(os.devnull, os.O_WRONLY)
+        os.dup2(devnull, sys.stdout.fileno())
+        raise SystemExit(1)
     except (KeyboardInterrupt, EOFError):
         if os.isatty(0):
             interpreter.say("\n")
