@@ -1,4 +1,4 @@
-#!/usr/bin/env python
+#! @PYSHEBANG@
 # -*- coding: utf-8 -*-
 """
 ntptrace - trace peers of an NTP server
@@ -10,6 +10,9 @@ Usage: ntptrace [-n | --numeric] [-m number | --max-hosts=number]
 
 See the manual page for details.
 """
+
+# Copyright the NTPsec project contributors
+#
 # SPDX-License-Identifier: BSD-2-Clause
 
 from __future__ import print_function
@@ -117,10 +120,7 @@ Options are specified by doubled hyphens and their name or by a single
 hyphen and the flag character.""" + "\n"
 
 bin_ver = "ntpsec-@NTPSEC_VERSION_EXTENDED@"
-if ntp.util.stdversion() != bin_ver:
-    sys.stderr.write("Module/Binary version mismatch\n")
-    sys.stderr.write("Binary: %s\n" % bin_ver)
-    sys.stderr.write("Module: %s\n" % ntp.util.stdversion())
+ntp.util.stdversioncheck(bin_ver)
 
 try:
     (options, arguments) = getopt.getopt(
