@@ -1,4 +1,4 @@
-#!/usr/bin/env python
+#! @PYSHEBANG@
 # -*- coding: utf-8 -*-
 """\
 ntpviz - visualizer for NTP log files
@@ -36,7 +36,11 @@ See the manual page for details.
 
 Python by ESR, concept and gnuplot code by Dan Drown.
 """
+
+# Copyright the NTPsec project contributors
+#
 # SPDX-License-Identifier: BSD-2-Clause
+
 from __future__ import print_function, division
 
 import atexit
@@ -137,7 +141,7 @@ class MyArgumentParser(argparse.ArgumentParser):
     def convert_arg_line_to_args(self, arg_line):
         '''Make options file more tolerant'''
         # strip out trailing comments
-        arg_line = re.sub('\s+#.*$', '', arg_line)
+        arg_line = re.sub(r'\s+#.*$', '', arg_line)
 
         # ignore blank lines
         if not arg_line:
@@ -160,18 +164,35 @@ def print_profile():
 # use this until we can guarantee Python 3.4 and the statistics module
 # http://stackoverflow.com/questions/15389768/standard-deviation-of-a-list#21505523
 
-# class to calc:
-#   Mean, Variance, Standard Deviation, Skewness and Kurtosis
+# RunningStats(), a class to calculate:
+#     Mean, Variance, Standard Deviation, Skewness and Kurtosis
+#
+# References:
+#     https://en.wikipedia.org/wiki/Standard_deviation
+#
+#     NIST/SEMATECH e-Handbook of Statistical Methods, 2012
+#     1.3.5.11.  Measures of Skewness and Kurtosis
+#     https://www.itl.nist.gov/div898/handbook/eda/section3/eda35b.htm
 
 class RunningStats(object):
-    "Calculate mean, variance, sigma, skewness and kurtosis"
+    """Calculate mean, variance, sigma, skewness and kurtosis"""
 
     def __init__(self, values):
+        """Initialize the class, and do all the caculations."""
+
         self.num = len(values)     # number of samples
         self.mu = 0.0              # simple arithmetic mean
         self.variance = 0.0        # variance
-        self.sigma = 0.0           # aka standard deviation
+        # standard deviation, NOT the sample standard deviation
+        self.sigma = 0.0
+        # the Fisher-Pearson coefficient of skewness.
+        # NOT the adjusted Fisher-Pearson coefficient of skewness.
+        # NOT the Galton skewness (a.k.a. Bowley's skewness)
         self.skewness = 0.0
+        # the standard Kurtosis
+        # NOT the Alternative Kurtosis
+        # NOT the Excess Kurtosis
+        # Excess Kurtosis === standard Kurtosis - 3
         self.kurtosis = 3.0
 
         if 0 >= self.num:
@@ -182,8 +203,12 @@ class RunningStats(object):
         self.variance = sum(pow((v-self.mu), 2) for v in values) / self.num
         self.sigma = math.sqrt(self.variance)
 
-        if math.isnan(self.sigma) or 1e-12 >= abs(self.sigma):
-            # punt
+        # Note: math.isnan(float("+Inf")) is false, so isnan() and isinf()
+        # Python 2 hasa no math.isfinite()
+        if ((math.isinf(self.sigma) or
+             math.isnan(self.sigma) or
+             1e-12 >= abs(self.sigma))):
+            # Not finite, or too small.  Punt
             self.skewness = float('nan')
             self.kurtosis = float('nan')
             return
@@ -191,8 +216,8 @@ class RunningStats(object):
         m3 = 0
         m4 = 0
         for val in values:
-            m3 += pow(val - self.sigma, 3)
-            m4 += pow(val - self.sigma, 4)
+            m3 += pow(val - self.mu, 3)
+            m4 += pow(val - self.mu, 4)
 
         self.skewness = m3 / (self.num * pow(self.sigma, 3))
         self.kurtosis = m4 / (self.num * pow(self.sigma, 4))
@@ -397,9 +422,13 @@ def gnuplot(template, outfile=None):
     if not template:
         # silently ignore empty plots
         return ''
+
     if outfile is None:
         out = None
     else:
+        if 2 <= args.debug_level:
+            sys.stderr.write("ntpviz: INFO: sending plot output "
+                             "to %s\n" % outfile)
         out = open(outfile, "w", encoding='utf-8')
     ##
 
@@ -475,8 +504,8 @@ set rmargin 10
         if item2:
             for row in rows:
                 try:
-                    values1.append(float(row[item1]))
-                    values2.append(float(row[item2]))
+                    val1 = float(row[item1])
+                    val2 = float(row[item2])
                     if 2200000 < row[0] - last_time:
                         # more than 2,200 seconds between points
                         # data loss, add a break in the plot line
@@ -486,7 +515,12 @@ set rmargin 10
                         + row[item2] + '\n'
                     last_time = row[0]
                 except IndexError:
-                    pass
+                    continue
+                except ValueError:
+                    continue
+                # both values are good, append them.
+                values1.append(val1)
+                values2.append(val2)
         else:
             for row in rows:
                 try:
@@ -500,6 +534,8 @@ set rmargin 10
                     last_time = row[0]
                 except IndexError:
                     pass
+                except ValueError:
+                    pass
 
         # I know you want to replace the plot_data string concat with
         # or more join()s, do not do it, it is slower
@@ -778,7 +814,7 @@ set terminal %(terminal)s size %(size)s
 set title "%(sitename)s: Local GPS%(clipped)s
 set ytics format "%(fmt)s TDOP" nomirror textcolor rgb '#0060ad'
 set yrange [%(min_y)s:%(max_y)s]
-set y2tics format "%%2.0f nSat"  nomirror textcolor rgb '#dd181f'
+set y2tics format "%%4.1f nSat"  nomirror textcolor rgb '#dd181f'
 set style line 1 lc rgb '#0060ad' lt 1 lw 1 pt 7 ps 0   # --- blue
 set style line 2 lc rgb '#dd181f' lt 1 lw 1 pt 5 ps 0   # --- red
 set key top right
@@ -787,9 +823,9 @@ plot \\
 
         for key in gpslist:
             plot_template += """\
-'-' using 1:2 title '%(key)s TDOP' with line ls 1, \\
-'-' using 1:3 title '%(key)s nSat' with line ls 2 axis x1y2, \\
-""" % locals()
+'-' using 1:2 title '%s TDOP' with line ls 1, \\
+'-' using 1:3 title '%s nSat' with line ls 2 axis x1y2, \\
+""" % (key, key)
 
         # strip the trailing ", \\n"
         plot_template = plot_template[:-4] + "\n"
@@ -798,9 +834,10 @@ plot \\
 The number of visible satellites (nSat) is plotted in red.</p>
 <p>TDOP is field 3, and nSats is field 4, from the gpsd log file.  The
 gpsd log file is created by the ntploggps program.</p>
-<p>TDOP is a dimensionless error factor.  TDOP ranges from 1 to greater
-than 20.  1 denotes the highest possible confidence level.  2 to 5 is good.
-Greater than 20 means there will be significant inaccuracy and error.</p>
+<p>TDOP is a dimensionless error factor.  Smaller numbers are better.
+TDOP ranges from 1 (ideal), 2 to 5 (good), to greater than 20 (poor).
+Some GNSS receivers report TDOP less than one which is theoretically
+impossible.</p>
 """
 
         ret = {'html': VizStats.table_head + stats.table +
@@ -1283,14 +1320,14 @@ plot \
 def local_offset_multiplot(statlist):
     "Plot comparative local offsets for a list of NTPViz objects."
 
-    out = []
+    out = {}
     out['size'] = args.img_size
     out['terminal'] = args.terminal
 
     plot = NTPViz.Common + '''\
 set terminal %(terminal)s size %(size)s
 set title "Multiplot Local Clock Offsets"
-set ytics format "%1.2f μs" nomirror textcolor rgb "#0060ad"
+set ytics format "%%1.2f μs" nomirror textcolor rgb "#0060ad"
 set key bottom right box
 plot \\
 ''' % out
@@ -1306,7 +1343,7 @@ plot \\
     for stats in statlist:
         # speed up by only sending gnuplot the data it will actually use
         # fields: time, offset
-        pt = NTPViz.plot_slice(stats.loopstats, 2)
+        pt = stats.plot_slice(stats.loopstats, 2)
         plot_data += pt[0]
 
     ret = {'html': '', 'stats': []}
@@ -1377,11 +1414,7 @@ AADRxQAAxBEAAA==
 
 if __name__ == '__main__':
     bin_ver = "ntpsec-@NTPSEC_VERSION_EXTENDED@"
-    if ntp.util.stdversion() != bin_ver:
-        sys.stderr.write("ntpviz: WARNING: Module/Binary version mismatch\n")
-        sys.stderr.write("ntpviz: WARNING: Binary: %s\n" % bin_ver)
-        sys.stderr.write("ntpviz: WARNING: Module: %s\n" %
-                         ntp.util.stdversion())
+    ntp.util.stdversioncheck(bin_ver)
 
     parser = MyArgumentParser(description="ntpd stats visualizer",
                               fromfile_prefix_chars='@',
@@ -1607,7 +1640,10 @@ Python by ESR, concept and gnuplot code by Dan Drown.
         if 2 < args.debug_level:
             sys.stderr.write("ntpviz: INFO: now running at nice: %s\n" % nice)
 
-    for fontpath in ("/usr/share/fonts/liberation",
+    for fontpath in ("@PREFIX@/share/fonts/liberation",
+                     "@PREFIX@/share/fonts/liberation-fonts",
+                     "@PREFIX@/share/fonts/truetype/liberation",
+                     "/usr/share/fonts/liberation",
                      "/usr/share/fonts/liberation-fonts",
                      "/usr/share/fonts/truetype/liberation"):
 
@@ -1775,10 +1811,16 @@ tbody tr:nth-child(6n+6) {
 
     # Ugh.  Not clear what to do in the multiplot case
     if len(statlist) == 1:
-        start_time = datetime.datetime.utcfromtimestamp(
-            stats.starttime).strftime('%c')
-        end_time = datetime.datetime.utcfromtimestamp(
-            stats.endtime).strftime('%c')
+        try:
+            start_time = datetime.datetime.fromtimestamp(
+                stats.starttime, datetime.timezone.utc).strftime('%c')
+            end_time = datetime.datetime.fromtimestamp(
+                stats.endtime, datetime.timezone.utc).strftime('%c')
+        except AttributeError:  # No datetime.timezone in Python 2
+            start_time = datetime.datetime.utcfromtimestamp(
+                stats.starttime).strftime('%c')
+            end_time = datetime.datetime.utcfromtimestamp(
+                stats.endtime).strftime('%c')
 
         index_header += '<b>Start Time:</b> %s UTC<br>\n' \
                         '<b>End Time:</b> %s UTC<br>\n' \
@@ -1814,13 +1856,6 @@ Refclock Jitter, and Server Jitter in seconds.  Local Frequency Jitter is
 in ppm or ppb.
 </dd>
 
-<dt>kurtosis, Kurt:</dt>
-<dd>The kurtosis of a random variable X is the fourth standardized
-moment and is a dimension-less ratio. ntpviz uses the Pearson's moment
-coefficient of kurtosis.  A normal distribution has a kurtosis of three.
-NIST describes a kurtosis over three as "heavy tailed" and one under
-three as "light tailed".</dd>
-
 <dt>ms, millisecond:</dt>
 <dd>One thousandth of a second = 0.001 seconds, 1e-3 seconds</dd>
 
@@ -1871,12 +1906,21 @@ The formula for sigma is: "σ = √[ ∑(x<sub>i</sub>-mu)^2 / N ]".
 Where x<sub>i</sub> denotes the data points and N is the number of data
 points.</dd>
 
-<dt>skewness, Skew:</dt>
-<dd>The skewness of a random variable X is the third standardized
-moment and is a dimension-less ratio. ntpviz uses the Pearson's moment
-coefficient of skewness.  Wikipedia describes it best: "The qualitative
-interpretation of the skew is complicated and unintuitive."<br> A normal
-distribution has a skewness of zero. </dd>
+<dt>Skewness, Skew:</dt>
+<dd>The skewness of a random variable X is the third standardized moment
+and is a dimension-less ratio. ntpviz uses the FIsher-Pearson moment
+of skewness.  There are other different ways to calculate Skewness
+Wikipedia describes Skewness best: "The qualitative interpretation of
+the skew is complicated and unintuitive."
+<br>A normal distribution has a skewness of zero. </dd>
+
+<dt>Kurtosis, Kurt:</dt>
+<dd>The kurtosis of a random variable X is the fourth standardized
+moment and is a dimension-less ratio. ntpviz uses standard
+Kurtosis.  There are other different ways to calculate Kurtosis.
+<br>A normal distribution has a Kurtosis of three.  NIST describes a
+kurtosis over three as "heavy tailed" and one under three as "light
+tailed".</dd>
 
 <dt>upstream clock:</dt>
 <dd>Any server or reference clock used as a source of time.</dd>
