@@ -1,4 +1,4 @@
-#!/usr/bin/env python
+#! @PYSHEBANG@
 # -*- coding: utf-8 -*-
 """\
 ntpwait - Wait for ntpd to stabilize the system clock.
@@ -15,6 +15,9 @@ hyphen and the flag character.
 
 A spurious 'not running' message can result from queries being disabled.
 """
+
+# Copyright the NTPsec project contributors
+#
 # SPDX-License-Identifier: BSD-2-Clause
 
 # Because we've actually seen this fail on a ^C during import of ntp.packet.
@@ -52,10 +55,7 @@ class Unbuffered(object):
 
 if __name__ == "__main__":
     bin_ver = "ntpsec-@NTPSEC_VERSION_EXTENDED@"
-    if ntp.util.stdversion() != bin_ver:
-        sys.stderr.write("Module/Binary version mismatch\n")
-        sys.stderr.write("Binary: %s\n" % bin_ver)
-        sys.stderr.write("Module: %s\n" % ntp.util.stdversion())
+    ntp.util.stdversioncheck(bin_ver)
     try:
         (options, arguments) = getopt.getopt(sys.argv[1:], "hn:s:vV", [
             "tries=", "sleep=", "verbose", "help", "version"
