@@ -44,6 +44,8 @@
 int qos = IPTOS_DSCP_EF;	/* QoS RFC 3246 */
 
 
+uint16_t extra_port = 0;	/* 0 => not used */
+
 /*
  * NIC rule entry
  */
@@ -92,13 +94,17 @@ struct packet_counters {
 	uint64_t received;	/* total number of packets received */
 	uint64_t sent;		/* total number of packets sent */
 	uint64_t notsent;	/* total number of packets which couldn't be sent */
-	/* There used to be a signal handler for received packets. */
-	/* It's not needed now that the kernel time stamps packets. */
-	uint64_t handler_calls;	/* number of calls to interrupt handler */
-	uint64_t handler_pkts;	/* number of pkts received by handler */
-	uptime_t io_timereset;	/* time counters were reset */
+
+	uint64_t handler_calls; /* wakeups -- may batch packets */
+	uint64_t handler_pkts;  /* input packets -- redundant */
+
+#ifdef REFCLOCK
+	uint64_t handler_refrds;/* refclock reads */
+#endif
+
 };
 volatile struct packet_counters pkt_count;
+uptime_t io_timereset;	/* time counters were reset */
 
 /*
  * Interface stuff
@@ -106,11 +112,11 @@ volatile struct packet_counters pkt_count;
 struct ntp_io_data io_data;
 static int ninterfaces;			/* total # of interfaces */
 
-extern  SOCKET  open_socket     (sockaddr_u *, bool, endpt *);
+static  SOCKET  open_socket     (sockaddr_u *, bool, endpt *);
 
 static bool
 netaddr_eqprefix(const isc_netaddr_t *, const isc_netaddr_t *,
-                    unsigned int) __attribute__((pure));
+		unsigned int) __attribute__((pure));
 
 /* Socket Address */
 typedef struct isc_sockaddr {
@@ -120,13 +126,12 @@ typedef struct isc_sockaddr {
 		struct sockaddr_in6	sin6;
 	}				type;
 	unsigned int			length;		/* XXXRTH beginning? */
-        struct { struct isc_sockaddr *prev, *next; } link;
+	struct { struct isc_sockaddr *prev, *next; } link;
 } isc_sockaddr_t;
 
 static void
 netaddr_fromsockaddr(isc_netaddr_t *netaddr, const isc_sockaddr_t *source);
 
-
 #ifdef REFCLOCK
 /*
  * Refclock stuff.	We keep a chain of structures with data concerning
@@ -142,14 +147,12 @@ static	struct refclockio *refio;
 static fd_set activefds;
 static int maxactivefd;
 
-/*
- * bit alternating value to detect verified interfaces during an update cycle
- */
-static unsigned short		sys_interphase = 0;
-
 static void	add_interface(endpt *);
-static bool	update_interfaces(unsigned short, interface_receiver_t,
-				  void *);
+static bool	update_interfaces(void);
+static void	update_interfaces_phase0(void);
+static bool	update_interfaces_phase1(uint16_t port);
+static void	update_interfaces_phase2(void);
+static void	update_interfaces_phase3(void);
 static void	remove_interface(endpt *);
 static endpt *	create_interface(unsigned short, endpt *);
 
@@ -209,7 +212,7 @@ static void init_async_notifications (void);
 
 static	bool	addr_eqprefix	(const sockaddr_u *, const sockaddr_u *,
 				 int);
-static	int	create_sockets	(unsigned short);
+static	int	create_sockets	(void);
 static	void	set_reuseaddr	(int);
 
 typedef struct remaddr remaddr_t;
@@ -222,9 +225,6 @@ struct remaddr {
 
 static remaddr_t * remoteaddr_list;
 
-static endpt *	wildipv4;
-static endpt *	wildipv6;
-
 static const int accept_wildcard_if_for_winnt = false;
 
 static void	add_fd_to_list		(SOCKET, enum desc_type);
@@ -233,7 +233,7 @@ static void	delete_interface_from_list(endpt *);
 static void	close_and_delete_fd_from_list(SOCKET);
 static void	add_addr_to_list	(sockaddr_u *, endpt *);
 static void	create_wildcards	(unsigned short);
-static endpt *	findlocalinterface	(sockaddr_u *, int, int);
+static endpt *	findlocalinterface	(sockaddr_u *);
 static endpt *	findclosestinterface	(sockaddr_u *, int);
 
 #ifdef DEBUG
@@ -372,7 +372,7 @@ io_open_sockets(void)
 	/*
 	 * Create the sockets
 	 */
-	create_sockets(NTP_PORT);
+	create_sockets();
 
 	init_async_notifications();
 
@@ -392,10 +392,6 @@ interface_dump(const endpt *itf)
 	printf("fd = %d\n", itf->fd);
 	printf("sin = %s,\n", socktoa(&itf->sin));
 	sockaddr_dump(&itf->sin);
-	printf("bcast = %s,\n", socktoa(&itf->bcast));
-	sockaddr_dump(&itf->bcast);
-	printf("mask = %s,\n", socktoa(&itf->mask));
-	sockaddr_dump(&itf->mask);
 	printf("name = %s\n", itf->name);
 	printf("flags = 0x%08x\n", itf->flags);
 	printf("addr_refid = %08x\n", itf->addr_refid);
@@ -404,7 +400,6 @@ interface_dump(const endpt *itf)
 	printf("notsent = %ld\n", itf->notsent);
 	printf("ifindex = %u\n", itf->ifindex);
 	printf("peercnt = %u\n", itf->peercnt);
-	printf("phase = %u\n", itf->phase);
 }
 
 /*
@@ -435,24 +430,18 @@ static void
 print_interface(const endpt *iface, const char *pfx, const char *sfx)
 {
 	printf("%sinterface #%u: fd=%d, name=%s, "
-               "flags=0x%x, ifindex=%u, sin=%s",
-	       pfx,
-	       iface->ifnum,
-	       iface->fd,
-	       iface->name,
-	       iface->flags,
-	       iface->ifindex,
-	       socktoa(&iface->sin));
-	if (AF_INET == iface->family) {
-		if (iface->flags & INT_BROADCAST)
-			printf(", bcast=%s", socktoa(&iface->bcast));
-		printf(", mask=%s", socktoa(&iface->mask));
-	}
+		"flags=0x%x, ifindex=%u, sin=%s",
+		pfx,
+		iface->ifnum,
+		iface->fd,
+		iface->name,
+		iface->flags,
+		iface->ifindex,
+		sockporttoa(&iface->sin));
 	printf(", %s:%s",
-	       (iface->ignore_packets)
-		   ? "Disabled"
-		   : "Enabled",
-	       sfx);
+		(iface->ignore_packets)
+		? "Disabled" : "Enabled",
+		sfx);
 	if (debug > 4)	/* in-depth debugging only */ /* SPECIAL DEBUG */
 		interface_dump(iface);
 }
@@ -702,7 +691,6 @@ init_interface(
 {
 	ZERO(*ep);
 	ep->fd = INVALID_SOCKET;
-	ep->phase = sys_interphase;
 }
 
 
@@ -779,8 +767,8 @@ remove_interface(
 	if (ep->fd != INVALID_SOCKET) {
 		msyslog(LOG_INFO,
 			"IO: Deleting interface #%u %s, %s#%d, interface stats: "
-                        "received=%ld, sent=%ld, dropped=%ld, "
-                        "active_time=%lu secs",
+			"received=%ld, sent=%ld, dropped=%ld, "
+			"active_time=%lu secs",
 			ep->ifnum,
 			ep->name,
 			socktoa(&ep->sin),
@@ -799,7 +787,7 @@ remove_interface(
 	/* remove restrict interface entry */
 	SET_HOSTMASK(&resmask, AF(&ep->sin));
 	hack_restrict(RESTRICT_REMOVEIF, &ep->sin, &resmask,
-		      RESM_NTPONLY | RESM_INTERFACE, RES_IGNORE, 0);
+		      RESM_NTPONLY | RESM_INTERFACE, RES_IGNORE);
 }
 
 
@@ -860,26 +848,26 @@ create_wildcards(
 		strlcpy(wildif->name, "v6wildcard", sizeof(wildif->name));
 		memcpy(&wildif->sin, &wildaddr, sizeof(wildif->sin));
 		wildif->family = AF_INET6;
-		AF(&wildif->mask) = AF_INET6;
-		SET_ONESMASK(&wildif->mask);
 
 		wildif->flags = INT_UP | INT_WILDCARD;
 		wildif->ignore_packets = (ACTION_DROP == action);
 
 		wildif->fd = open_socket(&wildif->sin, true, wildif);
 
-		if (wildif->fd != INVALID_SOCKET) {
-			wildipv6 = wildif;
-			io_data.any6_interface = wildif;
-			add_addr_to_list(&wildif->sin, wildif);
-			add_interface(wildif);
-			log_listen_address(wildif);
-		} else {
+		if (wildif->fd == INVALID_SOCKET) {
 			msyslog(LOG_ERR,
 				"IO: unable to bind to wildcard address %s - another process may be running: %s; EXITING",
 				socktoa(&wildif->sin), strerror(errno));
 			exit(1);
 		}
+		if (NTP_PORT == port) {
+			io_data.wild6_interface_NTP = wildif;
+		} else {
+			io_data.wild6_interface_extra = wildif;
+		}
+		add_addr_to_list(&wildif->sin, wildif);
+		add_interface(wildif);
+		log_listen_address(wildif);
 		DPRINT_INTERFACE(2, (wildif, "created ", "\n"));
 	}
 
@@ -903,26 +891,25 @@ create_wildcards(
 		strlcpy(wildif->name, "v4wildcard", sizeof(wildif->name));
 		memcpy(&wildif->sin, &wildaddr, sizeof(wildif->sin));
 		wildif->family = AF_INET;
-		AF(&wildif->mask) = AF_INET;
-		SET_ONESMASK(&wildif->mask);
 
-		wildif->flags = INT_BROADCAST | INT_UP | INT_WILDCARD;
+		wildif->flags = INT_UP | INT_WILDCARD;
 		wildif->ignore_packets = (ACTION_DROP == action);
 		wildif->fd = open_socket(&wildif->sin, true, wildif);
 
-		if (wildif->fd != INVALID_SOCKET) {
-			wildipv4 = wildif;
-			io_data.any_interface = wildif;
-
-			add_addr_to_list(&wildif->sin, wildif);
-			add_interface(wildif);
-			log_listen_address(wildif);
-		} else {
+		if (wildif->fd == INVALID_SOCKET) {
 			msyslog(LOG_ERR,
 				"IO: unable to bind to wildcard address %s - another process may be running: %s; EXITING",
 				socktoa(&wildif->sin), strerror(errno));
 			exit(1);
 		}
+		if (NTP_PORT == port) {
+			io_data.wild_interface_NTP = wildif;
+		} else {
+			io_data.wild_interface_extra = wildif;
+		}
+		add_addr_to_list(&wildif->sin, wildif);
+		add_interface(wildif);
+		log_listen_address(wildif);
 		DPRINT_INTERFACE(2, (wildif, "created ", "\n"));
 	}
 }
@@ -1001,12 +988,12 @@ static nic_rule_action
 interface_action(
 	char *		if_name,
 	sockaddr_u *	if_addr,
-	uint32_t		if_flags
+	uint32_t	if_flags
 	)
 {
 	nic_rule *	rule;
-	int		isloopback;
-	int		iswildcard;
+	bool		isloopback;
+	bool		iswildcard;
 
 	DPRINT(4, ("interface_action: interface %s ",
 		   (if_name != NULL) ? if_name : "wildcard"));
@@ -1021,8 +1008,8 @@ interface_action(
 	for (rule = nic_rule_list; rule != NULL; rule = rule->next) {
 
 		switch (rule->match_type) {
-                default:
-                        /* huh? */
+		default:
+			/* huh? */
 		case MATCH_ALL:
 			/* loopback and wildcard excluded from "all" */
 			if (isloopback || iswildcard)
@@ -1156,26 +1143,13 @@ convert_isc_if(
 	itf->ifindex = isc_if->ifindex;
 	itf->family = (unsigned short)isc_if->af;
 	AF(&itf->sin) = (sa_family_t)itf->family;
-	AF(&itf->mask) = (sa_family_t)itf->family;
-	AF(&itf->bcast) = (sa_family_t)itf->family;
 	SET_PORT(&itf->sin, port);
-	SET_PORT(&itf->mask, port);
-	SET_PORT(&itf->bcast, port);
 
 	if (IS_IPV4(&itf->sin)) {
 		NSRCADR(&itf->sin) = isc_if->address.type.in.s_addr;
-		NSRCADR(&itf->mask) = isc_if->netmask.type.in.s_addr;
-
-		if (isc_if->flags & INTERFACE_F_BROADCAST) {
-			itf->flags |= INT_BROADCAST;
-			NSRCADR(&itf->bcast) =
-			    isc_if->broadcast.type.in.s_addr;
-		}
 	}
 	else if (IS_IPV6(&itf->sin)) {
 		SET_ADDR6N(&itf->sin, isc_if->address.type.in6);
-		SET_ADDR6N(&itf->mask, isc_if->netmask.type.in6);
-
 		SET_SCOPE(&itf->sin, isc_if->address.zone);
 	}
 
@@ -1186,10 +1160,6 @@ convert_isc_if(
 			? INT_UP : 0)
 		| ((INTERFACE_F_LOOPBACK & isc_if->flags)
 			? INT_LOOPBACK : 0)
-		| ((INTERFACE_F_POINTTOPOINT & isc_if->flags)
-			? INT_PPP : 0)
-		| ((INTERFACE_F_PRIVACY & isc_if->flags)
-			? INT_PRIVACY : 0)
 		;
 
 	/*
@@ -1246,16 +1216,14 @@ refresh_interface(
  * interface_update - externally callable update function
  */
 void
-interface_update(
-	interface_receiver_t	receiver,
-	void *			data)
+interface_update(void)
 {
 	bool new_interface_found;
 
 	if (io_data.disable_dynamic_updates)
 		return;
 
-	new_interface_found = update_interfaces(NTP_PORT, receiver, data);
+	new_interface_found = update_interfaces();
 
 	if (!new_interface_found)
 		return;
@@ -1263,7 +1231,7 @@ interface_update(
 #ifdef DEBUG
 	msyslog(LOG_DEBUG, "IO: new interface(s) found: waking up resolver");
 #endif
-	dns_new_interface();
+	dns_try_again();
 }
 
 
@@ -1288,14 +1256,14 @@ is_wildcard_addr(
  */
 static void
 set_wildcard_reuse(
-	unsigned short	family,
+	sockaddr_u *psau,
 	int	on
 	)
 {
 	endpt *any;
 	SOCKET fd = INVALID_SOCKET;
 
-	any = ANY_INTERFACE_BYFAM(family);
+	any = wildcard_interface(psau);
 	if (any != NULL)
 		fd = any->fd;
 
@@ -1308,7 +1276,7 @@ set_wildcard_reuse(
 
 		DPRINT(4, ("set SO_REUSEADDR to %s on %s\n",
 			   on ? "on" : "off",
-			   socktoa(&any->sin)));
+			   sockporttoa(&any->sin)));
 	}
 }
 #endif /* NEED_REUSEADDR_FOR_IFADDRBIND */
@@ -1321,7 +1289,7 @@ check_flags6(
 	uint32_t flags6
 	)
 {
-#if defined(SIOCGIFAFLAG_IN6)
+#if defined(SIOCGIFAFLAG_IN6)  /* Apple */
 	struct in6_ifreq ifr6;
 	int fd;
 
@@ -1383,13 +1351,15 @@ is_valid(
 	return check_flags6(psau, name, flags6) ? false : true;
 }
 
-/*
+/* This used to be one giant routine before adding extra_port
+ * The old code couldn't handle a second port.
+ *
  * update_interface strategy
  *
  * toggle configuration phase
  *
  * Phase 1:
- * forall currently existing interfaces
+ * forall currently existing interfaces (and all their addresses)
  *   if address is known:
  *	drop socket - rebind again
  *
@@ -1409,14 +1379,32 @@ is_valid(
  */
 
 static bool
-update_interfaces(
-	unsigned short		port,
-	interface_receiver_t	receiver,
-	void *			data
-	)
+update_interfaces(void)
+{
+  bool have_new_interface = false;
+  update_interfaces_phase0();
+  if (extra_port)
+    /* do first so our requests are sent from extra_port
+     * see select_peerinterface() */
+    have_new_interface |= update_interfaces_phase1(extra_port);
+  have_new_interface |= update_interfaces_phase1(NTP_PORT);
+  update_interfaces_phase2();
+  update_interfaces_phase3();
+  return have_new_interface;
+}
+
+void update_interfaces_phase0(void)
+{
+  endpt *ep;
+  for (ep = io_data.ep_list; ep != NULL; ep = ep->elink) {
+    ep->inuse = false;
+  }
+}
+
+static bool
+update_interfaces_phase1(uint16_t port)
 {
 	isc_mem_t *		mctx = (void *)-1;
-	interface_info_t	ifi;
 	isc_interfaceiter_t *	iter;
 	bool			result;
 	isc_interface_t		isc_if;
@@ -1424,9 +1412,8 @@ update_interfaces(
 	unsigned int		family;
 	endpt			enumep;
 	endpt *			ep;
-	endpt *			next_ep;
 
-	DPRINT(3, ("update_interfaces(%d)\n", port));
+	DPRINT(3, ("\n\nupdate_interfaces(%d)\n", port));
 
 	/*
 	 * phase one - scan interfaces
@@ -1441,12 +1428,6 @@ update_interfaces(
 	if (!result)
 		return false;
 
-	/*
-	 * Toggle system interface scan phase to find untouched
-	 * interfaces to be deleted.
-	 */
-	sys_interphase ^= 0x1;
-
 	for (result = isc_interfaceiter_first_bool(iter);
 	     result;
 	     result = isc_interfaceiter_next_bool(iter)) {
@@ -1478,21 +1459,21 @@ update_interfaces(
 		switch (interface_action(enumep.name, &enumep.sin,
 					 enumep.flags)) {
 
-                default:
+		default:
 		case ACTION_IGNORE:
 			DPRINT(4, ("ignoring interface %s (%s) - by nic rules\n",
-				   enumep.name, socktoa(&enumep.sin)));
+				   enumep.name, sockporttoa(&enumep.sin)));
 			continue;
 
 		case ACTION_LISTEN:
 			DPRINT(4, ("listen interface %s (%s) - by nic rules\n",
-				   enumep.name, socktoa(&enumep.sin)));
+				   enumep.name, sockporttoa(&enumep.sin)));
 			enumep.ignore_packets = false;
 			break;
 
 		case ACTION_DROP:
 			DPRINT(4, ("drop on interface %s (%s) - by nic rules\n",
-				   enumep.name, socktoa(&enumep.sin)));
+				   enumep.name, sockporttoa(&enumep.sin)));
 			enumep.ignore_packets = true;
 			break;
 		}
@@ -1500,7 +1481,7 @@ update_interfaces(
 		 /* interfaces must be UP to be usable */
 		if (!(enumep.flags & INT_UP)) {
 			DPRINT(4, ("skipping interface %s (%s) - DOWN\n",
-				   enumep.name, socktoa(&enumep.sin)));
+				   enumep.name, sockporttoa(&enumep.sin)));
 			continue;
 		}
 
@@ -1509,32 +1490,42 @@ update_interfaces(
 		 * address - some dhcp clients produce that in the
 		 * wild
 		 */
-		if (is_wildcard_addr(&enumep.sin))
+		if (is_wildcard_addr(&enumep.sin)) {
+			DPRINT(4, ("skipping interface %s (%s) - WILD\n",
+				   enumep.name, sockporttoa(&enumep.sin)));
 			continue;
+			}
 
-		if (is_anycast(&enumep.sin, isc_if.name))
+		if (is_anycast(&enumep.sin, isc_if.name)) {
+			DPRINT(4, ("skipping interface %s (%s) - ANYCAST\n",
+				   enumep.name, sockporttoa(&enumep.sin)));
 			continue;
+			}
 
 		/*
 		 * skip any address that is an invalid state to be used
 		 */
-		if (!is_valid(&enumep.sin, isc_if.name))
+		if (!is_valid(&enumep.sin, isc_if.name)) {
+			DPRINT(4, ("skipping interface %s (%s) - ~VALID\n",
+				   enumep.name, sockporttoa(&enumep.sin)));
 			continue;
+			}
+
 
 		/*
 		 * map to local *address* in order to map all duplicate
 		 * interfaces to an endpt structure with the appropriate
-		 * socket.  Our name space is (ip-address), NOT
+		 * socket.  Our name space is (ip-address+port), NOT
 		 * (interface name, ip-address).
 		 */
 		ep = getinterface(&enumep.sin, INT_WILDCARD);
 
-		if (ep != NULL && refresh_interface(ep)) {
+		if (ep != NULL && SRCPORT(&ep->sin)==port && refresh_interface(ep)) {
 			/*
 			 * found existing and up to date interface -
 			 * mark present.
 			 */
-			if (ep->phase != sys_interphase) {
+			if (!ep->inuse) {
 				/*
 				 * On a new round we reset the name so
 				 * the interface name shows up again if
@@ -1587,12 +1578,8 @@ update_interfaces(
 				ep->ignore_packets = true;
 			}
 
-			ep->phase = sys_interphase;
+			ep->inuse = true;
 
-			ifi.action = IFS_EXISTS;
-			ifi.ep = ep;
-			if (receiver != NULL)
-				(*receiver)(data, &ifi);
 		} else {
 			/*
 			 * This is new or refreshing failed - add to
@@ -1604,35 +1591,37 @@ update_interfaces(
 			 */
 			ep = create_interface(port, &enumep);
 
-			if (ep != NULL) {
-				ifi.action = IFS_CREATED;
-				ifi.ep = ep;
-				if (receiver != NULL)
-					(*receiver)(data, &ifi);
-
-				new_interface_found = true;
-				DPRINT_INTERFACE(3,
-					(ep, "updating ",
-					 " new - created\n"));
-			} else {
+			if (ep == NULL) {
 				DPRINT_INTERFACE(3,
 					(&enumep, "updating ",
 					 " new - creation FAILED"));
-
 				msyslog(LOG_INFO,
-					"IO: failed to init interface for address %s",
-					socktoa(&enumep.sin));
+					"IO: failed to init interface for %s",
+					sockporttoa(&enumep.sin));
 				continue;
 			}
+
+			new_interface_found = true;
+			ep->inuse = true;
+			DPRINT_INTERFACE(3,
+				(ep, "updating ", " new - created\n"));
 		}
 	}
 
 	isc_interfaceiter_destroy(&iter);
 
-	/*
-	 * phase 2 - delete gone interfaces - reassigning peers to
-	 * other interfaces
-	 */
+	return new_interface_found;
+}
+
+/*
+ * phase 2 - delete gone interfaces - reassigning peers to
+ * other interfaces
+ */
+void update_interfaces_phase2(void)
+{
+	endpt *			ep;
+	endpt *			next_ep;
+
 	for (ep = io_data.ep_list; ep != NULL; ep = next_ep) {
 		next_ep = ep->elink;
 
@@ -1642,18 +1631,13 @@ update_interfaces(
 		 * is gone and will be deleted here unless it did not
 		 * originate from interface enumeration INT_WILDCARD,
 		 */
-		if ((INT_WILDCARD & ep->flags) || ep->phase == sys_interphase)
+		if ((INT_WILDCARD & ep->flags) || ep->inuse)
 			continue;
 
 		DPRINT_INTERFACE(3, (ep, "updating ",
 				     "GONE - deleting\n"));
 		remove_interface(ep);
 
-		ifi.action = IFS_DELETED;
-		ifi.ep = ep;
-		if (receiver != NULL)
-			(*receiver)(data, &ifi);
-
 		/* disconnect peers from deleted endpt. */
 		while (ep->peers != NULL)
 			set_peerdstadr(ep->peers, NULL);
@@ -1667,16 +1651,17 @@ update_interfaces(
 
 		delete_interface(ep);
 	}
+}
 
-	/*
-	 * phase 3 - re-configure as the world has possibly changed
-	 *
-	 * never ever make this conditional again - it is needed to track
-	 * routing updates. see bug #2506
-	 */
+/*
+ * phase 3 - re-configure as the world has possibly changed
+ *
+ * never ever make this conditional again - it is needed to track
+ * routing updates. see bug #2506
+ */
+void update_interfaces_phase3(void)
+{
 	refresh_all_peerinterfaces();
-
-	return new_interface_found;
 }
 
 
@@ -1685,18 +1670,17 @@ update_interfaces(
  *			socket for when we don't know where to send
  */
 static int
-create_sockets(
-	unsigned short port
-	)
+create_sockets(void)
 {
 	maxactivefd = 0;
 	FD_ZERO(&activefds);
 
-	DPRINT(2, ("create_sockets(%d)\n", port));
+	DPRINT(2, ("create_sockets(%d %u)\n", NTP_PORT, extra_port));
 
-	create_wildcards(port);
+	create_wildcards(NTP_PORT);
+	if (extra_port) create_wildcards(extra_port);
 
-	update_interfaces(port, NULL, NULL);
+	update_interfaces();
 
 	/*
 	 * Now that we have opened all the sockets, turn off the reuse
@@ -1721,7 +1705,7 @@ create_interface(
 {
 	sockaddr_u	resmask;
 	endpt *		iface;
-	DPRINT(2, ("create_interface(%s#%d)\n", socktoa(&protot->sin),
+	DPRINT(2, ("create_interface(%s#%d)\n", sockporttoa(&protot->sin),
 		    port));
 
 	/* build an interface */
@@ -1730,7 +1714,7 @@ create_interface(
 	/*
 	 * create socket
 	 */
-	iface->fd = open_socket(&iface->sin, false, iface);
+	iface->fd = open_socket(&iface->sin, true, iface);
 
 	if (iface->fd != INVALID_SOCKET)
 		log_listen_address(iface);
@@ -1750,7 +1734,7 @@ create_interface(
 	 */
 	SET_HOSTMASK(&resmask, AF(&iface->sin));
 	hack_restrict(RESTRICT_FLAGS, &iface->sin, &resmask,
-		      RESM_NTPONLY | RESM_INTERFACE, RES_IGNORE, 0);
+		      RESM_NTPONLY | RESM_INTERFACE, RES_IGNORE);
 
 	/*
 	 * set globals with the first found
@@ -1771,7 +1755,7 @@ create_interface(
 }
 
 
-#ifdef SO_EXCLUSIVEADDRUSE
+#ifdef SO_EXCLUSIVEADDRUSE  /* Windows */
 static void
 set_excladdruse(
 	SOCKET fd
@@ -1795,15 +1779,13 @@ set_excladdruse(
 
 /*
  * set_reuseaddr() - set/clear REUSEADDR on all sockets
- *			NB possible hole - should we be doing this on broadcast
- *			fd's also?
  */
 static void
 set_reuseaddr(
 	int flag
 	)
 {
-#ifndef SO_EXCLUSIVEADDRUSE
+#ifndef SO_EXCLUSIVEADDRUSE  /* Windows */
 	endpt *ep;
 
 	for (ep = io_data.ep_list; ep != NULL; ep = ep->elink) {
@@ -1836,7 +1818,7 @@ set_reuseaddr(
 SOCKET
 open_socket(
 	sockaddr_u *	addr,
-	bool		turn_off_reuse,
+	bool		turn_on_reuse,
 	endpt *		interf
 	)
 {
@@ -1882,24 +1864,24 @@ open_socket(
 
 	/*
 	 * set SO_REUSEADDR since we will be binding the same port
-	 * number on each interface according to turn_off_reuse.
+	 * number on each interface according to turn_on_reuse.
 	 */
 	if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR,
-		       (const void *)((turn_off_reuse)
-				    ? &off
-				    : &on),
+		       (const void *)((turn_on_reuse)
+				    ? &on
+				    : &off),
 		       sizeof(on))) {
 
 		msyslog(LOG_ERR,
 			"IO: setsockopt SO_REUSEADDR %s fails for address %s: %s",
-			(turn_off_reuse)
-			    ? "off"
-			    : "on",
+			(turn_on_reuse)
+			    ? "on"
+			    : "off",
 			socktoa(addr), strerror(errno));
 		close(fd);
 		return INVALID_SOCKET;
 	}
-#ifdef SO_EXCLUSIVEADDRUSE
+#ifdef SO_EXCLUSIVEADDRUSE  /* Windows */
 	/*
 	 * setting SO_EXCLUSIVEADDRUSE on the wildcard we open
 	 * first will cause more specific binds to fail.
@@ -1916,7 +1898,7 @@ open_socket(
 			       sizeof(qos)))
 			msyslog(LOG_ERR,
 				"IO: setsockopt IP_TOS (%02x) fails on "
-                                "address %s: %s",
+				"address %s: %s",
 				(unsigned)qos, socktoa(addr), strerror(errno));
 	}
 
@@ -1928,8 +1910,8 @@ open_socket(
 		if (setsockopt(fd, IPPROTO_IPV6, IPV6_TCLASS, (char*)&qos,
 			       sizeof(qos)))
 			msyslog(LOG_ERR, "IO: setsockopt IPV6_TCLASS (%02x) "
-                                         "fails on address %s: %s",
-				         (unsigned)qos, socktoa(addr), strerror(errno));
+					"fails on address %s: %s",
+					(unsigned)qos, socktoa(addr), strerror(errno));
 #endif /* IPV6_TCLASS */
 		if (isc_net_probe_ipv6only_bool()
 		    && setsockopt(fd, IPPROTO_IPV6, IPV6_V6ONLY,
@@ -1946,37 +1928,28 @@ open_socket(
 	 * to the port and SO_REUSEADDR is not set
 	 */
 	if (!is_wildcard_addr(addr))
-		set_wildcard_reuse(AF(addr), 1);
+		set_wildcard_reuse(addr, 1);
 #endif
 
 	/*
-	 * bind the local address.
+	 bind the local address.
 	 */
 	errval = bind(fd, &addr->sa, SOCKLEN(addr));
 
 #ifdef NEED_REUSEADDR_FOR_IFADDRBIND
+
 	if (!is_wildcard_addr(addr))
-		set_wildcard_reuse(AF(addr), 0);
+		set_wildcard_reuse(addr, 0);
 #endif
 
 	if (errval < 0) {
-		/*
-		 * Don't log this under all conditions
-		 */
-		if (turn_off_reuse == 0
-#ifdef DEBUG
-		    || debug > 1 /* SPECIAL DEBUG */
-#endif
-		    ) {
-			msyslog(LOG_ERR,
-				"IO: bind(%d) AF_INET%s %s#%d flags 0x%x failed: %s",
-				fd, IS_IPV6(addr) ? "6" : "",
-				socktoa(addr), SRCPORT(addr),
-				interf->flags, strerror(errno));
-		}
-
+		// FIXME: set_wildcard_reuse trashes errno??
+		msyslog(LOG_ERR,
+			"IO: bind(%d) AF_INET%s %s#%d flags 0x%x failed: %s",
+			fd, IS_IPV6(addr) ? "6" : "",
+			socktoa(addr), SRCPORT(addr),
+			interf->flags, strerror(errno));
 		close(fd);
-
 		return INVALID_SOCKET;
 	}
 
@@ -2086,7 +2059,7 @@ read_refclock_packet(
 
 	i = (rp->datalen == 0
 	     || rp->datalen > sizeof(rb->recv_buffer))
-	        ? sizeof(rb->recv_buffer)
+		? sizeof(rb->recv_buffer)
 		: rp->datalen;
 	do {
 		buflen = read(fd, (char *)&rb->recv_buffer, i);
@@ -2171,17 +2144,17 @@ read_network_packet(
 
 	fromlen = sizeof(rb->recv_srcadr);
 
-	iovec.iov_base        = &rb->recv_buffer;
-	iovec.iov_len         = sizeof(rb->recv_buffer);
+	iovec.iov_base		= &rb->recv_buffer;
+	iovec.iov_len		= sizeof(rb->recv_buffer);
 	memset(&msghdr, '\0', sizeof(msghdr));
-	msghdr.msg_name       = &rb->recv_srcadr;
-	msghdr.msg_namelen    = fromlen;
-	msghdr.msg_iov        = &iovec;
-	msghdr.msg_iovlen     = 1;
-	msghdr.msg_flags      = 0;
-	msghdr.msg_control    = (void *)&control;
-	msghdr.msg_controllen = sizeof(control);
-	buflen                = recvmsg(fd, &msghdr, 0);
+	msghdr.msg_name		= &rb->recv_srcadr;
+	msghdr.msg_namelen	= fromlen;
+	msghdr.msg_iov		= &iovec;
+	msghdr.msg_iovlen	= 1;
+	msghdr.msg_flags	= 0;
+	msghdr.msg_control	= (void *)&control;
+	msghdr.msg_controllen	= sizeof(control);
+	buflen			= recvmsg(fd, &msghdr, 0);
 
 	rb->recv_length = (size_t)buflen;
 
@@ -2320,7 +2293,6 @@ input_handler(
 	 * See below for the other cases (nothing left to do or error)
 	 */
 
-	++pkt_count.handler_pkts;
 
 #ifdef REFCLOCK
 	/*
@@ -2370,6 +2342,7 @@ input_handler(
 		if (FD_ISSET(fd, fds))
 			do {
 				++select_count;
+				++pkt_count.handler_pkts;
 				buflen = read_network_packet(fd, ep);
 			} while (buflen > 0);
 	}
@@ -2411,6 +2384,10 @@ input_handler(
 
 /*
  * find an interface suitable for the src address
+ * Called by newpeer() and peer_refresh_interface()
+ * if extra_port is active,
+ *   this selects the first one on the chain
+ *   see update_interfaces().
  */
 endpt *
 select_peerinterface(
@@ -2426,8 +2403,7 @@ select_peerinterface(
 	UNUSED_ARG(peer);
 #endif
 
-
-	wild = ANY_INTERFACE_CHOOSE(srcadr);
+	wild = wildcard_interface(srcadr);
 
 	/*
 	 * Initialize the peer structure and dance the interface jig.
@@ -2462,6 +2438,38 @@ select_peerinterface(
 	return ep;
 }
 
+/* FIXME SO_REUSEADDR, Hal 2024-Apr-19
+ * I'm not sure the NULLs below are correct.  It seems to work.
+ * I'm not sure it is worth the effort to turn off SO_REUSEADDR
+ */
+
+/*
+ * find wildcard interface
+ */
+endpt * wildcard_interface(const sockaddr_u *addr) {
+  unsigned short family = AF(addr);
+  uint16_t port = SRCPORT(addr);
+  if (AF_INET == family) {
+    if (NTP_PORT == port) {
+      return io_data.wild_interface_NTP;
+    } else if (extra_port == port) {
+      return io_data.wild_interface_extra;
+    } else {
+      return NULL;
+    }
+  } else if (AF_INET6 == family) {
+    if (NTP_PORT == port) {
+      return io_data.wild6_interface_NTP;
+    } else if (extra_port == port) {
+      return io_data.wild6_interface_extra;
+    } else {
+      return NULL;
+    }
+  } else {
+    return NULL;
+  }
+}
+
 
 /*
  * findinterface - find local interface corresponding to address
@@ -2473,13 +2481,13 @@ findinterface(
 {
 	endpt *iface;
 
-	iface = findlocalinterface(addr, INT_WILDCARD, 0);
+	iface = findlocalinterface(addr);
 
 	if (NULL == iface) {
 		DPRINT(4, ("Found no interface for address %s - returning wildcard\n",
 			   socktoa(addr)));
 
-		iface = ANY_INTERFACE_CHOOSE(addr);
+		iface = wildcard_interface(addr);
 	} else
 		DPRINT(4, ("Found interface #%u %s for address %s\n",
 			   iface->ifnum, iface->name, socktoa(addr)));
@@ -2488,9 +2496,7 @@ findinterface(
 }
 
 /*
- * findlocalinterface - find local interface corresponding to addr,
- * which does not have any of flags set.  If bast is nonzero, addr is
- * a broadcast address.
+ * findlocalinterface - find local interface corresponding to addr
  *
  * This code attempts to find the local sending address for an outgoing
  * address by connecting a new socket to destinationaddress:NTP_PORT
@@ -2504,9 +2510,7 @@ findinterface(
  */
 static endpt *
 findlocalinterface(
-	sockaddr_u *	addr,
-	int		flags,
-	int		bcast
+	sockaddr_u *	addr
 	)
 {
 	socklen_t	sockaddrlen;
@@ -2514,7 +2518,7 @@ findlocalinterface(
 	sockaddr_u	saddr;
 	SOCKET		s;
 	int		rtn;
-	int		on;
+	int		flags = INT_WILDCARD;
 
 	DPRINT(4, ("Finding interface for addr %s in list of addresses\n",
 		   socktoa(addr)));
@@ -2523,21 +2527,6 @@ findlocalinterface(
 	if (INVALID_SOCKET == s)
 		return NULL;
 
-	/*
-	 * If we are looking for broadcast interface we need to set this
-	 * socket to allow broadcast
-	 */
-	if (bcast) {
-		on = 1;
-		if (SOCKET_ERROR == setsockopt(s, SOL_SOCKET,
-						SO_BROADCAST,
-						(char *)&on,
-						sizeof(on))) {
-			close(s);
-			return NULL;
-		}
-	}
-
 	rtn = connect(s, &addr->sa, SOCKLEN(addr));
 	if (SOCKET_ERROR == rtn) {
 		close(s);
@@ -2726,7 +2715,7 @@ cmp_addr_distance(
 endpt *
 getinterface(
 	sockaddr_u *	addr,
-	uint32_t		flags
+	uint32_t	flags
 	)
 {
 	endpt *iface;
@@ -2753,7 +2742,10 @@ io_clr_stats(void)
 
 	pkt_count.handler_calls = 0;
 	pkt_count.handler_pkts = 0;
-	pkt_count.io_timereset = current_time;
+#ifdef REFCLOCK
+	pkt_count.handler_refrds = 0;
+#endif
+	io_timereset = current_time;
 }
 
 /*
@@ -2812,12 +2804,14 @@ uint64_t handler_pkts_count(void) {
   return pkt_count.handler_pkts;
 }
 
+#ifdef REFCLOCK
 /*
- * counter_reset_time - return the time of the last counter reset
+ * handler_pkts_refrds - return the number of refclock reads
  */
-uptime_t counter_reset_time(void) {
-  return pkt_count.io_timereset;
+uint64_t handler_refrds_count(void) {
+  return pkt_count.handler_refrds;
 }
+#endif
 
 
 #ifdef REFCLOCK
@@ -2999,7 +2993,7 @@ find_addr_in_list(
 	for (entry = remoteaddr_list;
 	     entry != NULL;
 	     entry = entry->link) {
-		if (SOCK_EQ(&entry->addr, addr)) {
+		if (ADDR_PORT_EQ(&entry->addr, addr)) {
 			DPRINT(4, ("FOUND\n"));
 			return entry->ep;
 		}
@@ -3030,7 +3024,7 @@ process_routing_msgs(struct asyncio_reader *reader)
 {
 	char buffer[5120];
 	ssize_t cnt;
-        int msg_type;
+	int msg_type;
 #ifdef HAVE_LINUX_RTNETLINK_H
 	struct nlmsghdr *nh;
 #else
@@ -3156,7 +3150,7 @@ process_routing_msgs(struct asyncio_reader *reader)
  * set up routing notifications
  */
 static void
-init_async_notifications()
+init_async_notifications(void)
 {
 	struct asyncio_reader *reader;
 #ifdef HAVE_LINUX_RTNETLINK_H
