@@ -175,12 +175,6 @@ static double sys_mindly;	/* huff-n'-puff filter min delay */
 /* Emacs cc-mode goes nuts if we split the next line... */
 #define MOD_BITS (MOD_OFFSET | MOD_MAXERROR | MOD_ESTERROR | \
     MOD_STATUS | MOD_TIMECONST)
-#ifdef SIGSYS
-static void pll_trap (int);	/* configuration trap */
-static struct sigaction sigsys;	/* current sigaction status */
-static struct sigaction newsigsys; /* new sigaction status */
-static sigjmp_buf env;		/* environment var. for pll_trap() */
-#endif /* SIGSYS */
 
 static void
 sync_status(const char *what, int ostatus, int nstatus) {
@@ -486,7 +480,7 @@ local_clock(
 	if (clock_ctl.mode_ntpdate) {
 		if (  ( fp_offset > loop_data.clock_max_fwd  && loop_data.clock_max_fwd  > 0)
 		   || (-fp_offset > loop_data.clock_max_back && loop_data.clock_max_back > 0)) {
-			step_systime(fp_offset, ntp_set_tod);
+			step_systime(fp_offset);
 			msyslog(LOG_NOTICE, "CLOCK: time set %+.6f s",
 			    fp_offset);
 			printf("ntpd: time set %+.6fs\n", fp_offset);
@@ -617,7 +611,7 @@ local_clock(
 			snprintf(tbuf, sizeof(tbuf), "%+.6f s",
 			    fp_offset);
 			report_event(EVNT_CLOCKRESET, NULL, tbuf);
-			step_systime(fp_offset, ntp_set_tod);
+			step_systime(fp_offset);
 			reinit_timer();
 			clkstate.tc_counter = 0;
 			clkstate.clock_jitter = LOGTOD(sys_vars.sys_precision);
@@ -751,14 +745,7 @@ local_clock(
 				dtemp = .5;
 			}
 			ntv.offset = (long)(clock_offset * NS_PER_S + dtemp);
-#ifdef STA_NANO
 			ntv.constant = clkstate.sys_poll;
-#else /* STA_NANO */
-			ntv.constant = clkstate.sys_poll - 4;
-#endif /* STA_NANO */
-			if (ntv.constant < 0)
-				ntv.constant = 0;
-
 			ntv.esterror = (long)(clkstate.clock_jitter * US_PER_S);
 			ntv.maxerror = (long)((sys_vars.sys_rootdelay / 2 +
 			    sys_vars.sys_rootdisp) * US_PER_S);
@@ -1065,35 +1052,10 @@ start_kern_loop(void)
 	ntv.status = STA_PLL;
 	ntv.maxerror = sys_maxdisp;
 	ntv.esterror = sys_maxdisp;
-	ntv.constant = clkstate.sys_poll; /* why is it that here constant is unconditionally set to sys_poll, whereas elsewhere is is modified depending on nanosecond vs. microsecond kernel? */
-#ifdef SIGSYS
-	/*
-	 * Use sigsetjmp() to save state and then call ntp_adjtime(); if
-	 * it fails, then pll_trap() will set pll_control false before
-	 * returning control using siglogjmp().
-	 */
-	newsigsys.sa_handler = pll_trap;
-	newsigsys.sa_flags = 0;
-	if (sigaction(SIGSYS, &newsigsys, &sigsys)) {
-		msyslog(LOG_ERR, "ERR: sigaction() trap SIGSYS: %s", strerror(errno));
-		clock_ctl.pll_control = false;
-	} else {
-		if (sigsetjmp(env, 1) == 0) {
-			if ((ntp_adj_ret = ntp_adjtime_ns(&ntv)) != 0) {
-			    ntp_adjtime_error_handler(__func__, &ntv, ntp_adj_ret, errno, false, false, __LINE__ - 1);
-			}
-		}
-		if (sigaction(SIGSYS, &sigsys, NULL)) {
-			msyslog(LOG_ERR,
-			    "ERR: sigaction() restore SIGSYS: %s", strerror(errno));
-			clock_ctl.pll_control = false;
-		}
-	}
-#else /* SIGSYS */
+	ntv.constant = clkstate.sys_poll;
 	if ((ntp_adj_ret = ntp_adjtime_ns(&ntv)) != 0) {
 	    ntp_adjtime_error_handler(__func__, &ntv, ntp_adj_ret, errno, false, false, __LINE__ - 1);
 	}
-#endif /* SIGSYS */
 
 	/*
 	 * Save the result status and light up an external clock
@@ -1298,9 +1260,11 @@ loop_config(
 		}
 		break;
 
+#ifdef ENABLE_FUZZ
 	case LOOP_TICK:		/* tick increment (tick) */
 		set_sys_tick_precision(freq);
 		break;
+#endif
 
 	case LOOP_LEAP:		/* not used, fall through */
 	default:
@@ -1309,23 +1273,3 @@ loop_config(
 	}
 }
 
-
-#if defined(SIGSYS)
-/*
- * _trap - trap processor for undefined syscalls
- *
- * This nugget is called by the kernel when the SYS_ntp_adjtime()
- * syscall bombs because the silly thing has not been implemented in
- * the kernel. In this case the phase-lock loop is emulated by
- * the stock adjtime() syscall and a lot of indelicate abuse.
- */
-static void
-pll_trap(
-	int arg
-	)
-{
-	UNUSED_ARG(arg);
-	clock_ctl.pll_control = false;
-	siglongjmp(env, 1);
-}
-#endif /* SIGSYS */
