@@ -98,7 +98,7 @@ init_peer(void)
 	 * Initialize our first association ID
 	 */
 	do
-		current_association_ID = ntp_random() & ASSOCID_MAX;
+		current_association_ID = random() & ASSOCID_MAX;
 	while (!current_association_ID);
 	initial_association_ID = current_association_ID;
 }
@@ -278,7 +278,7 @@ clear_all(void)
 	 * previously saved time values are untrusted.
 	 */
 	for (p = peer_list; p != NULL; p = p->p_link)
-		if (!(MDF_TXONLY_MASK & p->cast_flags))
+		if (!(MDF_POOL & p->cast_flags))
 		    peer_clear(p, "STEP", false);
 
 	DPRINT(1, ("clear_all: at %u\n", current_time));
@@ -334,7 +334,7 @@ score(
 	int	temp;
 
 	/*
-	 * This routine calculates the premption score from the peer
+	 * This routine calculates the preemption score from the peer
 	 * error bits and status. Increasing values are more cherished.
 	 */
 	temp = 0;
@@ -420,7 +420,7 @@ unpeer(
 	)
 {
 	mprintf_event(PEVNT_DEMOBIL, peer, "assoc %u", peer->associd);
-	restrict_source(&peer->srcadr, true, 0);
+	unrestrict_source(peer);
 	set_peerdstadr(peer, NULL);
 	peer_demobilizations++;
 	peer_associations--;
@@ -492,10 +492,6 @@ peer_refresh_interface(
 			   niface->fd, niface->name,
 			   niface->flags, niface->ifindex,
 			   socktoa(&niface->sin)));
-		if (niface->flags & INT_BROADCAST)
-			DPRINT(4, (", bcast=%s",
-				   socktoa(&niface->bcast)));
-		DPRINT(4, (", mask=%s\n", socktoa(&niface->mask)));
 	} else {
 		DPRINT(4, ("<NONE>\n"));
 	}
@@ -554,10 +550,11 @@ newpeer(
 	unsigned int	hash;
 	const char *	name;	/* for error messages */
 
+
 	if (NULL != hostname) {
 		name = hostname;
 	} else {
-		name = socktoa(srcadr); 
+		name = socktoa(srcadr);
 	}
 
 	/*
@@ -571,12 +568,10 @@ newpeer(
 	if (dstadr != NULL) {
 		peer = findexistingpeer(srcadr, hostname, NULL, hmode);
 		while (peer != NULL) {
-			if (peer->dstadr == dstadr ||
-			    ((MDF_BCLNT & cast_flags) &&
-			     (MDF_BCLNT & peer->cast_flags)))
+			if (peer->dstadr == dstadr)
 				break;
 
-			if (dstadr == ANY_INTERFACE_CHOOSE(srcadr) &&
+			if (dstadr == wildcard_interface(srcadr) &&
 			    peer->dstadr == findinterface(srcadr))
 				break;
 
@@ -589,9 +584,7 @@ newpeer(
 
 	/*
 	 * If a peer is found, this would be a duplicate and we don't
-	 * allow that. This avoids duplicate ephemeral (broadcast/
-	 * multicast) and preemptible (manycast and pool) client
-	 * associations.
+	 * allow that.
 	 */
 	if (peer != NULL) {
 		DPRINT(2, ("newpeer(%s) found existing association\n", name));
@@ -689,14 +682,16 @@ newpeer(
 	/*
 	 * Put the new peer in the hash tables.
 	 */
-	if ((MDF_UCAST & cast_flags) && !(FLAG_LOOKUP & ctl->flags))
+	if ((MDF_UCAST & cast_flags) && !(FLAG_LOOKUP & ctl->flags)) {
+		/* simple server with numeric address */
 		peer_add_hash(peer);
+		restrict_source(peer);
+	}
 	hash = peer->associd & NTP_HASH_MASK;
 	LINK_SLIST(assoc_hash[hash], peer, aid_link);
 	assoc_hash_count[hash]++;
 	LINK_SLIST(peer_list, peer, p_link);
 
-	restrict_source(&peer->srcadr, false, 0);
 	mprintf_event(PEVNT_MOBIL, peer, "assoc %d", peer->associd);
 	DPRINT(1, ("newpeer: %s->%s mode %u vers %u poll %u %u flags 0x%x 0x%x mode %u key %08x\n",
 		   latoa(peer->dstadr), socktoa(&peer->srcadr), peer->hmode,
