@@ -4,6 +4,7 @@
  */
 #include "config.h"
 
+#include "ntp.h"
 #include "ntpd.h"
 #include "ntp_endian.h"
 #include "ntp_stdlib.h"
@@ -19,6 +20,8 @@
 #endif
 #include <unistd.h>
 
+#define MSSNTP_QUERY_MAC_LEN 16
+
 /*
  * Byte order conversion
  */
@@ -61,15 +64,10 @@ static inline l_fp_w htonl_fp(l_fp lfp) {
  */
 #define	NTP_MINCLOCK	3	/* min survivors */
 #define	NTP_MAXCLOCK	10	/* max candidates */
-#define MINDISPERSE	.001	/* min distance */
+#define MINDISTANCE	.001	/* min distance */
 #define CLOCK_SGATE	3.	/* popcorn spike gate */
 #define	NTP_ORPHWAIT	300	/* orphan wait (s) */
 
-/*
- * pool soliciting restriction duration (s)
- */
-#define	POOL_SOLICIT_WINDOW	8
-
 #define DIFF(x, y) (SQUARE((x) - (y)))
 
 /*
@@ -119,7 +117,7 @@ bool leap_sec_in_progress;
  * Nonspecified system state variables
  */
 l_fp	sys_authdelay;		/* authentication delay */
-double	sys_mindisp = MINDISPERSE; /* minimum distance (s) */
+double	sys_mindist = MINDISTANCE; /* minimum distance (s) */
 static double	sys_maxdist = MAXDISTANCE; /* selection threshold */
 double	sys_maxdisp = MAXDISPERSE; /* maximum dispersion */
 static unsigned long	sys_epoch;	/* last clock update time */
@@ -141,110 +139,98 @@ int	sys_maxclock = NTP_MAXCLOCK; /* maximum candidates */
 int	sys_orphan = STRATUM_UNSPEC + 1; /* orphan stratum */
 static int sys_orphwait = NTP_ORPHWAIT; /* orphan wait */
 
-/*
- * Statistics counters - first the good, then the bad
- * These get reset every hour if sysstats is enabled.
- */
+// proto stats structure and variables
 struct statistics_counters {
-	uptime_t	sys_stattime;		/* time since sysstats reset */
 	uint64_t	sys_received;		/* packets received */
 	uint64_t	sys_processed;		/* packets for this host */
 	uint64_t	sys_restricted;		/* restricted packets */
 	uint64_t	sys_newversion;		/* current version  */
 	uint64_t	sys_oldversion;		/* old version */
+	uint64_t	sys_version1;		/* old cruft: NTPv1 */
+	uint64_t	sys_version1client;	/*  NTPv1: mode 3 many */
+	uint64_t	sys_version1zero;	/*  NTPv1: mode 0 as per RFC */
+	uint64_t	sys_version1symm;	/*  NTPv1: mode 1 old Windows */
 	uint64_t	sys_badlength;		/* bad length or format */
 	uint64_t	sys_badauth;		/* bad authentication */
 	uint64_t	sys_declined;		/* declined */
 	uint64_t	sys_limitrejected;	/* rate exceeded */
 	uint64_t	sys_kodsent;		/* KoD sent */
-	uptime_t	use_stattime;		/* time since usestats reset */
 };
-volatile struct statistics_counters stat_count;
+volatile struct statistics_counters stat_proto_hourago, stat_proto_total;
+uptime_t	sys_stattime;		/* time since sysstats "reset" */
+
+uptime_t	use_stattime;		/* time since usestats reset */
 
 uptime_t stat_stattime(void)
 {
-  return stat_count.sys_stattime;
+  return current_time - sys_stattime;
 }
 
-uint64_t stat_received(void)
+uptime_t stat_total_stattime(void)
 {
-  return stat_count.sys_received;
+  return current_time;
 }
 
-uint64_t stat_processed(void)
-{
-  return stat_count.sys_processed;
+#define stat_sys_dumps(member)\
+uint64_t stat_##member(void) {\
+  return stat_proto_total.sys_##member - stat_proto_hourago.sys_##member;\
+}\
+uint64_t stat_total_##member(void) {\
+  return stat_proto_total.sys_##member;\
 }
 
-uint64_t stat_restricted(void)
-{
-  return stat_count.sys_restricted;
-}
+stat_sys_dumps(received)
+stat_sys_dumps(processed)
+stat_sys_dumps(restricted)
+stat_sys_dumps(newversion)
+stat_sys_dumps(oldversion)
+stat_sys_dumps(version1)
+stat_sys_dumps(version1client)
+stat_sys_dumps(version1zero)
+stat_sys_dumps(version1symm)
+stat_sys_dumps(badlength)
+stat_sys_dumps(badauth)
+stat_sys_dumps(declined)
+stat_sys_dumps(limitrejected)
+stat_sys_dumps(kodsent)
+
+#undef stat_sys_dumps
 
 void increment_restricted(void)
 {
-  stat_count.sys_restricted++;
-}
-
-uint64_t stat_newversion(void)
-{
-  return stat_count.sys_newversion;
-}
-
-uint64_t stat_oldversion(void)
-{
-  return stat_count.sys_oldversion;
-}
-
-uint64_t stat_badlength(void)
-{
-  return stat_count.sys_badlength;
-}
-
-uint64_t stat_badauth(void)
-{
-  return stat_count.sys_badauth;
-}
-
-uint64_t stat_declined(void)
-{
-  return stat_count.sys_declined;
-}
-
-uint64_t stat_limitrejected(void)
-{
-  return stat_count.sys_limitrejected;
-}
-
-uint64_t stat_kodsent(void)
-{
-  return stat_count.sys_kodsent;
+  stat_proto_total.sys_restricted++;
 }
 
 uptime_t stat_use_stattime(void)
 {
-  return stat_count.use_stattime;
+  return current_time - use_stattime;
 }
 
 void set_use_stattime(uptime_t stattime) {
-  stat_count.use_stattime = stattime;
+  use_stattime = stattime;
 }
 
 
+#ifdef ENABLE_FUZZ
 double	measured_tick;		/* non-overridable sys_tick (s) */
+#endif
 
 static	void	clock_combine	(peer_select *, int, int);
 static	void	clock_select	(void);
 static	void	clock_update	(struct peer *);
-static	void	fast_xmit	(struct recvbuf *, int, auth_info*, int);
+static	void	fast_xmit	(struct recvbuf *, auth_info*, int);
 static	int	local_refid	(struct peer *);
+#ifdef ENABLE_FUZZ
 static	void	measure_precision(const bool);
 static	double	measure_tick_fuzz(void);
+#endif
 static	void	peer_xmit	(struct peer *);
 static	int	peer_unfit	(struct peer *);
 static	double	root_distance	(struct peer *);
+#ifndef DISABLE_NTS
 static	void	restart_nts_ke	(struct peer *);
-static	void	maybe_log_junk	(struct recvbuf *rbuf);
+#endif
+static	void	maybe_log_junk	(const char *tag, struct recvbuf *rbuf);
 
 void
 set_sys_leap(unsigned char new_sys_leap) {
@@ -275,18 +261,19 @@ set_sys_leap(unsigned char new_sys_leap) {
 }
 
 /* Returns false for packets we want to reject out of hand: those with an
-   out-of-range version number or an unsupported mode.
+   out-of-range version number, an unsupported mode, or too short.
 */
 static bool
-is_vn_mode_acceptable(
+is_packet_not_low_rot(
 	struct recvbuf const* rbufp
 	)
 {
-	return rbufp->recv_length >= 1 &&
-	    PKT_VERSION(rbufp->recv_buffer[0]) >= 1 &&
-	    PKT_VERSION(rbufp->recv_buffer[0]) <= 4 &&
-	    PKT_MODE(rbufp->recv_buffer[0]) != MODE_PRIVATE &&
-	    PKT_MODE(rbufp->recv_buffer[0]) != MODE_UNSPEC;
+	return rbufp->recv_length >= 12 &&
+	    PKT_VERSION(rbufp->recv_buffer[0]) >= NTP_OLDVERSION &&
+	    PKT_VERSION(rbufp->recv_buffer[0]) <= NTP_VERSION &&
+	    ( PKT_MODE(rbufp->recv_buffer[0]) == MODE_CLIENT ||
+	      PKT_MODE(rbufp->recv_buffer[0]) == MODE_SERVER ||
+              PKT_MODE(rbufp->recv_buffer[0]) == MODE_CONTROL);
 }
 
 static bool
@@ -295,7 +282,7 @@ is_control_packet(
 	)
 {
 	return rbufp->recv_length >= 1 &&
-	    PKT_VERSION(rbufp->recv_buffer[0]) <= 4 &&
+	    PKT_VERSION(rbufp->recv_buffer[0]) <= NTP_VERSION &&
 	    PKT_MODE(rbufp->recv_buffer[0]) == MODE_CONTROL;
 }
 
@@ -338,10 +325,10 @@ parse_packet(
 	rbufp->extens_present = false;
 	rbufp->ntspacket.valid = false;
 
-	if(PKT_VERSION(pkt->li_vn_mode) > 4) {
+	if(PKT_VERSION(pkt->li_vn_mode) > NTP_VERSION) {
 		/* Unsupported version */
 		return false;
-	} else if(PKT_VERSION(pkt->li_vn_mode) == 4) {
+	} else if(PKT_VERSION(pkt->li_vn_mode) == MODE_SERVER) {
 		/* Only version 4 packets support extensions. */
 		/* But they also support shared key authentication. */
 		if (recv_length > (LEN_PKT_NOMAC+MAX_MAC_LEN)) {
@@ -464,7 +451,7 @@ static bool is_kod(
 
 /* Check the restrictions which can be checked just based on the source
    IP address and the first byte of the packet, namely RES_IGNORE,
-   RES_FLAKE, RES_FLAKE, RES_NOQUERY, RES_DONTSERVE, and RES_VERSION. */
+   RES_FLAKE, RES_NOQUERY, RES_DONTSERVE, and RES_VERSION. */
 
 static bool check_early_restrictions(
 	struct recvbuf const* rbufp,
@@ -473,7 +460,7 @@ static bool check_early_restrictions(
 {
 	return (restrict_mask & RES_IGNORE) ||
 	    ((restrict_mask & RES_FLAKE) &&
-	     (double)ntp_random() / 0x7fffffff < .1) ||
+	     (double)random() / RAND_MAX < .1) ||
 	    (restrict_mask & (is_control_packet(rbufp) ? RES_NOQUERY : RES_DONTSERVE)) ||
 	    rbufp->recv_length < 1 ||
 	    ((restrict_mask & RES_VERSION) &&
@@ -481,54 +468,73 @@ static bool check_early_restrictions(
 	      PKT_VERSION(rbufp->recv_buffer[0]) != NTP_VERSION));
 }
 
-static void
-handle_fastxmit(
-	struct recvbuf *rbufp,
-	unsigned short restrict_mask,
-	auth_info* auth
-	)
-{
-        int xmode =
-            PKT_MODE(rbufp->pkt.li_vn_mode) == MODE_ACTIVE ? MODE_PASSIVE : MODE_SERVER;
-	fast_xmit(rbufp, xmode, auth, restrict_mask);
+/* rawstats_filter
+ * Don't print all rejectioned packets or we could get DoSed.
+ * Print the packet we use.
+ * Print the first rejection.
+ *   In particular, we get to see why a response is rejected.
+ *   Took-too-long, BOGON14 is common.
+ *   This lets us see how long it did take.
+ */
+static void rawstats_filter(
+  struct peer *peer,
+  struct recvbuf *rbufp,
+  unsigned int flag,
+  unsigned int outcount) {
+	peer->flash |= flag;
+	peer->bogons += 1;
+	if (peer->bogons > 1) {
+		/* only print one bogon */
+		return;
+	}
+	record_raw_stats(peer, rbufp, flag, outcount);
+#if 0
+	record_raw_stats(peer,
+			 PKT_LEAP(rbufp->pkt.li_vn_mode),
+			 PKT_VERSION(rbufp->pkt.li_vn_mode),
+			 PKT_MODE(rbufp->pkt.li_vn_mode),
+			 PKT_TO_STRATUM(rbufp->pkt.stratum),
+			 rbufp->pkt.ppoll, rbufp->pkt.precision,
+			 rbufp->pkt.rootdelay, rbufp->pkt.rootdisp,
+			 /* FIXME: this cast is disgusting */
+			 *(const uint32_t*)rbufp->pkt.refid,
+			 outcount);
+#endif
 }
 
+/* Handle MODE_SERVER, replies to our requests.
+ * Authentication done upstream.
+ */
 static void
 handle_procpkt(
 	struct recvbuf *rbufp,
 	struct peer *peer
 	)
 {
-	int outcount = peer->outcount;
+	unsigned int outcount = peer->outcount;
 
 	peer->flash &= ~PKT_BOGON_MASK;
 
 	/* Duplicate detection */
 	if(rbufp->pkt.xmt == peer->xmt) {
-		peer->flash |= BOGON1;
+		rawstats_filter(peer, rbufp, BOGON1, outcount);
+		peer->oldpkt++;
+		return;
+	}
+	if(outcount == 0) {
+		rawstats_filter(peer, rbufp, BOGON1, outcount);
 		peer->oldpkt++;
 		return;
 	}
 
 	/* Origin timestamp validation */
-	if(PKT_MODE(rbufp->pkt.li_vn_mode) == MODE_SERVER) {
-		if(outcount == 0) {
-			peer->flash |= BOGON1;
-			peer->oldpkt++;
-			return;
-		}
-		if(rbufp->pkt.org == 0) {
-			peer->flash |= BOGON3;
-			peer->bogusorg++;
-			return;
-		} else if(rbufp->pkt.org != peer->org_rand) {
-			peer->flash |= BOGON2;
-			peer->bogusorg++;
-			return;
-		}
-	} else {
-		/* This case should be unreachable. */
-		stat_count.sys_declined++;
+	if(rbufp->pkt.org == 0) {
+		rawstats_filter(peer, rbufp, BOGON3, outcount);
+		peer->bogusorg++;
+		return;
+	} else if(rbufp->pkt.org != peer->org_rand) {
+		rawstats_filter(peer, rbufp, BOGON2, outcount);
+		peer->bogusorg++;
 		return;
 	}
 
@@ -557,13 +563,13 @@ handle_procpkt(
 	if (PKT_LEAP(rbufp->pkt.li_vn_mode) == LEAP_NOTINSYNC ||
 	    PKT_TO_STRATUM(rbufp->pkt.stratum) < sys_floor ||
 	    PKT_TO_STRATUM(rbufp->pkt.stratum) >= sys_ceiling) {
-		peer->flash |= BOGON6;
+		rawstats_filter(peer, rbufp, BOGON6, outcount);
 		return;
 	}
 
 	if(scalbn((double)rbufp->pkt.rootdelay/2.0 + (double)rbufp->pkt.rootdisp, -16) >=
 	   sys_maxdisp) {
-		peer->flash |= BOGON7;
+		rawstats_filter(peer, rbufp, BOGON7, outcount);
 		return;
 	}
 
@@ -599,9 +605,7 @@ handle_procpkt(
 	   makes the desired security invariant easier to verify.
 	*/
 	if(delta > sys_maxdist) {
-	  peer->flash |= BOGON1; /*XXX we should probably allocate a
-				   new bogon bit here rather than
-				   recycling BOGON1. */
+	  rawstats_filter(peer, rbufp, BOGON14, outcount);
 	  peer->oldpkt++;
 	  return;
 	}
@@ -618,23 +622,8 @@ handle_procpkt(
 	peer->xmt = rbufp->pkt.xmt;
 	peer->dst = rbufp->recv_time;
 
-	record_raw_stats(peer,
-			 /* What we want to be reporting is values in
-			    the packet, not the values in the peer
-			    structure, but when we reach here they're
-			    the same thing. Passing the values in the
-			    peer structure is a convenience, because
-			    they're already in the l_fp format that
-			    record_raw_stats() expects. */
-			 PKT_LEAP(rbufp->pkt.li_vn_mode),
-			 PKT_VERSION(rbufp->pkt.li_vn_mode),
-			 PKT_MODE(rbufp->pkt.li_vn_mode),
-			 PKT_TO_STRATUM(rbufp->pkt.stratum),
-			 rbufp->pkt.ppoll, rbufp->pkt.precision,
-			 rbufp->pkt.rootdelay, rbufp->pkt.rootdisp,
-			 /* FIXME: this cast is disgusting */
-			 *(const uint32_t*)rbufp->pkt.refid,
-			 outcount);
+	/* Record good packet */
+	record_raw_stats(peer, rbufp, 0, outcount);
 
 	/* If either burst mode is armed, enable the burst.
 	 * Compute the headway for the next packet and delay if
@@ -672,11 +661,48 @@ receive(
 	struct peer *peer = NULL;
 	unsigned short restrict_mask;
 	auth_info* auth = NULL;  /* !NULL if authenticated */
+	int mode;
 
-	stat_count.sys_received++;
+#ifdef ENABLE_MSSNTP
+	uint8_t zero_key[MSSNTP_QUERY_MAC_LEN];
+	memset(&zero_key, 0, MSSNTP_QUERY_MAC_LEN);
+#endif /* ENABLE_MSSNTP */
 
-	if(!is_vn_mode_acceptable(rbufp)) {
-		stat_count.sys_badlength++;
+	stat_proto_total.sys_received++;
+
+#ifdef NTPv1
+	/*Hack for NTPv1.  See #707 */
+	if ((NTPv1 == PKT_VERSION(rbufp->recv_buffer[0])) && \
+		(48 == rbufp->recv_length)) {
+	  mode = PKT_MODE(rbufp->recv_buffer[0]);
+	  stat_proto_total.sys_version1++;
+	  /* There is a lot of crufty old NTPv1 SNTP traffic.
+	   * NTPv1 has no mode field. */
+	  switch (mode) {
+	    case MODE_CLIENT:
+		/* Some packets come with MODE_CLIENT. They will just work. */
+		stat_proto_total.sys_version1client++;
+		break;
+	    case MODE_UNSPEC:
+		/* Some packets come with 0. Patch them.
+		 *   This assumes the client doesn't check the returned mode. */
+		rbufp->recv_buffer[0] = PKT_LI_VN_MODE(0, NTPv1, MODE_CLIENT);
+		stat_proto_total.sys_version1zero++;
+		break;
+	    case MODE_ACTIVEx:
+		/* Windows XP and Server 2003 send MODE_ACTIVE
+		 *   https://kb.meinbergglobal.com/kb/time_sync/timekeeping_on_windows/configuring_w32time_as_ntp_client
+		 * Again, this assumes the client doesn't check the returned mode. */
+		rbufp->recv_buffer[0] = PKT_LI_VN_MODE(0, NTPv1, MODE_CLIENT);
+		stat_proto_total.sys_version1symm++;
+		break;
+	    default:
+		break;
+	  }
+	}
+#endif
+	if(!is_packet_not_low_rot(rbufp)) {
+		stat_proto_total.sys_badlength++;
 		return;
 	}
 
@@ -685,19 +711,19 @@ receive(
 	restrict_mask = restrictions(&rbufp->recv_srcadr);
 
 	if(check_early_restrictions(rbufp, restrict_mask)) {
-		stat_count.sys_restricted++;
+		stat_proto_total.sys_restricted++;
 		return;
 	}
 
 	restrict_mask = ntp_monitor(rbufp, restrict_mask);
 	if (restrict_mask & RES_LIMITED) {
-		stat_count.sys_limitrejected++;
+		stat_proto_total.sys_limitrejected++;
 		if(!(restrict_mask & RES_KOD)) { return; }
 	}
 
 	if(is_control_packet(rbufp)) {
 		process_control(rbufp, restrict_mask);
-		stat_count.sys_processed++;
+		stat_proto_total.sys_processed++;
 		return;
 	}
 
@@ -708,33 +734,49 @@ receive(
 	{
 	uint8_t hisversion = PKT_VERSION(rbufp->recv_buffer[0]);
 	if (hisversion == NTP_VERSION) {
-		stat_count.sys_newversion++;		/* new version */
+		stat_proto_total.sys_newversion++;		/* new version */
 	} else if (!(restrict_mask & RES_VERSION) && hisversion >=
 	    NTP_OLDVERSION) {
-		stat_count.sys_oldversion++;		/* previous version */
+		stat_proto_total.sys_oldversion++;		/* previous version */
 	} else {
-		stat_count.sys_badlength++;
+		stat_proto_total.sys_badlength++;
 		return;			/* old version */
 	}
 	}
 
 	if (!parse_packet(rbufp)) {
-		stat_count.sys_badlength++;
+		stat_proto_total.sys_badlength++;
 		return;
 	}
 
-	if (MODE_SERVER == PKT_MODE(rbufp->pkt.li_vn_mode)) {
+	mode = PKT_MODE(rbufp->pkt.li_vn_mode);
+	if (MODE_SERVER == mode) {
 	    /* Reply to our request:
 	     * Auth check breaks if we findpeer for MODE_CLIENT and
 	     * a site we are using as a server uses us as a server
 	     * with a different key. */
 	    peer = findpeer(rbufp);
 	    if (NULL == peer) {
-		stat_count.sys_declined++;
+		stat_proto_total.sys_declined++;
 		return;
 	    }
 	}
 
+#ifdef ENABLE_MSSNTP
+	// If in an MS-SNTP restrict range, with a 16-byte all zero authenticator,
+	if((RES_MSSNTP == (restrict_mask & RES_MSSNTP))
+	    &&(MSSNTP_QUERY_MAC_LEN == rbufp->mac_len)
+	    &&(0 == memcmp(&(rbufp->recv_buffer[LEN_PKT_NOMAC + 4]),
+		                 zero_key, MSSNTP_QUERY_MAC_LEN))
+	) {
+	    // switch off the check that was breaking MS-SNTP;
+	    rbufp->keyid_present = false;
+	} else {
+	    // otherwise, switch off MS-SNTP to not break all other time service.
+	    restrict_mask &= ~RES_MSSNTP;
+	}
+#endif /* ENABLE_MSSNTP */
+
 	if(i_require_authentication(peer, restrict_mask) ||
 	    /* He wants authentication */
 	    rbufp->keyid_present) {
@@ -763,7 +805,7 @@ receive(
 				 (int)(rbufp->recv_length - (rbufp->mac_len + 4)),
 				 (int)(rbufp->mac_len + 4))) {
 
-			stat_count.sys_badauth++;
+			stat_proto_total.sys_badauth++;
 			if(peer != NULL) {
 				peer->badauth++;
 				peer->cfg.flags &= ~FLAG_AUTHENTIC;
@@ -773,36 +815,39 @@ receive(
 		}
 	}
 
-	switch (PKT_MODE(rbufp->pkt.li_vn_mode)) {
-	    case MODE_ACTIVE:  /* remote site using "peer" in config file */
+	switch (mode) {
 	    case MODE_CLIENT:  /* Request for us as a server. */
 		if (rbufp->extens_present
+#ifndef DISABLE_NTS
 		    && !extens_server_recv(&rbufp->ntspacket,
-			  rbufp->recv_buffer, rbufp->recv_length)) {
-			stat_count.sys_declined++;
-			maybe_log_junk(rbufp);
+			  rbufp->recv_buffer, rbufp->recv_length)
+#endif
+) {
+			stat_proto_total.sys_declined++;
+			maybe_log_junk("EX-REQ", rbufp);
 			break;
 		}
-		handle_fastxmit(rbufp, restrict_mask, auth);
-		stat_count.sys_processed++;
+		fast_xmit(rbufp, auth, restrict_mask);
+		stat_proto_total.sys_processed++;
 		break;
 	    case MODE_SERVER:  /* Reply to our request to a server. */
-		if (NULL == peer) {
-		    stat_count.sys_declined++;
-		    break;
-		}
+/* FIXME: Where is the shared key case tested? */
 		if ((peer->cfg.flags & FLAG_NTS)
-		     && (!rbufp->extens_present || !extens_client_recv(peer,
-		          rbufp->recv_buffer, rbufp->recv_length))) {
-		    stat_count.sys_declined++;
-		    maybe_log_junk(rbufp);
+		     && (!rbufp->extens_present
+#ifndef DISABLE_NTS
+ || !extens_client_recv(peer,
+		          rbufp->recv_buffer, rbufp->recv_length)
+#endif
+)) {
+		    stat_proto_total.sys_declined++;
+		    maybe_log_junk("EX-REP", rbufp);
 		    break;
 		}
 		peer->received++;
 		peer->cfg.flags |= FLAG_AUTHENTIC;
 		peer->timereceived = current_time;
 		handle_procpkt(rbufp, peer);
-		stat_count.sys_processed++;
+		stat_proto_total.sys_processed++;
 		peer->processed++;
 		break;
 	    default:
@@ -810,7 +855,7 @@ receive(
 		   which are a security nightmare.  So they go to the
 		   bit bucket until this improves.
 		*/
-		stat_count.sys_declined++;
+		stat_proto_total.sys_declined++;
 		break;
 	}
 
@@ -1028,10 +1073,10 @@ clock_update(
 		+ loop_data.clock_phi * (current_time - peer->update)
 		+ fabs(clkstate.sys_offset);
 
-	if (dtemp > sys_mindisp)
+	if (dtemp > sys_mindist)
 		sys_vars.sys_rootdisp = dtemp;
 	else
-		sys_vars.sys_rootdisp = sys_mindisp;
+		sys_vars.sys_rootdisp = sys_mindist;
 	sys_vars.sys_rootdelay = peer->delay + peer->rootdelay;
 	sys_vars.sys_reftime = peer->dst;
 
@@ -1066,7 +1111,7 @@ clock_update(
 		}
 #endif /* HAVE_LIBSCF_H */
 		msyslog(LOG_ERR, "CLOCK: Panic: offset too big: %.3f",
-            clkstate.sys_offset);
+			clkstate.sys_offset);
 		exit (1);
 		/* not reached */
 
@@ -1101,13 +1146,11 @@ clock_update(
 			 * If our parent process is waiting for the
 			 * first clock sync, send them home satisfied.
 			 */
-#ifdef HAVE_WORKING_FORK
 			if (waitsync_fd_to_close != -1) {
 				close(waitsync_fd_to_close);
 				waitsync_fd_to_close = -1;
 				DPRINT(1, ("notified parent --wait-sync is done\n"));
 			}
-#endif /* HAVE_WORKING_FORK */
 
 		}
 
@@ -1217,7 +1260,9 @@ poll_update(
 			next = 1U << hpoll;
 		else
 #endif /* REFCLOCK */
-			next = ((0x1000UL | (ntp_random() & 0x0ff)) <<
+			/* add a bit of randomess to next polling time
+			 * to disperse traffic */
+			next = ((0x1000UL | (random() & 0x0ff)) <<
 			    hpoll) >> 12;
 		next += peer->outdate;
 		if (next > utemp)
@@ -1252,7 +1297,7 @@ peer_clear(
 	 * Clear all values, including the optional crypto values above.
 	 */
 	memset(CLEAR_TO_ZERO(peer), 0, LEN_CLEAR_TO_ZERO(peer));
-	peer->ppoll = peer->cfg.maxpoll;
+	peer->ppoll = NTP_MAXPOLL_UNK;
 	peer->hpoll = peer->cfg.minpoll;
 	peer->disp = sys_maxdisp;
 	peer->flash = peer_unfit(peer);
@@ -1283,8 +1328,6 @@ peer_clear(
 	peer->nextdate = peer->update = peer->outdate = current_time;
 	if (initializing1) {
 		peer->nextdate += (unsigned long)peer_associations;
-	} else if (MODE_PASSIVE == peer->hmode) {
-		peer->nextdate += (unsigned long)rstrct.ntp_minpkt;
 	} else {
 	    /*
 	     * Randomizing the next poll interval used to be done with
@@ -1296,8 +1339,8 @@ peer_clear(
 	     * to be invisible from outside ntpd; the internal
 	     * association ID fits the bill.
 	     */
-	    int pseudorandom = peer->associd ^ sock_hash(&peer->srcadr);
-	    peer->nextdate += (unsigned long)(pseudorandom % (1 << peer->cfg.minpoll));
+	    unsigned int pseudorand = peer->associd ^ sock_hash(&peer->srcadr);
+	    peer->nextdate += (pseudorand % (1 << peer->cfg.minpoll));
 	}
 	DPRINT(1, ("peer_clear: at %u next %u associd %d refid %s\n",
 		   current_time, peer->nextdate, peer->associd,
@@ -1570,6 +1613,9 @@ clock_select(void)
 		 * Leave the island immediately if the peer is
 		 * unfit to synchronize.
 		 */
+		if ((FLAG_LOOKUP & peer->cfg.flags) ||
+                        (MDF_POOL & peer->cast_flags))
+			continue;
 		if (peer_unfit(peer)) {
 			continue;
 }
@@ -1792,7 +1838,7 @@ clock_select(void)
 	 */
 	if (nlist == 0) {
 		peers[0].error = 0;
-		peers[0].synch = sys_mindisp;
+		peers[0].synch = sys_mindist;
 #ifdef REFCLOCK
 		if (typemodem != NULL) {
 			peers[0].peer = typemodem;
@@ -1874,11 +1920,11 @@ clock_select(void)
 	 *
 	 * Choose the system peer using a hybrid metric composed of the
 	 * selection jitter scaled by the root distance augmented by
-	 * stratum scaled by sys_mindisp (.001 by default). The goal of
+	 * stratum scaled by sys_mindist (.001 by default). The goal of
 	 * the small stratum factor is to avoid clockhop between a
 	 * reference clock and a network peer which has a refclock and
 	 * is using an older ntpd, which does not floor sys_rootdisp at
-	 * sys_mindisp.
+	 * sys_mindist.
 	 *
 	 * In contrast, ntpd 4.2.6 and earlier used stratum primarily
 	 * in selecting the system peer, using a weight of 1 second of
@@ -1912,7 +1958,7 @@ clock_select(void)
 		if (peer->cfg.flags & FLAG_PREFER)
 			sys_prefer = peer;
 		speermet = peers[i].seljit * peers[i].synch +
-		    peer->stratum * sys_mindisp;
+		    peer->stratum * sys_mindist;
 		if (speermet < e) {
 			e = speermet;
 			speer = i;
@@ -1933,9 +1979,9 @@ clock_select(void)
 		if (osys_peer == NULL || osys_peer == typesystem) {
 			sys_clockhop = 0;
 		} else if ((x = fabs(typesystem->offset -
-		    osys_peer->offset)) < sys_mindisp) {
+		    osys_peer->offset)) < sys_mindist) {
 			if ( D_ISZERO_NS(sys_clockhop) ) {
-				sys_clockhop = sys_mindisp;
+				sys_clockhop = sys_mindist;
 			} else {
 				sys_clockhop *= .5;
 			}
@@ -2100,8 +2146,8 @@ root_distance(
 	 * cannot exceed the sys_maxdist, as this is the cutoff by the
 	 * selection algorithm.
 	 */
-	if (dtemp < sys_mindisp) {
-		dtemp = sys_mindisp;
+	if (dtemp < sys_mindist) {
+		dtemp = sys_mindist;
 	}
 	return (dtemp);
 }
@@ -2124,7 +2170,7 @@ peer_xmit(
 	sendlen = LEN_PKT_NOMAC;
 	if (NTP_VERSION == peer->cfg.version) {
 		/* Hide most of info for privacy
-		 * RFC in progress - draft-ietf-ntp-data-minimization, 2018-Jul-07
+		 * RFC in progress - draft-ietf-ntp-data-minimization, 2020-Feb-16
 		 */
 		xpkt.li_vn_mode = PKT_LI_VN_MODE(
 			LEAP_NOWARNING, peer->cfg.version, MODE_CLIENT);
@@ -2137,7 +2183,8 @@ peer_xmit(
 		xpkt.reftime = htonl_fp(0);
 		xpkt.org = htonl_fp(0);
 		xpkt.rec = htonl_fp(0);
-		peer->org_rand = ntp_random64();
+		ntp_RAND_bytes((unsigned char *)&peer->org_rand,
+			sizeof(peer->org_rand));
 		get_systime(&peer->org_ts);	/* as late as possible */
 	} else {
 		xpkt.li_vn_mode = PKT_LI_VN_MODE(
@@ -2159,18 +2206,20 @@ peer_xmit(
 	xpkt.xmt = htonl_fp(peer->org_rand);	/* out in xmt, back in org */
 
 	/* 3 way branch to add authentication:
-         *  1) NTS
-         *  2) Shared KEY
-         *  3) none
+	 *  1) NTS
+	 *  2) Shared KEY
+	 *  3) none
 	 */
 	if (FLAG_NTS & peer->cfg.flags) {
+#ifndef DISABLE_NTS
 		if (0 < peer->nts_state.count)
 		  sendlen += extens_client_send(peer, &xpkt);
 		else {
 		  restart_nts_ke(peer);  /* out of cookies */
 		  return;
 		}
-        } else if (0 != peer->cfg.peerkey) {
+#endif
+	} else if (0 != peer->cfg.peerkey) {
 		auth_info *auth = authlookup(peer->cfg.peerkey, true);
 		if (NULL == auth) {
 			report_event(PEVNT_AUTH, peer, "no key");
@@ -2185,6 +2234,7 @@ peer_xmit(
 
 	peer->sent++;
         peer->outcount++;
+        peer->bogons = 0;
 	peer->throttle += (1 << peer->cfg.minpoll) - 2;
 	DPRINT(1, ("transmit: at %u %s->%s mode %d keyid %08x len %u\n",
 		   current_time, peer->dstadr ?
@@ -2197,9 +2247,8 @@ peer_xmit(
 #ifdef ENABLE_LEAP_SMEAR
 
 static void
-leap_smear_add_offs(l_fp *t, l_fp *t_recv) {
-	UNUSED_ARG(t_recv);
-	t += leap_smear.offset;
+leap_smear_add_offs(l_fp *t) {
+	*t += leap_smear.offset;
 }
 
 #endif	/* ENABLE_LEAP_SMEAR */
@@ -2211,13 +2260,13 @@ leap_smear_add_offs(l_fp *t, l_fp *t_recv) {
 static void
 fast_xmit(
 	struct recvbuf *rbufp,	/* receive packet pointer */
-	int	xmode,		/* receive mode */
 	auth_info *auth,	/* !NULL for authentication */
 	int	flags		/* restrict mask */
 	)
 {
 	struct pkt xpkt;	/* transmit packet structure */
-	l_fp	xmt_tx, xmt_ty;
+	l_fp	xmt_tx;
+	struct timespec	start, finish;
 	size_t	sendlen;
 
 	/*
@@ -2237,9 +2286,9 @@ fast_xmit(
 	 * synchronization.
 	 */
 	if (flags & RES_KOD) {
-		stat_count.sys_kodsent++;
+		stat_proto_total.sys_kodsent++;
 		xpkt.li_vn_mode = PKT_LI_VN_MODE(LEAP_NOTINSYNC,
-		    PKT_VERSION(rbufp->pkt.li_vn_mode), xmode);
+		    PKT_VERSION(rbufp->pkt.li_vn_mode), MODE_SERVER);
 		xpkt.stratum = STRATUM_PKT_UNSPEC;
 		xpkt.ppoll = max(rbufp->pkt.ppoll, rstrct.ntp_minpoll);
 		xpkt.precision = rbufp->pkt.precision;
@@ -2274,8 +2323,15 @@ fast_xmit(
 		 * reftime to make sure the reftime isn't later than
 		 * the transmit/receive times.
 		 */
+		/* Note: This returns the same data for all versions.
+		 * Currently, the mode is always Server.
+		 * The version is copied from the request.
+		 * There are minor differences between v3 and v4.
+		 * So far, nobody cares.
+		 * Note: There is significant NTPv1 traffic.  See #707
+		 */
 		xpkt.li_vn_mode = PKT_LI_VN_MODE(sys_vars.sys_leap,
-		    PKT_VERSION(rbufp->pkt.li_vn_mode), xmode);
+		    PKT_VERSION(rbufp->pkt.li_vn_mode), MODE_SERVER);
 		xpkt.stratum = STRATUM_TO_PKT(sys_vars.sys_stratum);
 		xpkt.ppoll = max(rbufp->pkt.ppoll, rstrct.ntp_minpoll);
 		xpkt.precision = sys_vars.sys_precision;
@@ -2286,7 +2342,7 @@ fast_xmit(
 #ifdef ENABLE_LEAP_SMEAR
 		this_ref_time = sys_vars.sys_reftime;
 		if (leap_smear.in_progress) {
-			leap_smear_add_offs(&this_ref_time, NULL);
+			leap_smear_add_offs(&this_ref_time);
 			xpkt.refid = convertLFPToRefID(leap_smear.offset);
 			DPRINT(2, ("fast_xmit: leap_smear.in_progress: refid %8x, smear %s\n",
 				ntohl(xpkt.refid),
@@ -2304,7 +2360,7 @@ fast_xmit(
 #ifdef ENABLE_LEAP_SMEAR
 		this_recv_time = rbufp->recv_time;
 		if (leap_smear.in_progress)
-			leap_smear_add_offs(&this_recv_time, NULL);
+			leap_smear_add_offs(&this_recv_time);
 		xpkt.rec = htonl_fp(this_recv_time);
 #else
 		xpkt.rec = htonl_fp(rbufp->recv_time);
@@ -2313,7 +2369,7 @@ fast_xmit(
 		get_systime(&xmt_tx);
 #ifdef ENABLE_LEAP_SMEAR
 		if (leap_smear.in_progress)
-			leap_smear_add_offs(&xmt_tx, &this_recv_time);
+			leap_smear_add_offs(&xmt_tx);
 #endif
 		xpkt.xmt = htonl_fp(xmt_tx);
 	}
@@ -2321,10 +2377,7 @@ fast_xmit(
 
 #ifdef ENABLE_MSSNTP
 	if (flags & RES_MSSNTP) {
-		keyid_t keyid = 0;
-		if (NULL != auth) keyid = auth->keyid;
-		// FIXME need counter
-		send_via_ntp_signd(rbufp, xmode, keyid, flags, &xpkt);
+		send_via_ntp_signd(rbufp, &xpkt); // Simplified the API
 		return;
 	}
 #endif /* ENABLE_MSSNTP */
@@ -2336,22 +2389,31 @@ fast_xmit(
          *  3) none
 	 */
 	sendlen = LEN_PKT_NOMAC;
-	get_systime(&xmt_tx);
+	clock_gettime(CLOCK_MONOTONIC, &start);
 	if (rbufp->ntspacket.valid) {
+#ifndef DISABLE_NTS
 	  sendlen += extens_server_send(&rbufp->ntspacket, &xpkt);
+#endif
         } else if (NULL != auth) {
 	  sendlen += (size_t)authencrypt(auth, (uint32_t *)&xpkt, (int)sendlen);
         }
+	if (sendlen > rbufp->recv_length) {
+	  /* About to send a response that is bigger than the request.
+	   * That can be used for DDoS amplification, so don't do that.
+	   * This shouldn't happen, but check here in case of a bug.
+	   */
+	  maybe_log_junk("DDoS", rbufp);	/* needs a counter */
+	  return;
+	}
 	sendpkt(&rbufp->recv_srcadr, rbufp->dstadr, &xpkt, (int)sendlen);
-	get_systime(&xmt_ty);
-	xmt_ty -= xmt_tx;
-	sys_authdelay = xmt_ty;
+	clock_gettime(CLOCK_MONOTONIC, &finish);
+	sys_authdelay = tspec_intv_to_lfp(sub_tspec(finish, start));
 	/* Previous versions of this code had separate DPRINT-s so it
 	 * could print the key on the auth case.  That requires separate
 	 * sendpkt-s on each branch or the DPRINT pollutes the timing. */
-	DPRINT(1, ("transmit: at %u %s->%s mode %d len %zu\n",
+	DPRINT(1, ("transmit: at %u %s->%s len %zu\n",
 		   current_time, socktoa(&rbufp->dstadr->sin),
-		   socktoa(&rbufp->recv_srcadr), xmode, sendlen));
+		   socktoa(&rbufp->recv_srcadr), sendlen));
 }
 
 
@@ -2364,7 +2426,6 @@ dns_take_server(
 	sockaddr_u *rmtadr
 	)
 {
-	int		restrict_mask;
 	struct peer *	pp;
 
 	if(!(server->cfg.flags & FLAG_LOOKUP))
@@ -2382,17 +2443,30 @@ dns_take_server(
           msyslog(LOG_INFO, "DNS: Server taking: %s", socktoa(rmtadr));
         else
           msyslog(LOG_INFO, "DNS: Server taking: %s", sockporttoa(rmtadr));
+
 	server->cfg.flags &= (unsigned)~FLAG_LOOKUP;
-
 	server->srcadr = *rmtadr;
-	peer_add_hash(server);
-
-	restrict_mask = restrictions(&server->srcadr);
-	if (RES_FLAGS & restrict_mask) {
-		msyslog(LOG_INFO, "DNS: Server poking hole in restrictions for: %s",
-			socktoa(&server->srcadr));
-		restrict_source(&server->srcadr, false, 0);
+	{ /* HACK: fixup refid to show progress */
+	char tag[REFIDLEN];
+	if (server->cfg.flags & FLAG_NTS) {
+		memcpy(tag, "NTS?", REFIDLEN);
+	} else {
+		memcpy(tag, "DNS?", REFIDLEN);
 	}
+	switch(AF(rmtadr)) {
+	  case AF_INET:
+		tag[REFIDLEN-1] = '4';
+		break;
+	  case AF_INET6:
+		tag[REFIDLEN-1] = '6';
+		break;
+	  default:
+		break;
+	}
+	memcpy(&server->refid, tag, REFIDLEN);
+	}
+	peer_add_hash(server);
+	restrict_source(server);
 
 	peer_refresh_interface(server);
 
@@ -2415,7 +2489,6 @@ dns_take_pool(
 {
 	struct peer_ctl		pctl;
 	struct peer *		peer;
-	int			restrict_mask;
 	endpt *			lcladr;
 
 	peer = findexistingpeer(rmtadr, NULL, NULL, MODE_CLIENT);
@@ -2436,21 +2509,12 @@ dns_take_pool(
 	pctl.mode = 0;
 	pctl.peerkey = 0;
 	peer = newpeer(rmtadr, NULL, lcladr,
-		       MODE_CLIENT, &pctl, MDF_UCAST | MDF_UCLNT, false);
+		       MODE_CLIENT, &pctl, MDF_UCAST, false);
 	peer_xmit(peer);
 	if (peer->cfg.flags & FLAG_IBURST)
 	  peer->retry = NTP_RETRY;
 	poll_update(peer, peer->hpoll);
 
-	restrict_mask = restrictions(&peer->srcadr);
-	/* FIXME-DNS: RES_FLAGS includes RES_DONTSERVE?? */
-	if (RES_FLAGS & restrict_mask) {
-		msyslog(LOG_INFO, "DNS: Pool poking hole in restrictions for: %s",
-				socktoa(&peer->srcadr));
-		restrict_source(&peer->srcadr, false,
-				current_time + POOL_SOLICIT_WINDOW + 1);
-	}
-
 	DPRINT(1, ("dns_take_pool: at %u %s->%s pool\n",
 		   current_time, latoa(lcladr), socktoa(rmtadr)));
 }
@@ -2516,6 +2580,7 @@ void dns_take_status(struct peer* peer, DNS_Status status) {
 	peer->nextdate = current_time + (1U << hpoll);
 }
 
+#ifndef DISABLE_NTS
 /* NTS out of cookies
  * Beware of clutter in NTS-KE server logs
  * There are actually several cases:
@@ -2531,20 +2596,22 @@ static void restart_nts_ke(struct peer *peer) {
 		hpoll = 8;      /* min retry: 256 seconds, ~5 min */
 	if (hpoll > 12)
 		hpoll = 12;	/* 4096, a bit over an hour */
+	peer->ppoll = NTP_MAXPOLL_UNK;
 	peer->hpoll = hpoll;
 	peer->nextdate = current_time + (1U << hpoll);
 	peer->cfg.flags |= FLAG_LOOKUP;
 };
+#endif
 
 /*
- * dns_new_interface
- *   A new interface is now active
- *   retry danging DNS lookups
+ * dns_try_again
+ *   retry danging DNS and NTS lookups
  */
-void dns_new_interface(void) {
+void dns_try_again(void) {
 	struct peer *p;
 	for (p = peer_list; p != NULL; p = p->p_link) {
 		if ((p->cfg.flags & FLAG_LOOKUP) || (p->cast_flags & MDF_POOL)) {
+			p->ppoll = NTP_MAXPOLL_UNK;
 			p->hpoll = p->cfg.minpoll;
 			transmit(p);   /* does all the work */
 		}
@@ -2552,7 +2619,6 @@ void dns_new_interface(void) {
 }
 
 
-
 /*
  * local_refid(peer) - check peer refid to avoid selecting peers
  *		       currently synced to this ntpd.
@@ -2564,7 +2630,9 @@ local_refid(
 {
 	endpt *	unicast_ep;
 
-	if (p->dstadr != NULL && !(INT_MCASTIF & p->dstadr->flags))
+// FIXME: Is this ever NULL?  MCAST leftover??
+// (actually, just check)
+	if (p->dstadr != NULL)
 		unicast_ep = p->dstadr;
 	else
 		unicast_ep = findinterface(&p->srcadr);
@@ -2639,6 +2707,7 @@ peer_unfit(
 }
 
 
+#ifdef ENABLE_FUZZ
 /*
  * Find the precision of this particular machine
  */
@@ -2746,7 +2815,6 @@ measure_tick_fuzz(void)
 	return tick;
 }
 
-
 void
 set_sys_tick_precision(
 	double tick
@@ -2784,7 +2852,7 @@ set_sys_tick_precision(
 
 	sys_vars.sys_precision = (int8_t)i;
 }
-
+#endif
 
 /*
  * init_proto - initialize the protocol module's data
@@ -2806,13 +2874,18 @@ init_proto(const bool verbose)
 	sys_vars.sys_rootdisp = 0;
 	sys_vars.sys_reftime = 0;
 	clkstate.sys_jitter = 0;
+#ifdef ENABLE_FUZZ
 	measure_precision(verbose);
+#else
+	UNUSED_ARG(verbose);
+	sys_vars.sys_precision = -30; /* ns */
+#endif
 	get_systime(&dummy);
 	sys_survivors = 0;
-	stat_count.sys_stattime = current_time;
+	sys_stattime = current_time;
 	orphwait = current_time + (unsigned long)sys_orphwait;
 	proto_clr_stats();
-	stat_count.use_stattime = current_time;
+	use_stattime = current_time;
 	clock_ctl.hardpps_enable = false;
 	stats_control = true;
 }
@@ -2851,9 +2924,9 @@ proto_config(
 
 	case PROTO_MONITOR:	/* monitoring (monitor) */
 		if (value) {
-			mon_start(MON_ON);
+			mon_setup(MON_ON);
 		} else {
-			mon_stop(MON_ON);
+			mon_setdown(MON_ON);
 			if (mon_data.mon_enabled)
 				msyslog(LOG_WARNING,
 					"CONFIG: 'monitor' cannot be disabled while 'limited' is enabled");
@@ -2896,15 +2969,12 @@ proto_config(
 		sys_maxdist = dvalue;
 		break;
 
-	case PROTO_CALLDELAY:	/* modem call delay (mdelay) */
-		break;		/* NOT USED */
-
 	case PROTO_MINCLOCK:	/* minimum candidates (minclock) */
 		sys_minclock = (int)dvalue;
 		break;
 
-	case PROTO_MINDISP:	/* minimum distance (mindist) */
-		sys_mindisp = dvalue;
+	case PROTO_MINDIST:	/* minimum distance (mindist) */
+		sys_mindist = dvalue;
 		break;
 
 	case PROTO_MINSANE:	/* minimum survivors (minsane) */
@@ -2934,38 +3004,51 @@ proto_config(
 void
 proto_clr_stats(void)
 {
-	stat_count.sys_stattime = current_time;
-	stat_count.sys_received = 0;
-	stat_count.sys_processed = 0;
-	stat_count.sys_newversion = 0;
-	stat_count.sys_oldversion = 0;
-	stat_count.sys_declined = 0;
-	stat_count.sys_restricted = 0;
-	stat_count.sys_badlength = 0;
-	stat_count.sys_badauth = 0;
-	stat_count.sys_limitrejected = 0;
-	stat_count.sys_kodsent = 0;
+    stat_proto_hourago = stat_proto_total;
+    sys_stattime = current_time;
 }
 
 
-/* limit logging so bad guys can't DDoS us by sending crap
- * log first 100 and 10/hour
- * This gets too-old cookies
- */
+/* limit logging so bad guys can't DDoS us by sending crap */
+
+void maybe_log_junk(const char *tag, struct recvbuf *rbufp) {
+  static float junk_limit = 2.0;         /* packets per hour */
+  static float junk_score = 0;           /* score, packets/hour */
+  static float junk_decay = 2.0;         /* hours, exponential decay time */
+  static l_fp  junk_last = 0;            /* time of last attempted print */
+  static long  junk_count = 0;           /* total count */
+  static long  junk_print = 0;           /* printed count */
+#define JUNKSIZE 500
+    char buf[JUNKSIZE];
+    int lng = rbufp->recv_length;
+    int i, j;
+
+    junk_count++;
+    if (0 == junk_last) {
+      /* first time */
+      junk_last = rbufp->recv_time;
+    } else {
+      l_fp interval_fp = rbufp->recv_time - junk_last;
+      float since_last = ldexpf(interval_fp, -32)/3600.0;
+      junk_last = rbufp->recv_time;
+      junk_score *= expf(-since_last/junk_decay);
+      if (junk_limit < junk_score)
+	return;
+    }
+    junk_print++;
+    junk_score += 1.0/junk_decay;  /* only count the ones we print */
 
-void maybe_log_junk(struct recvbuf *rbufp) {
-    static unsigned int noise_try = 0;
-    noise_try++;
-    if ((noise_try>100) && (((noise_try-90)*3600/current_time) > 10))
-      return;
     msyslog(LOG_INFO,
-	"JUNK: M%d V%d 0/%2x%2x%2x%2x 48/%2x%2x%2x%2x from %s, lng=%ld",
-	PKT_MODE(rbufp->pkt.li_vn_mode), PKT_VERSION(rbufp->pkt.li_vn_mode),
-	rbufp->recv_buffer[0], rbufp->recv_buffer[1],
-	rbufp->recv_buffer[2], rbufp->recv_buffer[3],
-	rbufp->recv_buffer[48+0], rbufp->recv_buffer[48+1],
-	rbufp->recv_buffer[48+2], rbufp->recv_buffer[48+3],
-	sockporttoa(&rbufp->recv_srcadr),
-	(long)rbufp->recv_length);
+	"%s: Count=%ld Print=%ld, Score=%.3f, M%d V%d from %s, lng=%d",
+	tag, junk_count, junk_print, junk_score,
+        PKT_MODE(rbufp->pkt.li_vn_mode), PKT_VERSION(rbufp->pkt.li_vn_mode),
+        sockporttoa(&rbufp->recv_srcadr), lng);
+    for (i=0,j=0; i<lng; i++) {
+      if ((j+4)>JUNKSIZE) break;
+      if (0 == (i%4)) buf[j++] = ' ';
+      j += snprintf(&buf[j], (JUNKSIZE-j), "%02x", rbufp->recv_buffer[i]);
+    }
+    msyslog(LOG_INFO,
+	"%s: %s", tag, buf);
 }
 
