@@ -406,7 +406,7 @@ refclock_process_f(
 	 * to deducing the year from the receipt time of the sample if
 	 * it finds only a 2-digit year in the timecode.
 	 */
-	if (!clocktime(pp->year, pp->day, pp->hour, pp->minute, pp->second,
+	if (!clocktime(pp->year, pp->yday, pp->hour, pp->minute, pp->second,
 		       time(NULL), lfpuint(pp->lastrec), &pp->yearstart, &sec)) {
 		return false;
 	}
@@ -469,7 +469,10 @@ refclock_sample(
 	 * approximately 60 percent of the samples remain.
 	 */
 	i = 0; j = n;
-	m = n - (n * 4) / 10;
+	/* m=n*6/10 doesn't work.  It rounds down.  We need up. */
+	m = n-(n*4/10);
+	/* CID 497295, Bug in Coverity: goes around loop when it shouldn't */
+	/* coverity[INTEGER_OVERFLOW] */
 	while ((j - i) > m) {
 		offset = off[(j + i) / 2];
 		if (off[j - 1] - offset < offset - off[i])
@@ -541,7 +544,7 @@ refclock_receive(
 		return;
 
 	clock_filter(peer, pp->offset, 0., pp->jitter);
-	if (cal_enable && fabs(clkstate.last_offset) < sys_mindisp &&
+	if (cal_enable && fabs(clkstate.last_offset) < sys_mindist &&
 		sys_vars.sys_peer != NULL) {
 		if (sys_vars.sys_peer->is_pps_driver &&
 		    !peer->is_pps_driver)
