@@ -6,8 +6,8 @@
  * Written By:	Sachin Kamboj
  *		University of Delaware
  *		Newark, DE 19711
- * Copyright (c) 2006
- * Copyright 2015 by the NTPsec project contributors
+ * Copyright Sachin Kamboj
+ * Copyright the NTPsec project contributors
  * SPDX-License-Identifier: BSD-2-Clause
  */
 
@@ -153,6 +153,8 @@ lex_open(
 		stream->fpi = fopen(path, mode);
 		if (NULL == stream->fpi) {
 			free(stream);
+			msyslog(LOG_ERR, "CONFIG: failed to open \'%s\': %s",
+				path, strerror(errno));
 			stream = NULL;
 		}
 	}
@@ -180,6 +182,7 @@ lex_getch(
 			conf_file_sum += (unsigned int)ch;
 	} else if (stream->fpi) {
 		/* fetch next 7-bit ASCII char (or EOF) from file */
+		/* coverity[tainted_scalar] */
 		while ((ch = fgetc(stream->fpi)) != EOF && ch > SCHAR_MAX) {
 			stream->curpos.ncol++;
 		}
@@ -326,7 +329,7 @@ lex_init_stack(
  * anything until the next 'lex_init_stack()' succeeded.
  */
 void
-lex_drop_stack()
+lex_drop_stack(void)
 {
 	lex_stack = _drop_stack_do(lex_stack);
 }
@@ -340,7 +343,7 @@ lex_drop_stack()
  * in the force-eof mode before this call.
  */
 bool
-lex_flush_stack() {
+lex_flush_stack(void) {
 	bool retv = false;
 
 	if (NULL != lex_stack) {
@@ -419,6 +422,10 @@ bool lex_push_file(
 			if ((dfd = opendir(fullpath)) == NULL)
 				return false;
 			baselist = (char **)malloc(sizeof(char *));
+			if (NULL == baselist) {
+				msyslog(LOG_ERR, "CONFIG: lex_push_file: NULL from malloc");
+				exit(3);
+			}
 			while ((dp = readdir(dfd)) != NULL)
 			{
 				if (!CONF_ENABLE(dp->d_name)) {
@@ -515,7 +522,7 @@ lex_from_file(void)
 }
 
 struct FILE_INFO *
-lex_current()
+lex_current(void)
 {
 	/* this became so simple, it could be a macro. But then,
 	 * lex_stack needed to be global...
