@@ -24,6 +24,7 @@
 # include <sys/ioctl.h>
 #endif /* HAVE_SYS_IOCTL_H */
 #include <sched.h>
+#include <pthread.h>  /* For pthread_setschedparam() */
 #include <sys/mman.h>
 
 #include <termios.h>
@@ -78,15 +79,11 @@ static char *user;		/* User to switch to */
 static char *group;		/* group to switch to */
 static const char *chrootdir;	/* directory to chroot to */
 
-#ifdef HAVE_WORKING_FORK
 int	waitsync_fd_to_close = -1;	/* -w/--wait-sync */
-#endif
 
 const char *progname;
 
-#if defined(HAVE_WORKING_FORK)
 static int	wait_child_sync_if	(int, long);
-#endif
 
 static	void	catchHUP	(int);
 static	void	catchDNS	(int);
@@ -101,15 +98,12 @@ static	void	no_debug	(int);
 static int	saved_argc;
 static char **	saved_argv;
 
-static int	ntpdmain(int, char **) __attribute__((noreturn));
 static void	mainloop		(void)
 			__attribute__	((__noreturn__));
 static void	set_process_priority	(void);
-static  void    close_all_beyond(int);
-static  void    close_all_except(int);
 
 
-#define ALL_OPTIONS "46abc:dD:f:gGhi:I:k:l:LmnNp:Pqr:Rs:t:u:U:Vw:xzZ"
+#define ALL_OPTIONS "46abc:dD:f:gGhi:I:k:l:LmnNp:P:qr:Rs:t:u:U:Vw:xzZ"
 static const struct option longoptions[] = {
     { "ipv4",		    0, 0, '4' },
     { "ipv6",		    0, 0, '6' },
@@ -412,19 +406,6 @@ parse_cmdline_opts(
 	}
 }
 
-#ifdef NO_MAIN_ALLOWED
-CALL(ntpd,"ntpd",ntpdmain);
-#else	/* !NO_MAIN_ALLOWED follows */
-int
-main(
-	int argc,
-	char *argv[]
-	)
-{
-	return ntpdmain(argc, argv);
-}
-#endif /* !NO_MAIN_ALLOWED */
-
 #ifdef SIGDANGER
 /*
  * Signal handler for SIGDANGER. (AIX)
@@ -452,7 +433,6 @@ set_process_priority(void)
 			 : "Attempt to set priority"
 				));
 # endif /* DEBUG */
-#ifdef HAVE_SCHED_SETSCHEDULER
 	if (need_priority) {
 		int pmax, pmin;
 		struct sched_param sched;
@@ -468,43 +448,40 @@ set_process_priority(void)
 			else
 				sched.sched_priority = config_priority;
 		}
-		if ( sched_setscheduler(0, SCHED_FIFO, &sched) == -1 )
-			msyslog(LOG_ERR, "INIT: sched_setscheduler(): %s", strerror(errno));
+		if ( pthread_setschedparam(pthread_self(), SCHED_FIFO, &sched) != 0 )
+			msyslog(LOG_ERR, "INIT: pthread_setschedparam(): %s", strerror(errno));
 		else
 			need_priority = false;
 	}
-#endif
 	if (need_priority)
 		msyslog(LOG_ERR, "INIT: set_process_priority: No way found to improve our priority");
 }
 
+const char *ntpd_version_string = "ntpd ntpsec-" NTPSEC_VERSION_EXTENDED;
 const char *ntpd_version(void)
 {
-    static char versionbuf[64];
-    snprintf(versionbuf, sizeof(versionbuf),
-	     "ntpd ntpsec-%s", NTPSEC_VERSION_EXTENDED);
-    return versionbuf;
+    return ntpd_version_string;
 }
 
 /*
  * Main program.  Initialize us, disconnect us from the tty if necessary,
  * and loop waiting for I/O and/or timer expiries.
  */
-static int
-ntpdmain(
+int
+main(
 	int argc,
 	char *argv[]
 	)
 {
 	mode_t		uv;
 	uid_t		uid;
-# if defined(HAVE_WORKING_FORK)
 	int		pipe_fds[2];
 	int		rc;
 	int		exit_code;
-	struct sigaction sa;
-# endif	/* HAVE_WORKING_FORK*/
 	int op;
+#ifdef SIGDANGER
+	struct sigaction sa;
+#endif
 
 	uv = umask(0);
 	if (uv) {
@@ -523,11 +500,19 @@ ntpdmain(
 # endif
 
 	init_logging(progname, NLOG_SYNCMASK, true);
+
+	if (!dumpopts) {
+		/* log to syslog before setting up log file */
+		announce_starting();
+	}
+
 	/* honor -l/--logfile option to log to a file */
 	if (logfilename != NULL) {
 		syslogit = false;
 		termlogit = false;
 		change_logfile(logfilename, false);
+		/* Repeat critical info in logfile. Helps debugging. */
+		announce_starting();
 	} else {
 		if (nofork)
 		    termlogit = true;
@@ -535,25 +520,6 @@ ntpdmain(
 			syslogit = false;
 	}
 
-	if (!dumpopts)
-	{
-		char buf[1024];	/* Secret knowledge of msyslog buf length */
-		char *cp = buf;
-
-		msyslog(LOG_NOTICE, "INIT: %s: Starting", ntpd_version());
-
-		/* Note that every arg has an initial space character */
-		snprintf(cp, sizeof(buf), "Command line:");
-		cp += strlen(cp);
-
-		for (int i = 0; i < saved_argc ; ++i) {
-			snprintf(cp, sizeof(buf) - (size_t)(cp - buf),
-				" %s", saved_argv[i]);
-			cp += strlen(cp);
-		}
-		msyslog(LOG_INFO, "INIT: %s", buf);
-	}
-
 	uid = getuid();
 	if (uid && !dumpopts) {
 		termlogit = true;
@@ -562,9 +528,6 @@ ntpdmain(
 		exit(1);
 	}
 
-	set_prettydate_pivot(time(NULL));
-
-# ifdef HAVE_WORKING_FORK
 	/* make sure the FDs are initialised */
 	pipe_fds[0] = -1;
 	pipe_fds[1] = -1;
@@ -583,7 +546,6 @@ ntpdmain(
 	    }
 	    waitsync_fd_to_close = pipe_fds[1];
 	}
-# endif	/* HAVE_WORKING_FORK */
 
 	init_network();
 	/*
@@ -591,7 +553,6 @@ ntpdmain(
 	 */
 	if (!nofork) {
 
-# ifdef HAVE_WORKING_FORK
 		rc = fork();
 		if (-1 == rc) {
 			exit_code = (errno) ? errno : -1;
@@ -600,6 +561,8 @@ ntpdmain(
 		}
 		if (rc > 0) {
 			/* parent */
+			if (pidfile)
+				write_pidfile(pidfile, rc);
 			exit_code = wait_child_sync_if(pipe_fds[0],
 						       wait_sync);
 			exit(exit_code);
@@ -607,37 +570,28 @@ ntpdmain(
 
 		/*
 		 * child/daemon
-		 * close all open files excepting waitsync_fd_to_close.
-		 * msyslog() unreliable until after init_logging().
 		 */
 		termlogit = false;    /* do not use stderr after fork */
-		closelog();
-		close_all_except(waitsync_fd_to_close);
-		INSIST(0 == open("/dev/null", 0) && 1 == dup2(0, 1) \
-			&& 2 == dup2(0, 2));
-
-		init_logging(progname, 0, true);
-		/* we lost our logfile (if any) daemonizing */
-		setup_logfile(logfilename);
+		fclose(stdin);
+		fclose(stdout);
+		fclose(stderr);
+		INSIST(STDIN_FILENO == open("/dev/null", 0) \
+			&& STDOUT_FILENO == dup2(0, 1) \
+			&& STDERR_FILENO == dup2(0, 2));
 
 		if (setsid() == (pid_t)-1)
 			msyslog(LOG_ERR, "INIT: setsid(): %s", strerror(errno));
-#  ifdef SIGDANGER
+#ifdef SIGDANGER
 		/* Don't get killed by low-on-memory signal. */
 		sa.sa_handler = catch_danger;
 		sigemptyset(&sa.sa_mask);
 		sa.sa_flags = SA_RESTART;
 		sigaction(SIGDANGER, &sa, NULL);
-#  endif	/* SIGDANGER */
-# endif		/* HAVE_WORKING_FORK */
+#endif	/* SIGDANGER */
+	} else if (pidfile) {
+		write_pidfile(pidfile, getpid());
 	}
 
-	/* Ignore SIGPIPE - from OpenSSL */
-	sa.sa_handler = SIG_IGN;
-	sigemptyset(&sa.sa_mask);
-	sa.sa_flags = SA_RESTART;
- 	(void)sigaction(SIGPIPE, &sa, NULL);
-
 	/*
 	 * Set up signals we pay attention to locally.
 	 */
@@ -677,12 +631,10 @@ ntpdmain(
 	init_refclock();
 # endif
 	set_process_priority();
-	init_proto(!dumpopts);		/* Call at high priority */
+	init_proto(!dumpopts);	/* Call at high priority */
 	init_io();
 	init_loopfilter();
 	init_readconfig();	/* see readconfig() */
-	mon_start(MON_ON);	/* monitor on by default now	  */
-				/* turn off in config if unwanted */
 
 	/*
 	 * Some option settings have to be deferred until after
@@ -711,6 +663,7 @@ ntpdmain(
 		stats_config(STATS_FREQ_FILE, driftfile);
 		break;
             case 'g':
+            case 'G':
             case 'h':
             case 'i':
                 /* handled elsewhere */
@@ -735,11 +688,7 @@ ntpdmain(
             case 'm':
             case 'n':
             case 'N':
-                /* handled elsewhere */
-                break;
-	    case 'p':
-		stats_config(STATS_PID_FILE, pidfile);
-		break;
+            case 'p':
             case 'P':
             case 'q':
                 /* handled elsewhere */
@@ -806,7 +755,8 @@ ntpdmain(
 	    struct rlimit rlim;
 	    rlim.rlim_max = rlim.rlim_cur = RLIM_INFINITY;
 	    if (setrlimit(RLIMIT_MEMLOCK, &rlim) < 0)
-		msyslog(LOG_WARNING, "INIT: setrlimit() failed: not locking into RAM");
+		msyslog(LOG_DEBUG, "INIT: setrlimit(RLIMIT_CORE, [%ju, %ju]) failed",
+			(uintmax_t)rlim.rlim_cur, (uintmax_t)rlim.rlim_max);
 	    else
 #endif
 
@@ -898,14 +848,17 @@ ntpdmain(
 	readconfig(getconfig(explicit_config));
 	check_minsane();
         if ( 8 > sizeof(time_t) ) {
-	    msyslog(LOG_ERR, "INIT: This system has a 32-bit time_t.");
-	    msyslog(LOG_ERR, "INIT: This ntpd will fail on 2038-01-19T03:14:07Z.");
+	    msyslog(LOG_NOTICE, "INIT: This system has a 32-bit time_t.");
+	    msyslog(LOG_NOTICE, "INIT: This ntpd will fail on 2038-01-19T03:14:07Z.");
         }
 
+	mon_start();
 	loop_config(LOOP_DRIFTINIT, 0);
 	report_event(EVNT_SYSRESTART, NULL, NULL);
 
+#ifndef DISABLE_NTS
 	nts_init();		/* Before droproot */
+#endif
 
 #ifndef ENABLE_EARLY_DROPROOT
 	/* drop root privileges */
@@ -915,7 +868,9 @@ ntpdmain(
 	}
 #endif
 
+#ifndef DISABLE_NTS
 	nts_init2();		/* After droproot */
+#endif
 
 	if (access(statsdir, W_OK) != 0) {
 	    msyslog(LOG_ERR, "statistics directory %s does not exist or is unwriteable, error %s", statsdir, strerror(errno));
@@ -925,6 +880,45 @@ ntpdmain(
         /* unreachable, mainloop() never returns */
 }
 
+
+/* This goes to syslog.
+ * And again to a log file if you are using one.
+ *
+ * The first copy also goes to stderr.
+ * systemd adds that to syslog.
+ *
+ * Switching log files also logs a message before switching.
+ *
+ * If using a log file, there should be enough info in syslog
+ * to debug things with minimal extra clutter.
+ */
+void announce_starting(void) {
+	char buf[1024];	/* Secret knowledge of msyslog buf length */
+	char *cp = buf;
+
+	msyslog(LOG_NOTICE, "INIT: %s: Starting", ntpd_version());
+
+	/* Note that every arg gets an initial space character */
+	snprintf(cp, sizeof(buf), "Command line:");
+	cp += strlen(cp);
+
+	for (int i = 0; i < saved_argc ; ++i) {
+		snprintf(cp, sizeof(buf) - (size_t)(cp - buf),
+			" %s", saved_argv[i]);
+		cp += strlen(cp);
+	}
+	msyslog(LOG_NOTICE, "INIT: %s", buf);
+
+	/* This is helpful if you specify a log file in ntp.conf
+	 * The error messages while parsing ntp.conf go to syslog.
+	 * You might forget to look there while debugging things.
+	 */
+	if (0 < parsing_errors) {
+		msyslog(LOG_ERR, "INIT: saw %d parsing errors", parsing_errors);
+		parsing_errors = 0;
+	}
+}
+
 /*
  * Process incoming packets until exit or interrupted.
  */
@@ -965,10 +959,12 @@ static void mainloop(void)
 			sig_flags.sawHUP = false;
 			msyslog(LOG_INFO, "LOG: Saw SIGHUP");
 
-			reopen_logfile();
-
+			check_logfile();
 			check_leap_file(false, time(NULL));
+#ifndef DISABLE_NTS
 			check_cert_file();
+#endif
+			dns_try_again();
 		}
 
 		/*
@@ -984,7 +980,7 @@ static void mainloop(void)
 				if (!--mdnstries) {
 					msyslog(LOG_ERR, "INIT: Unable to register mDNS, giving up.");
 				} else {
-					msyslog(LOG_INFO, "INIT: Unable to register mDNS, will try later.");
+					msyslog(LOG_NOTICE, "INIT: Unable to register mDNS, will try later.");
 				}
 			} else {
 				msyslog(LOG_INFO, "INIT: mDNS service registered.");
@@ -1052,7 +1048,6 @@ static void catchDNS(int sig)
 /*
  * wait_child_sync_if - implements parent side of -w/--wait-sync
  */
-# ifdef HAVE_WORKING_FORK
 static int
 wait_child_sync_if(
 	int	pipe_read_fd,
@@ -1116,7 +1111,6 @@ wait_child_sync_if(
 		progname, wait_sync1);
 	return ETIMEDOUT;
 }
-# endif	/* HAVE_WORKING_FORK */
 
 
 /*
@@ -1130,7 +1124,7 @@ wait_child_sync_if(
  * With 3 working servers, 2 can outvote a falseticker
  * With 4 servers, you still have 3 if one is down.
  */
-static void check_minsane()
+static void check_minsane(void)
 {
 	struct peer *peer;
 	int servers = 0;
@@ -1214,60 +1208,3 @@ no_debug(
 }
 # endif	/* !DEBUG */
 
-/*
- * close_all_except()
- *
- * Close all file descriptors except the given keep_fd.
- */
-static void
-close_all_except(
-	int keep_fd
-	)
-{
-	int fd;
-
-	for (fd = 0; fd < keep_fd; fd++) {
-		close(fd);
-	}
-
-	close_all_beyond(keep_fd);
-}
-
-
-/*
- * close_all_beyond()
- *
- * Close all file descriptors after the given keep_fd, which is the
- * highest fd to keep open.  See
- *
- * http://stackoverflow.com/questions/899038/getting-the-highest-allocated-file-descriptor
- */
-static void
-close_all_beyond(
-	int keep_fd
-	)
-{
-# ifdef HAVE_CLOSEFROM
-	closefrom(keep_fd + 1);
-# elif defined(F_CLOSEM)
-	/*
-	 * From 'Writing Reliable AIX Daemons,' SG24-4946-00,
-	 * by Eric Agar (saves us from doing 32767 system
-	 * calls)
-	 */
-	if (fcntl(keep_fd + 1, F_CLOSEM, 0) == -1)
-		msyslog(LOG_ERR, "INIT: F_CLOSEM(%d): %s", keep_fd + 1, strerror(errno));
-# else  /* !HAVE_CLOSEFROM && !F_CLOSEM follows */
-	int fd;
-	int max_fd;
-
-	/* includes POSIX case */
-	max_fd = sysconf(_SC_OPEN_MAX);
-	if (10000 < max_fd)
-		msyslog(LOG_ERR, "INIT: close_all_beyond: closing %d files", max_fd);
-	for (fd = keep_fd + 1; fd < max_fd; fd++) {
-		close(fd);
-	}
-# endif /* !HAVE_CLOSEFROM && !F_CLOSEM */
-}
-
