@@ -1,18 +1,20 @@
 /*
  * nts_client.c - Network Time Security (NTS) client side support
- * Copyright 2019 by the NTPsec project contributors
- * SPDX-License-Identifier: BSD-4-Clause-UC
+ * Copyright the NTPsec project contributors
+ * SPDX-License-Identifier: BSD-2-Clause
  *
  * Section references are to
- * https://tools.ietf.org/html/draft-ietf-ntp-using-nts-for-ntp-15
+ * https://tools.ietf.org/html/rfc8915
  *
  */
 #include "config.h"
 
 #include <ctype.h>
 #include <sys/types.h>
+#include <sys/socket.h>
 #include <sys/stat.h>
 #include <unistd.h>
+#include <fcntl.h>
 
 #ifdef HAVE_RES_INIT
 #include <netinet/in.h>
@@ -30,40 +32,35 @@
 #include "nts2.h"
 #include "ntp_dns.h"
 #include "ntp_stdlib.h"
+#include "timespecops.h"
 
 SSL_CTX* make_ssl_client_ctx(const char *filename);
 int open_TCP_socket(struct peer *peer, const char *hostname);
+struct addrinfo * find_best_addr(struct addrinfo *answer);
+bool connect_TCP_socket(int sockfd, struct addrinfo *addr);
 bool nts_set_cert_search(SSL_CTX *ctx, const char *filename);
 void set_hostname(SSL *ssl, const char *hostname);
 bool check_certificate(SSL *ssl, struct peer *peer);
-bool check_aead(SSL *ssl, struct peer *peer, const char *hostname);
+bool check_alpn(SSL *ssl, struct peer *peer, const char *hostname);
 bool nts_client_send_request(SSL *ssl, struct peer *peer);
+bool nts_client_send_request_core(uint8_t *buff, int buf_size, int *used, struct peer* peer);
 bool nts_client_process_response(SSL *ssl, struct peer *peer);
 bool nts_client_process_response_core(uint8_t *buff, int transferred, struct peer* peer);
-bool nts_client_send_request_core(uint8_t *buff, int buf_size, int *used, struct peer* peer);
 bool nts_server_lookup(char *server, sockaddr_u *addr, int af);
 
 static SSL_CTX *client_ctx = NULL;
+
+/* Ugly global variables passed from worker thread back to main thread. */
 static sockaddr_u sockaddr;
 static bool addrOK;
 
-// Fedora 30:  0x1010104fL  1.1.1d
-// Fedora 29:  0x1010102fL  1.1.1b
-// Fedora 28:  0x1010009fL  1.1.0i
-// Debian 10:  0x1010104fL  1.1.1d
-// Debian 9:   0x101000afL  1.1.0j
-// Debian 8:   0x1000114fL  1.0.1t
-// CentOS 7:   0x100020bfL  1.0.2k
-// CentOS 6:   0x1000105fL  1.0.1e
-// NetBSD 8:   0x100020bfL  1.0.2k
-// NetBSD 7:   0x1000115fL  1.0.1u (1.0.2s via pkgin)
-// FreeBSD 12: 0x1010101fL  1.1.1a-freebsd
-// FreeBSD 11: 0x100020ffL  1.0.2o-freebsd
 
 bool nts_client_init(void) {
 
 	client_ctx = make_ssl_client_ctx(ntsconfig.ca);
 
+
+/* Ugly global variables passed from worker thread back to main thread. */
 	return true;
 }
 
@@ -74,14 +71,14 @@ bool nts_probe(struct peer * peer) {
 	char errbuf[100];
 	SSL     *ssl;
 	int      server;
-	l_fp     start, finish;
+	struct timespec start, finish;
 	int      err;
 
 	if (NULL == client_ctx)
 		return false;
 
 	addrOK = false;
-	get_systime(&start);
+	clock_gettime(CLOCK_MONOTONIC, &start);
 
 	if (NULL == hostname) {
 		/* IP Address case */
@@ -101,7 +98,7 @@ bool nts_probe(struct peer * peer) {
 
 	server = open_TCP_socket(peer, hostname);
 	if (-1 == server) {
-		nts_ke_probes_bad++;
+		ntske_cnt.probes_bad++;
 		return false;
 	}
 
@@ -110,14 +107,14 @@ bool nts_probe(struct peer * peer) {
 		ntp_strerror_r(errno, errbuf, sizeof(errbuf));
 		msyslog(LOG_ERR, "NTSc: can't setsockopt: %s", errbuf);
 		close(server);
-		nts_ke_probes_bad++;
+		ntske_cnt.probes_bad++;
 		return false;
 	}
 
 	if (NULL == peer->cfg.nts_cfg.ca)
 		ssl = SSL_new(client_ctx);
 	else {
-		SSL_CTX *ctx; 
+		SSL_CTX *ctx;
 		ctx = make_ssl_client_ctx(peer->cfg.nts_cfg.ca);
 		if (NULL == ctx) {
 			close(server);
@@ -148,7 +145,7 @@ bool nts_probe(struct peer * peer) {
 
 	if (!check_certificate(ssl, peer))
 		goto bail;
-	if (!check_aead(ssl, peer, hostname))
+	if (!check_alpn(ssl, peer, hostname))
 		goto bail;
 
 	if (!nts_client_send_request(ssl, peer))
@@ -171,21 +168,21 @@ bool nts_probe(struct peer * peer) {
 		goto bail;
 
 	addrOK = true;
-	nts_ke_probes_good++;
+	ntske_cnt.probes_good++;
 
   bail:
 	if (!addrOK) {
-		nts_ke_probes_bad++;
+		ntske_cnt.probes_bad++;
 		peer->nts_state.count = -1;
 	}
 	SSL_shutdown(ssl);
 	SSL_free(ssl);
 	close(server);
 
-	get_systime(&finish);
-	finish -= start;
-	msyslog(LOG_INFO, "NTSc: NTS-KE req to %s took %.3Lf sec, %s",
-		hostname, lfptod(finish),
+	clock_gettime(CLOCK_MONOTONIC, &finish);
+	finish = sub_tspec(finish, start);
+	msyslog(LOG_INFO, "NTSc: NTS-KE req to %s took %.3f sec, %s",
+		hostname, tspec_to_d(finish),
 		addrOK? "OK" : "fail");
 
 	return addrOK;
@@ -209,12 +206,7 @@ SSL_CTX* make_ssl_client_ctx(const char * filename) {
 	bool ok = true;
 	SSL_CTX *ctx;
 
-#if (OPENSSL_VERSION_NUMBER > 0x1010000fL)
 	ctx = SSL_CTX_new(TLS_client_method());
-#else
-	// OpenSSL_add_all_ciphers();  // Maybe was needed on NetBSD ??
-	ctx = SSL_CTX_new(TLSv1_2_client_method());
-#endif
 	if (NULL == ctx) {
 		/* Happens if no ciphers */
 		msyslog(LOG_ERR, "NTSc: NULL ctx");
@@ -222,19 +214,18 @@ SSL_CTX* make_ssl_client_ctx(const char * filename) {
 		return NULL;
 	}
 
-#if (OPENSSL_VERSION_NUMBER > 0x1000200fL)
 	{
 		// 4., ALPN, RFC 7301
 		static unsigned char alpn [] = { 7, 'n', 't', 's', 'k', 'e', '/', '1' };
 		SSL_CTX_set_alpn_protos(ctx, alpn, sizeof(alpn));
 	}
-#endif
 
 	SSL_CTX_set_session_cache_mode(ctx, SSL_SESS_CACHE_OFF);
-	SSL_CTX_set_timeout(ctx, NTS_KE_TIMEOUT);
+	SSL_CTX_set_timeout(ctx, NTS_KE_TIMEOUT);   /* session lifetime */
 
 	ok &= nts_load_versions(ctx);
 	ok &= nts_load_ciphers(ctx);
+	ok &= nts_load_ecdhcurves(ctx);
 	ok &= nts_set_cert_search(ctx, filename);
 
 	if (!ok) {
@@ -252,10 +243,10 @@ int open_TCP_socket(struct peer *peer, const char *hostname) {
 	char errbuf[100];
 	char *tmp;
 	struct addrinfo hints;
-	struct addrinfo *answer;
-	int gai_rc, err;
+	struct addrinfo *answer, *worker;
+	int gai_rc;
 	int sockfd;
-	l_fp start, finish;
+	struct timespec start, finish;
 
 	/* copy avoids dancing around const warnings */
 	strlcpy(host, hostname, sizeof(host));
@@ -270,7 +261,7 @@ int open_TCP_socket(struct peer *peer, const char *hostname) {
 	}
 	if (NULL == tmp) {
 		/* simple case, no : */
-		strlcpy(port, "123", sizeof(port));
+		strlcpy(port, NTS_KE_PORTA, sizeof(port));
 	} else {
 		/* Complicated case, found a : */
 		*tmp++ = 0;
@@ -281,36 +272,39 @@ int open_TCP_socket(struct peer *peer, const char *hostname) {
 	hints.ai_protocol = IPPROTO_TCP;
 	hints.ai_socktype = SOCK_STREAM;
 	hints.ai_family = AF(&peer->srcadr);  /* -4, -6 switch */
-	get_systime(&start);
+	clock_gettime(CLOCK_MONOTONIC, &start);
 	gai_rc = getaddrinfo(host, port, &hints, &answer);
 	if (0 != gai_rc) {
-		msyslog(LOG_INFO, "NTSc: nts_probe: DNS error trying to contact %s: %d, %s",
+		msyslog(LOG_INFO, "NTSc: open_TCP_socket: DNS error trying to contact %s: %d, %s",
 			hostname, gai_rc, gai_strerror(gai_rc));
 		return -1;
 	}
-	get_systime(&finish);
-	finish -= start;
-	msyslog(LOG_INFO, "NTSc: DNS lookup of %s took %.3Lf sec",
-		hostname, lfptod(finish));
-
-	/* Save first answer for NTP, switch to NTP port in case of server-name:port */
-	memcpy(&sockaddr, answer->ai_addr, answer->ai_addrlen);
-	SET_PORT(&sockaddr, NTP_PORT);
+	clock_gettime(CLOCK_MONOTONIC, &finish);
+	finish = sub_tspec(finish, start);
+	msyslog(LOG_INFO, "NTSc: DNS lookup of %s took %.3f sec",
+		hostname, tspec_to_d(finish));
 
+	/* Use first answer
+	 * sockaddr is global for NTP address
+	 * also use as temp for printing here
+	 */
+	worker = find_best_addr(answer);
+	memcpy(&sockaddr, worker->ai_addr, worker->ai_addrlen);
 	sockporttoa_r(&sockaddr, errbuf, sizeof(errbuf));
-	msyslog(LOG_INFO, "NTSc: nts_probe connecting to %s:%s => %s",
+	msyslog(LOG_INFO, "NTSc: connecting to %s:%s => %s",
 		host, port, errbuf);
 
-	sockfd = socket(answer->ai_family, SOCK_STREAM, 0);
+	/* setup default NTP port now
+	 *   in case of server-name:port later on
+	 */
+	SET_PORT(&sockaddr, NTP_PORT);
+	sockfd = socket(worker->ai_family, SOCK_STREAM, 0);
 	if (-1 == sockfd) {
 		ntp_strerror_r(errno, errbuf, sizeof(errbuf));
-		msyslog(LOG_INFO, "NTSc: nts_probe: no socket: %s", errbuf);
+		msyslog(LOG_INFO, "NTSc: open_TCP_socket: no socket: %s", errbuf);
 	} else {
-		// Use first answer
-		err = connect(sockfd, answer->ai_addr, answer->ai_addrlen);
-		if (-1 == err) {
-			ntp_strerror_r(errno, errbuf, sizeof(errbuf));
-			msyslog(LOG_INFO, "NTSc: nts_probe: connect failed: %s", errbuf);
+		/* Use first IP Address */
+		if (!connect_TCP_socket(sockfd, worker)) {
 			close(sockfd);
 			sockfd = -1;
 		}
@@ -318,8 +312,81 @@ int open_TCP_socket(struct peer *peer, const char *hostname) {
 
 	freeaddrinfo(answer);
 	return sockfd;
+
 }
 
+struct addrinfo *find_best_addr(struct addrinfo *answer) {
+	/* default to first one */
+	return(answer);
+}
+
+
+/* This kludgery is needed to get a sane timeout.
+ * The default is unspecified but long.
+ * On Linux, man connect gets man 2 which doesn't mention O_NONBLOCK
+ * Use man 3 connect.
+ */
+bool connect_TCP_socket(int sockfd, struct addrinfo *addr) {
+	char errbuf[100];
+	int err;
+	fd_set fdset;
+	struct timeval timeout;
+	int so_error;
+	socklen_t so_len = sizeof(so_error);
+
+	err = fcntl(sockfd, F_SETFL, O_NONBLOCK);
+	if (-1 == err) {
+		ntp_strerror_r(errno, errbuf, sizeof(errbuf));
+		msyslog(LOG_INFO, "NTSc: can't set O_NONBLOCK %s", errbuf);
+		return false;
+	}
+	err = connect(sockfd, addr->ai_addr, addr->ai_addrlen);
+	/* The normal case is -1 and errno == EINPROGRESS
+	 * Getting connected should be possible if the scheduler
+	 * avoids us for long enough.
+	 * Other errors may be possible.  No route?
+	 * I haven't seen that yet.  HGM, 2020 Jan 19
+	 */
+	if (-1 != err || EINPROGRESS != errno) {
+		ntp_strerror_r(errno, errbuf, sizeof(errbuf));
+		msyslog(LOG_INFO, "NTSc: connect_TCP_socket: connect failed: %s", errbuf);
+		return false;
+	}
+
+	FD_ZERO(&fdset);
+	FD_SET(sockfd, &fdset);
+	timeout.tv_sec = NTS_KE_TIMEOUT;
+	timeout.tv_usec = 0;
+
+	if (0 == select(sockfd + 1, NULL, &fdset, NULL, &timeout)) {
+		msyslog(LOG_INFO, "NTSc: connect_TCP_socket: timeout");
+		return false;
+	}
+
+	/* It's ready, either connected or error. */
+	if (-1 == getsockopt(sockfd, SOL_SOCKET, SO_ERROR, &so_error, &so_len)) {
+		ntp_strerror_r(errno, errbuf, sizeof(errbuf));
+		msyslog(LOG_INFO, "NTSc: connect_TCP_socket: getsockopt failed: %s", errbuf);
+		return false;
+	}
+
+	if (0 != so_error) {
+		ntp_strerror_r(so_error, errbuf, sizeof(errbuf));
+		msyslog(LOG_INFO, "NTSc: connect_TCP_socket: connect failed: %s", errbuf);
+		return false;
+	}
+
+	err = fcntl(sockfd, F_SETFL, 0); /* turn off O_NONBLOCK */
+	if (-1 == err) {
+		ntp_strerror_r(errno, errbuf, sizeof(errbuf));
+		msyslog(LOG_INFO, "NTSc: can't unset O_NONBLOCK %s", errbuf);
+		return false;
+	}
+
+	return true;
+}
+
+
 void set_hostname(SSL *ssl, const char *hostname) {
 	char host[256], *tmp;
 
@@ -334,113 +401,123 @@ void set_hostname(SSL *ssl, const char *hostname) {
 		*tmp = 0;
 	}
 
-// https://wiki.openssl.org/index.php/Hostname_validation
-#if (OPENSSL_VERSION_NUMBER > 0x1010000fL)
-	SSL_set_hostflags(ssl, X509_CHECK_FLAG_NO_WILDCARDS);
+/* https://wiki.openssl.org/index.php/Hostname_validation
+ * draft-ietf-uta-rfc6125bis section 3 relaxes the restrictions around the use
+ * of wildcards to make it clear that they're permitted unless specifically
+ * prohibited in an RFC
+ */
+	SSL_set_hostflags(ssl, X509_CHECK_FLAG_NO_PARTIAL_WILDCARDS);
 	SSL_set1_host(ssl, host);
 	msyslog(LOG_DEBUG, "NTSc: set cert host: %s", host);
-#elif (OPENSSL_VERSION_NUMBER > 0x1000200fL)
-	{  /* enable automatic hostname checks */
-	X509_VERIFY_PARAM *param = SSL_get0_param(ssl);
-	X509_VERIFY_PARAM_set_hostflags(param, X509_CHECK_FLAG_NO_WILDCARDS);
-	if (1 != X509_VERIFY_PARAM_set1_host(param, host, strlen(host))) {
-		msyslog(LOG_ERR, "NTSc: troubles setting hostflags");
-	}
-	SSL_set_verify(ssl, SSL_VERIFY_PEER, NULL);
-	}
-#else
-	/*  Versions prior to 1.0.2 did not perform hostname validation */
-	UNUSED_ARG(ssl);
-	msyslog(LOG_ERR, "NTSc: can't check hostname/certificate");
-#endif
 
 }
 
 bool check_certificate(SSL *ssl, struct peer* peer) {
 	X509 *cert = SSL_get_peer_certificate(ssl);
+	X509_NAME *certname;
+	GENERAL_NAMES *gens;
+	char name[200];
+	int certok;
+	int numgens = 0;
 
 	if (NULL == cert) {
 		msyslog(LOG_INFO, "NTSc: No certificate");
 		if (!(FLAG_NTS_NOVAL & peer->cfg.flags))
 			return false;
 		return true;
+	}
+
+	certname = X509_get_subject_name(cert);
+	X509_NAME_oneline(certname, name, sizeof(name));
+	msyslog(LOG_INFO, "NTSc: certificate subject name: %s", name);
+	certname = X509_get_issuer_name(cert);
+	X509_NAME_oneline(certname, name, sizeof(name));
+	msyslog(LOG_INFO, "NTSc: certificate issuer name: %s", name);
+	/* print SAN:DNS strings */
+	gens = X509_get_ext_d2i(cert, NID_subject_alt_name, 0, 0);
+	if (gens) {
+	  char buff[150];
+	  numgens = sk_GENERAL_NAME_num(gens);
+	  buff[0] = 0;
+	  for (int i = 0; i<numgens; i++) {
+	    const GENERAL_NAME *gen;
+	    const char *dnsname;
+	    unsigned int len;
+	    gen = sk_GENERAL_NAME_value(gens, i);
+	    if (gen->type != GEN_DNS)
+	      continue;
+	    // string is NUL terminated but may have internal NULs
+	    len = (unsigned int)ASN1_STRING_length(gen->d.ia5);
+	    dnsname = (const char *)ASN1_STRING_get0_data(gen->d.ia5);
+	    if (0 != buff[0])
+	      strlcat(buff, ", ", sizeof(buff));
+	    strlcat(buff, dnsname, sizeof(buff));
+	    if (len != strlen(dnsname))
+	      strlcat(buff, "??", sizeof(buff));
+	  }
+	  msyslog(LOG_INFO, "NTSc: SAN:DNS %s", buff);
+	  GENERAL_NAMES_free(gens);
+	}
+	if (0 == numgens) {
+	  const char *peername = SSL_get0_peername(ssl);
+	  msyslog(LOG_INFO, "NTSc: matching with subject:CN %s", peername);
+	} else if (1 > numgens) {
+	  const char *peername = SSL_get0_peername(ssl);
+	  msyslog(LOG_INFO, "NTSc: matching with SAN:DNS: %s", peername);
+	}
+	X509_free(cert);
+	certok = SSL_get_verify_result(ssl);
+	if (X509_V_OK == certok) {
+		msyslog(LOG_INFO, "NTSc: certificate is valid.");
 	} else {
-		X509_NAME *certname;
-		char name[200];
-		int certok;
-		certname = X509_get_subject_name(cert);
-		X509_NAME_oneline(certname, name, sizeof(name));
-		msyslog(LOG_INFO, "NTSc: certificate subject name: %s", name);
-		certname = X509_get_issuer_name(cert);
-		X509_NAME_oneline(certname, name, sizeof(name));
-		msyslog(LOG_INFO, "NTSc: certificate issuer name: %s", name);
-		X509_free(cert);
-		certok = SSL_get_verify_result(ssl);
-		if (X509_V_OK == certok) {
-			msyslog(LOG_INFO, "NTSc: certificate is valid.");
-		} else {
-			msyslog(LOG_ERR, "NTSc: certificate invalid: %d=>%s",
-				certok, X509_verify_cert_error_string(certok));
-			if (!(FLAG_NTS_NOVAL & peer->cfg.flags))
-				return false;
+		msyslog(LOG_ERR, "NTSc: certificate invalid: %d=>%s",
+			certok, X509_verify_cert_error_string(certok));
+		if (FLAG_NTS_NOVAL & peer->cfg.flags) {
+			msyslog(LOG_INFO, "NTSc: noval - accepting invalid cert.");
+			return true;
 		}
+		return false;
 	}
 	return true;
 }
 
-bool check_aead(SSL *ssl, struct peer* peer, const char *hostname) {
+bool check_alpn(SSL *ssl, struct peer* peer, const char *hostname) {
 	UNUSED_ARG(peer);
-#if (OPENSSL_VERSION_NUMBER > 0x1000200fL)
-	bool bad = true;  /* Always return OK for now. */
 	const unsigned char *data;
 	unsigned int len;
-	unsigned int i;
-	char buff [100];
 	SSL_get0_alpn_selected(ssl, &data, &len);
 	if (0 == len) {
-		/* This happens when talking to old/TLSv1.2 systems. */
 		msyslog(LOG_DEBUG, "NTSc: No ALPN from %s (%s)",
 			hostname, SSL_get_version(ssl));
-		return bad;
-	}
-	if (sizeof(buff) <= len) {
-		/* Broken or malicious server */
-		msyslog(LOG_DEBUG, "NTSc: Very long ALPN from %s (%u)",
-			hostname, len);
-		return bad;
-	}
-	memcpy(buff, data, len);
-	buff[len] = '\0';
-	for (i=0; i<len; i++) {
-		if (!isgraph(buff[i])) {
-			buff[i] = '*'; /* fix non-printing crap */
-		}
+		return false;
 	}
 	/* For now, we only support one version.
-	 * This gets more complicated when version 2 arrives. */
-	if (0 != strcmp((const char*)data, "ntske/1")) {
-		msyslog(LOG_DEBUG, "NTSc: Strange ALPN returned: %s (%u) from %s",
+	 * This will get more complicated when version 2 arrives. */
+	if (len != 7 ||
+	    0 != memcmp(data, "ntske/1", len)) {
+		/* copy data over so we can print it. */
+		/* don't read past end of data */
+		unsigned int i, l;
+		char buff [16];
+		l = min(len, sizeof(buff)-1);
+		memcpy(buff, data, l);
+		buff[l] = '\0';
+		for (i=0; i<l; i++) {
+			if (!isgraph((int)buff[i])) {
+			  buff[i] = '*'; /* fix non-printing crap */
+			}
+		}
+		msyslog(LOG_DEBUG, "NTSc: Strange ALPN %s (%u) from %s",
 			buff, len, hostname);
-		return bad;
+		return false;
 	}
-        msyslog(LOG_DEBUG, "NTSc: Good ALPN from: %s", hostname);
+	msyslog(LOG_DEBUG, "NTSc: Good ALPN from %s", hostname);
 
-#else
-	UNUSED_ARG(ssl);
-	UNUSED_ARG(hostname);
-#endif
 	return true;
 }
 
 bool nts_make_keys(SSL *ssl, uint16_t aead, uint8_t *c2s, uint8_t *s2c, int keylen) {
-	// There is a bug in OpenSSL 1.1.1a
-	// Until Mar-23, we were using:
-	//    const char *label = "EXPORTER-nts/1";
-	// Subject: [Ntp] [NTS4NTP] info for NTS developers
-	// From: Martin Langer <mart.langer@ostfalia.de>
-	// Date: Tue, 15 Jan 2019 11:40:13 +0100
-	// https://mailarchive.ietf.org/arch/msg/ntp/nkc-9n6XOPt5Glgi_ueLvuD9EfY
-	const char *label = "EXPORTER-network-time-security/1";
+	const char *label = "EXPORTER-network-time-security";
 	unsigned char context[5];
 	context[0] = (nts_protocol_NTP >> 8) & 0xFF;
 	context[1] = nts_protocol_NTP & 0xFF;
@@ -483,7 +560,6 @@ bool nts_client_send_request(SSL *ssl, struct peer* peer) {
 }
 
 bool nts_client_send_request_core(uint8_t *buff, int buf_size, int *used, struct peer* peer) {
-	char errbuf[100];
 	struct  BufCtl_t buf;
 	uint16_t aead = NO_AEAD;
 
@@ -510,10 +586,9 @@ bool nts_client_send_request_core(uint8_t *buff, int buf_size, int *used, struct
 
 	*used = buf_size-buf.left;
 	if (*used >= (int)(buf_size - 10)) {
-		ntp_strerror_r(errno, errbuf, sizeof(errbuf));
-		msyslog(LOG_ERR, "NTSc: write failed: %d, %ld, %s",
-			*used, (long)buf_size, errbuf);
-		return false;
+		msyslog(LOG_ERR, "ERR-NTSc: buffer overflow: %d, %ld",
+			*used, (long)buf_size);
+		exit(2);
 	}
 	return true;
 }
@@ -542,7 +617,7 @@ bool nts_client_process_response_core(uint8_t *buff, int transferred, struct pee
 
 	buf.next = buff;
 	buf.left = transferred;
-	while (buf.left > 0) {
+	while (buf.left >= NTS_KE_HDR_LNG) {
 		uint16_t type, data, port;
 		bool critical = false;
 		int length, keylength;
@@ -551,6 +626,10 @@ bool nts_client_process_response_core(uint8_t *buff, int transferred, struct pee
 		char server[MAX_SERVER];
 
 		type = ke_next_record(&buf, &length);
+		if(buf.left < length){
+			msyslog(LOG_ERR, "NTSc: length cannot be more than buf.left: %d", length);
+			return false;
+		}
 		if (NTS_CRITICAL & type) {
 			critical = true;
 			type &= ~NTS_CRITICAL;
@@ -559,25 +638,30 @@ bool nts_client_process_response_core(uint8_t *buff, int transferred, struct pee
 			msyslog(LOG_ERR, "NTSc: Record: T=%d, L=%d, C=%d", type, length, critical);
 		switch (type) {
 		    case nts_error:
-			data = next_uint16(&buf);
-			if (sizeof(data) != length)
+			if (sizeof(data) != length) {
 				msyslog(LOG_ERR, "NTSc: wrong length on error: %d", length);
+				return false;
+			}
+			data = next_uint16(&buf);
 			msyslog(LOG_ERR, "NTSc: error: %d", data);
 			return false;
 		    case nts_next_protocol_negotiation:
+			if (sizeof(data) != length) {
+				msyslog(LOG_ERR, "NTSc: NPN-Wrong length: %d", length);
+				return false;
+			}
 			data = next_uint16(&buf);
-			if ((sizeof(data) != length) || (data != nts_protocol_NTP)) {
-				msyslog(LOG_ERR, "NTSc: NPN-Wrong length or bad data: %d, %d",
-					length, data);
+			if (data != nts_protocol_NTP) {
+				msyslog(LOG_ERR, "NTSc: NPN-Bad data: %d", data);
 				return false;
 			}
 			break;
 		    case nts_algorithm_negotiation:
-			data = next_uint16(&buf);
 			if (sizeof(data) != length) {
 				msyslog(LOG_ERR, "NTSc: AN-Wrong length: %d", length);
 				return false;
 			}
+			data = next_uint16(&buf);
 			keylength = nts_get_key_length(data);
 			if (0 == keylength) {
 				msyslog(LOG_ERR, "NTSc: AN-Unsupported AEAN type: %d", data);
@@ -655,6 +739,9 @@ bool nts_client_process_response_core(uint8_t *buff, int transferred, struct pee
 		} /* case */
 	}   /* while */
 
+	if (buf.left > 0)
+		return false;
+
 	if (NO_AEAD == peer->nts_state.aead) {
 		msyslog(LOG_ERR, "NTSc: No AEAD algorithim.");
 		return false;
@@ -679,13 +766,23 @@ bool nts_set_cert_search(SSL_CTX *ctx, const char *filename) {
 	}
 	if (0 == stat(filename, &statbuf)) {
 		if (S_ISDIR(statbuf.st_mode)) {
+			if (1 != SSL_CTX_load_verify_locations(
+				ctx, NULL, filename)) {
+			msyslog(LOG_INFO, "NTSc: Can't use %s as dir for root certificates.", filename);
+			    nts_log_ssl_error();
+			    return false;
+			}
 			msyslog(LOG_INFO, "NTSc: Using dir %s for root certificates.", filename);
-			SSL_CTX_load_verify_locations(ctx, NULL, filename);
 			return true;
 		}
 		if (S_ISREG(statbuf.st_mode)) {
+			if (1 != SSL_CTX_load_verify_locations(
+				ctx, filename, NULL)) {
+			    msyslog(LOG_INFO, "NTSc: Can't use %s as file for root certificates.", filename);
+			    nts_log_ssl_error();
+			    return false;
+			}
 			msyslog(LOG_INFO, "NTSc: Using file %s for root certificates.", filename);
-			SSL_CTX_load_verify_locations(ctx, filename, NULL);
 			return true;
 		}
 		msyslog(LOG_ERR, "NTSc: cert dir/file isn't dir or file: %s. mode 0x%x",
@@ -704,7 +801,7 @@ bool nts_set_cert_search(SSL_CTX *ctx, const char *filename) {
  */
 bool nts_server_lookup(char *server, sockaddr_u *addr, int af) {
 	struct addrinfo hints;
-	struct addrinfo *answer;
+	struct addrinfo *answer = NULL;  /* init to keep oss-fuzz happy */
 	int gai_rc;
 
 	ZERO(hints);
@@ -712,7 +809,7 @@ bool nts_server_lookup(char *server, sockaddr_u *addr, int af) {
 	hints.ai_socktype = SOCK_DGRAM;
 	hints.ai_family = af;
 
-	gai_rc = getaddrinfo(server, "123", &hints, &answer);
+	gai_rc = getaddrinfo(server, NTS_KE_PORTA, &hints, &answer);
 	if (0 != gai_rc) {
 		msyslog(LOG_INFO, "NTSc: DNS error trying to lookup %s: %d, %s",
 			server, gai_rc, gai_strerror(gai_rc));
