@@ -1,10 +1,10 @@
 /*
  * nts_server.c - Network Time Security (NTS) server side support
- * Copyright 2019 by the NTPsec project contributors
- * SPDX-License-Identifier: BSD-4-Clause-UC
+ * Copyright the NTPsec project contributors
+ * SPDX-License-Identifier: BSD-2-Clause
  *
  * Section references are to
- * https://tools.ietf.org/html/draft-ietf-ntp-using-nts-for-ntp-15
+ * https://tools.ietf.org/html/rfc8915
  *
  */
 #include "config.h"
@@ -13,8 +13,11 @@
 #include <pthread.h>
 #include <unistd.h>
 #include <sys/socket.h>
+#include <sys/time.h>
+#include <sys/resource.h>
 
 #include <openssl/ssl.h>
+#include <openssl/err.h>
 #include <openssl/x509.h>
 
 #include "ntp.h"
@@ -22,15 +25,18 @@
 #include "ntp_stdlib.h"
 #include "nts.h"
 #include "nts2.h"
+#include "timespecops.h"
 
 /* Beware: bind and accept take type sockaddr, but that's not big
  *         enough for an IPv6 address.
  */
 
+
 static bool create_listener4(int port);
 static bool create_listener6(int port);
 static void* nts_ke_listener(void*);
 static bool nts_ke_request(SSL *ssl);
+static void nts_ke_accept_fail(char* addrbuf, double sec);
 
 static void nts_lock_certlock(void);
 static void nts_unlock_certlock(void);
@@ -44,13 +50,6 @@ static int listener6_sock = -1;
  * This seems like overkill, but it doesn't happen often. */
 pthread_mutex_t certificate_lock = PTHREAD_MUTEX_INITIALIZER;
 
-/* Statistics for ntpq */
-uint64_t nts_ke_serves_good = 0;
-uint64_t nts_ke_serves_bad = 0;
-uint64_t nts_ke_probes_good = 0;
-uint64_t nts_ke_probes_bad = 0;
-
-#if (OPENSSL_VERSION_NUMBER > 0x1000200fL)
 static int alpn_select_cb(SSL *ssl,
 			  const unsigned char **out,
 			  unsigned char *outlen,
@@ -84,7 +83,6 @@ static int alpn_select_cb(SSL *ssl,
 
 	return SSL_TLSEXT_ERR_NOACK;
 }
-#endif
 
 bool nts_server_init(void) {
 	bool ok = true;
@@ -92,12 +90,7 @@ bool nts_server_init(void) {
 	msyslog(LOG_INFO, "NTSs: starting NTS-KE server listening on port %d",
 		NTS_KE_PORT);
 
-#if (OPENSSL_VERSION_NUMBER > 0x1010000fL)
 	server_ctx = SSL_CTX_new(TLS_server_method());
-#else
-	// OpenSSL_add_all_ciphers();  // maybe was needed on NetBSD ??
-	server_ctx = SSL_CTX_new(TLSv1_2_server_method());
-#endif
 	if (NULL == server_ctx) {
 		/* Happens if no ciphers */
 		msyslog(LOG_INFO, "NTSs: NULL server_ctx");
@@ -105,15 +98,13 @@ bool nts_server_init(void) {
 		return false;
 	}
 
-#if (OPENSSL_VERSION_NUMBER > 0x1000200fL)
 	SSL_CTX_set_alpn_select_cb(server_ctx, alpn_select_cb, NULL);
-#endif
-
 	SSL_CTX_set_session_cache_mode(server_ctx, SSL_SESS_CACHE_OFF);
+	SSL_CTX_set_timeout(server_ctx, NTS_KE_TIMEOUT);  /* session lifetime */
 
 	ok &= nts_load_versions(server_ctx);
 	ok &= nts_load_ciphers(server_ctx);
-	ok &= nts_load_certificate(server_ctx);
+	ok &= nts_load_ecdhcurves(server_ctx);
 
 	if (!ok) {
 		msyslog(LOG_ERR, "NTSs: Disabling NTS-KE server");
@@ -122,11 +113,11 @@ bool nts_server_init(void) {
 		return false;
 	};
 
-#if (OPENSSL_VERSION_NUMBER > 0x1010000fL)
 	msyslog(LOG_INFO, "NTSs: OpenSSL security level is %d",
 		SSL_CTX_get_security_level(server_ctx));
-#endif
 
+	msyslog(LOG_INFO, "NTSs: starting NTS-KE server listening on port %d",
+		NTS_KE_PORT);
 	ok &= create_listener4(NTS_KE_PORT);
 	ok &= create_listener6(NTS_KE_PORT);
 
@@ -139,6 +130,10 @@ bool nts_server_init2(void) {
 	int rc;
 	char errbuf[100];
 
+	if (!nts_load_certificate(server_ctx)) {
+		return false;
+	}
+
 	sigfillset(&block_mask);
 	pthread_sigmask(SIG_BLOCK, &block_mask, &saved_sig_mask);
 	if (listener4_sock != -1) {
@@ -160,8 +155,15 @@ bool nts_server_init2(void) {
 	return true;
 }
 
-void check_cert_file(void) {
+/* called every hour */
+void nts_cert_timer(void) {
+	check_cert_file();
+}
 
+/* call hourly and by SIGHUP */
+void check_cert_file(void) {
+	if (NULL == server_ctx)
+		return;
 	nts_lock_certlock();
 	nts_reload_certificate(server_ctx);
 	nts_unlock_certlock();
@@ -183,16 +185,48 @@ void nts_unlock_certlock(void) {
 	}
 }
 
+/* lfptod goes to long double */
+static inline double lfptox(l_fp r) {
+/* l_fp to double */
+        return ldexp((double)((int64_t)r), -32);
+}
+
+
 void* nts_ke_listener(void* arg) {
 	struct timeval timeout = {.tv_sec = NTS_KE_TIMEOUT, .tv_usec = 0};
 	int sock = *(int*)arg;
 	char errbuf[100];
+	char addrbuf[100];
+	char usingbuf[100];
+	struct timespec start, finish;		/* wall clock */
+	l_fp wall;
+	bool worked;
+	const char *good;
+#ifdef RUSAGE_THREAD
+	struct timespec start_u, finish_u;	/* CPU user */
+	struct timespec start_s, finish_s;	/* CPU system */
+	l_fp usr, sys;
+	struct rusage usage;
+#endif
+
+#ifdef HAVE_SECCOMP_H
+        setup_SIGSYS_trap();   /* enable trap for this thread */
+#endif
+
+#ifdef RUSAGE_THREAD
+	/* NB: start_u and start_s are from near the end of the previous cycle.
+	 * Thus usage timing includes the TCP accept and
+	 * writing the previous msyslog message.
+	 */
+	getrusage(RUSAGE_THREAD, &usage);
+	start_u = tval_to_tspec(usage.ru_utime);
+	start_s = tval_to_tspec(usage.ru_stime);
+#endif
 
 	while(1) {
 		sockaddr_u addr;
 		socklen_t len = sizeof(addr);
 		SSL *ssl;
-		l_fp start, finish;
 		int client, err;
 
 		client = accept(sock, &addr.sa, &len);
@@ -204,16 +238,33 @@ void* nts_ke_listener(void* arg) {
 			sleep(1);		/* avoid log clutter on bug */
 			continue;
 		}
-		get_systime(&start);
+		clock_gettime(CLOCK_MONOTONIC, &start);
+		sockporttoa_r(&addr, addrbuf, sizeof(addrbuf));
 
-		sockporttoa_r(&addr, errbuf, sizeof(errbuf));
-		msyslog(LOG_INFO, "NTSs: TCP accept-ed from %s", errbuf);
-		err = setsockopt(client, SOL_SOCKET, SO_RCVTIMEO, &timeout, sizeof(timeout));
+/* This is disabled in order to reduce clutter in the log file.
+ * The client's address is now included in the final message.
+ * That works fine in the normal successful case.  There is one line
+ * per connection.
+ * The failed cases are more complicated.
+ * The common fail case is bad guys probing which fails in SSL_accept.
+ * That branch has its own error handling.  Again, the common cases
+ * have one line per connection and include the client address.
+ * Uncommon cases will include two (or more) lines.
+ * There are many possible error messages after SSL_accept works.
+ * In practice, they don't happen, at least not often enough to notice.
+ * They currently get logged without the client's address.  Then they
+ * fall into the normal (non-error) path which does include the address.
+ * Enabling this might make strange cases easier to understand.
+ */
+/*		msyslog(LOG_INFO, "NTSs: TCP accept-ed from %s", addrbuf); */
+
+		err = setsockopt(client, SOL_SOCKET, SO_RCVTIMEO,
+			&timeout, sizeof(timeout));
 		if (0 > err) {
 			ntp_strerror_r(errno, errbuf, sizeof(errbuf));
 			msyslog(LOG_ERR, "NTSs: can't setsockopt: %s", errbuf);
 			close(client);
-			nts_ke_serves_bad++;
+			ntske_cnt.serves_bad++;
 			continue;
 		}
 
@@ -224,39 +275,106 @@ void* nts_ke_listener(void* arg) {
 		SSL_set_fd(ssl, client);
 
 		if (SSL_accept(ssl) <= 0) {
-			get_systime(&finish);
-			finish -= start;
-			sockporttoa_r(&addr, errbuf, sizeof(errbuf));
-			msyslog(LOG_ERR, "NTSs: SSL accept from %s failed, %.3Lf sec",
-				errbuf, lfptod(finish));
-			nts_log_ssl_error();
+			clock_gettime(CLOCK_MONOTONIC, &finish);
+			wall = tspec_intv_to_lfp(sub_tspec(finish, start));
+			nts_ke_accept_fail(addrbuf, lfptox(wall));
 			SSL_free(ssl);
 			close(client);
-			nts_ke_serves_bad++;
+			ntske_cnt.serves_nossl++;
+			ntske_cnt.serves_nossl_wall += wall;
+#ifdef RUSAGE_THREAD
+			getrusage(RUSAGE_THREAD, &usage);
+			finish_u = tval_to_tspec(usage.ru_utime);
+			finish_s = tval_to_tspec(usage.ru_stime);
+			usr = tspec_intv_to_lfp(sub_tspec(finish_u, start_u));
+			sys = tspec_intv_to_lfp(sub_tspec(finish_s, start_s));
+			start_u = finish_u;
+			start_s = finish_s;
+			ntske_cnt.serves_nossl_cpu += usr;
+			ntske_cnt.serves_nossl_cpu += sys;
+#endif
 			continue;
 		}
-		msyslog(LOG_INFO, "NTSs: Using %s, %s (%d)",
+
+		/* Save info for final message. */
+		snprintf(usingbuf, sizeof(usingbuf), "%s, %s (%d)",
 			SSL_get_version(ssl),
 			SSL_get_cipher_name(ssl),
 			SSL_get_cipher_bits(ssl, NULL));
 
-
-		if (!nts_ke_request(ssl))
-			nts_ke_serves_bad++;
+		if (nts_ke_request(ssl)) {
+			worked = true;
+			good = "OK";
+		} else {
+			worked = false;
+			good = "Failed";
+		}
 
 		SSL_shutdown(ssl);
 		SSL_free(ssl);
 		close(client);
 
-		get_systime(&finish);
-		finish -= start;
-		nts_ke_serves_good++;
-		msyslog(LOG_INFO, "NTSs: NTS-KE server took %.3Lf sec", lfptod(finish));
-
+		clock_gettime(CLOCK_MONOTONIC, &finish);
+		wall = tspec_intv_to_lfp(sub_tspec(finish, start));
+		if (worked) {
+			ntske_cnt.serves_good++;
+			ntske_cnt.serves_good_wall += wall;
+		} else {
+			ntske_cnt.serves_bad++;
+			ntske_cnt.serves_bad_wall += wall;
+		}
+#ifdef RUSAGE_THREAD
+		getrusage(RUSAGE_THREAD, &usage);
+		finish_u = tval_to_tspec(usage.ru_utime);
+		finish_s = tval_to_tspec(usage.ru_stime);
+		usr = tspec_intv_to_lfp(sub_tspec(finish_u, start_u));
+		sys = tspec_intv_to_lfp(sub_tspec(finish_s, start_s));
+		start_u = finish_u;
+		start_s = finish_s;
+		if (worked) {
+			ntske_cnt.serves_good_cpu += usr;
+			ntske_cnt.serves_good_cpu += sys;
+		} else {
+			ntske_cnt.serves_bad_cpu += usr;
+			ntske_cnt.serves_bad_cpu += sys;
+		}
+		msyslog(LOG_INFO, "NTSs: NTS-KE from %s, %s, Using %s, took %.3f sec, CPU: %.3f+%.3f ms",
+			addrbuf, good, usingbuf, lfptox(wall),
+			lfptox(usr*1000), lfptox(sys*1000));
+#else
+		msyslog(LOG_INFO, "NTSs: NTS-KE from %s, %s, Using %s, took %.3f sec",
+			addrbuf, good, usingbuf, lfptox(wall));
+#endif
 	}
 	return NULL;
 }
 
+/* Analyze failure from SSL_accept
+ * print single error message for common cases.
+ */
+void nts_ke_accept_fail(char* addrbuf, double sec) {
+	unsigned long err = ERR_peek_error();
+	int lib = ERR_GET_LIB(err);
+	int reason = ERR_GET_REASON(err);
+	const char *msg = NULL;
+	if (ERR_LIB_SSL == lib && SSL_R_WRONG_VERSION_NUMBER == reason)
+		msg = "wrong version number";
+	if (ERR_LIB_SSL == lib && SSL_R_HTTP_REQUEST == reason)
+		msg = "http request";
+	if (ERR_LIB_SSL == lib && SSL_R_NO_SHARED_CIPHER == reason)
+		msg = "no shared cipher";
+	if (ERR_LIB_SSL == lib && SSL_R_UNSUPPORTED_PROTOCOL == reason)
+		msg = "unsupported protocol (TLSv1.2?)";
+	if (NULL == msg) {
+		msyslog(LOG_INFO, "NTSs: SSL accept from %s failed, took %.3f sec",
+			addrbuf, sec);
+		nts_log_ssl_error();
+		return;
+	}
+	msyslog(LOG_INFO, "NTSs: SSL accept from %s failed: %s, took %.3f sec",
+		addrbuf, msg, sec);
+}
+
 bool nts_ke_request(SSL *ssl) {
 	/* RFC 4: servers must accept 1024
 	 * Our cookies can be 104, 136, or 168 for AES_SIV_CMAC_xxx
@@ -298,8 +416,10 @@ bool nts_ke_request(SSL *ssl) {
 	if (bytes_written != used)
 		return false;
 
-	msyslog(LOG_INFO, "NTSs: Read %d, wrote %d bytes.  AEAD=%d",
-		bytes_read, bytes_written, aead);
+	/* Skip logging the normal case. */
+	if ((bytes_read!=16) || (aead!=15) )
+		msyslog(LOG_INFO, "NTSs: Read %d, wrote %d bytes.  AEAD=%d",
+			bytes_read, bytes_written, aead);
 
 	return true;
 }
@@ -405,7 +525,7 @@ bool create_listener6(int port) {
 }
 
 bool nts_ke_process_receive(struct BufCtl_t *buf, int *aead) {
-	while (buf->left > 0) {
+	while (buf->left >= NTS_KE_HDR_LNG) {
 		uint16_t type, data;
 		int length;
 		bool critical = false;
@@ -416,7 +536,7 @@ bool nts_ke_process_receive(struct BufCtl_t *buf, int *aead) {
 			type &= ~NTS_CRITICAL;
 		}
 		if (0) // Handy for debugging but very verbose
-			msyslog(LOG_ERR, "NTSs: Record: T=%d, L=%d, C=%d", type, length, critical);
+			msyslog(LOG_INFO, "NTSs: Record: T=%d, L=%d, C=%d", type, length, critical);
 		switch (type) {
 		    case nts_error:
 			data = next_uint16(buf);
@@ -468,6 +588,9 @@ bool nts_ke_process_receive(struct BufCtl_t *buf, int *aead) {
 		} /* case */
 	}   /* while */
 
+	if (buf->left > 0)
+		return false;
+
 	return true;
 
 }
