@@ -1,5 +1,9 @@
 # -*- coding: utf-8 -*-
 
+# Copyright the NTPsec project contributors
+#
+# SPDX-License-Identifier: BSD-2-Clause
+
 """
 packet.py - definitions and classes for Python querying of NTP
 
@@ -56,7 +60,7 @@ the general structure of an NTP packet (Figure 8):
       |                          Key Identifier                       |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       |                                                               |
-      |                            dgst (128)                         |
+      |                          MAC (128/160)                        |
       |                                                               |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 
@@ -66,7 +70,7 @@ Stratum and all following fields zeroed out to byte 47.
 
 How to interpret these fields:
 
-Mode is decoded as follows:
+The modes are as follows:
 
 +-------+--------------------------+
 | Value | Meaning                  |
@@ -82,10 +86,10 @@ Mode is decoded as follows:
 +-------+--------------------------+
 
 While the Stratum field has 8 bytes, only values 0-16 (low 5 bits)
-are legal. Value 16 means 'unsychronized' Value 17-255 are reserved.
+are legal. Value 16 means 'unsynchronized' Values 17-255 are reserved.
 
 LI (Leap Indicator), Version, Poll, and Precision are not described
-here; see RFC5905.
+here; see RFC 5905.
 
 t_1, the origin timestamp, is the time according to the client at
 which the request was sent.
@@ -96,7 +100,7 @@ which the request was received.
 t_3, the transmit timestamp, is the time according to the server at
 which the reply was sent.
 
-You also need t_4, the destination timestamp, is the time according to
+You also need t_4, the destination timestamp, which is the time according to
 the client at which the reply was received.  This is not in the reply packet,
 it's the packet receipt time collected by the client.
 
@@ -107,14 +111,14 @@ most recent reference-clock.
 
 Theta is the thing we want to estimate: the offset between the server
 clock and the client clock. The sign convention is that theta is
-positive iff the server is ahead of the client.
+positive if the server is ahead of the client.
 
 Theta is estimated by [(t_2-t_1)+(t_3-t_4)]/2. The accuracy of this
 estimate is predicated upon network latency being symmetrical.
 
 Delta is the network round trip time, i.e. (t_4-t_1)-(t_3-t_2). Here's
 how the terms work: (t_4-t_1) is the total time that the request was
-in flight, and (t_3-t_2) is time that the server spent processing it;
+in flight, and (t_3-t_2) is the time that the server spent processing it;
 when you subtract that out you're left with just network delays.
 
 Lambda nominally represents the maximum amount by which theta could be
@@ -143,17 +147,16 @@ If you look at the raw data, there are 3 unknowns:
    * clock offset
 but there are only two equations, so you can't solve it.
 
-NTP gets a 3rd equation by assuming the transit times are equal.  That lets
-it solve for the clock offset.
+NTP gets the 3rd equation by assuming the transit times are equal.  That 
+lets it solve for the clock offset.
 
-If you assume that both clocks are accurate which is reasonable if you have
-GPS at both ends, then you can easily solve for the transit times in each
-direction.
+If you assume that both clocks are accurate which is reasonable if you 
+have GPS at both ends, then you can easily solve for the transit times 
+in each direction.
 
-The RFC 5905 diagram is slightly out of date in that the digest header assumes
-a 128-bit (16-octet) MD5 hash, but it is also possible for the field to be a
-128-bit AES_CMAC hash or 160-bit (20-octet) SHA-1 hash.  NTPsec will
-support any 128- or 160-bit MAC type in lincrypto.
+The RFC 5905 diagram is slightly out of date in that the MAC header 
+assumes a 128 or 160-bit (16 or 20-octet) MAC. NTPsec supports all 
+appropriate algorithms in libcrypto.
 
 An extension field consists of a 16-bit network-order type field
 length, followed by a 16-bit network-order payload length in octets,
@@ -189,11 +192,11 @@ Here's what a Mode 6 packet looks like:
       |                          Key Identifier                       |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       |                                                               |
-      |                            dgst (128)                         |
+      |                          MAC (128/160)                        |
       |                                                               |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 
-In this case the fixed header is 24 bytes long.
+In this case, the fixed header is 24 bytes long.
 
 R = Response bit
 E = Error bit
@@ -202,11 +205,12 @@ M = More bit.
 A Mode 6 packet cannot have extension fields.
 
 """
-# SPDX-License-Identifier: BSD-2-Clause
+
 from __future__ import print_function, division
 import getpass
-import hashlib
+import hmac
 import os
+import random
 import select
 import socket
 import string
@@ -253,6 +257,13 @@ DEFSTIMEOUT = 3000
 # The maximum keyid for authentication, keyid is a 16-bit field
 MAX_KEYID = 0xFFFF
 
+# Some constants (in Bytes) for mode6 and authentication
+MODE_SIX_HEADER_LENGTH = 12
+MINIMUM_MAC_LENGTH = 16
+KEYID_LENGTH = 4
+MODE_SIX_ALIGNMENT = 4
+MAX_BARE_MAC_LENGTH = 20
+
 
 class Packet:
     "Encapsulate an NTP fragment"
@@ -351,16 +362,14 @@ class SyncPacket(Packet):
          self.transmit_timestamp) = struct.unpack(
             SyncPacket.format, data[:SyncPacket.HEADER_LEN])
         self.extension = data[SyncPacket.HEADER_LEN:]
-        # Parse the extension field if present. We figure out whether
-        # an extension field is present by measuring the MAC size. If
-        # the number of 4-octet words following the packet header is
-        # 0, no MAC is present and the packet is not authenticated. If
-        # 1, the packet is a crypto-NAK; if 3, the packet is
-        # authenticated with DES; if 5, the packet is authenticated
-        # with MD5; if 6, the packet is authenticated with SHA1. If 2
-        # or 4, the packet is a runt and discarded forthwith. If
-        # greater than 6, an extension field is present, so we
-        # subtract the length of the field and go around again.
+        # Parse any present extension fields. We determine the presence
+        # of an extension field by measuring the trailer size. If the
+        # number of 4-octet words following the packet header is 0, there
+        # is no MAC, and the packet is unauthenticated. If 1, the missive
+        # is a crypto-NAK; if 5 or 6, register authentication. If 2, 3, or
+        # 4 the datagram is a runt and discarded immediately. If greater
+        # than 6, an extension field is present, so we subtract the fields'
+        # length and repeat.
         payload = self.extension  # Keep extension intact for flatten()
         while len(payload) > 24:
             (ftype, flen) = struct.unpack("!II", payload[:8])
@@ -368,11 +377,9 @@ class SyncPacket(Packet):
             payload = payload[8+flen:]
         if len(payload) == 4:    # Crypto-NAK
             self.mac = payload
-        elif len(payload) == 12:   # DES
-            raise SyncException("Unsupported DES authentication")
-        elif len(payload) in (8, 16):
+        elif len(payload) in (8, 12, 16):
             raise SyncException("Packet is a runt")
-        elif len(payload) in (20, 24):   # MD5 or SHA1
+        elif len(payload) in (20, 24):   # Symmetric authentication
             self.mac = payload
 
     @staticmethod
@@ -576,7 +583,7 @@ class Peer:
         self.variables = {}
 
     def readvars(self):
-        self.variables = self.session.readvars()
+        self.variables = self.session.readvar()
 
     def __str__(self):
         return "<Peer: associd=%s status=%0x>" % (self.associd, self.status)
@@ -637,9 +644,11 @@ class MRUEntry:
         self.addr = None        # text of IPv4 or IPv6 address and port
         self.last = None        # timestamp of last receipt
         self.first = None       # timestamp of first receipt
-        self.ct = 0             # count of packets received
         self.mv = None          # mode and version
         self.rs = None          # restriction mask (RES_* bits)
+        self.ct = 0             # count of packets received
+        self.sc = None          # score
+        self.dr = None          # dropped packets
 
     def avgint(self):
         last = ntp.ntpc.lfptofloat(self.last)
@@ -696,6 +705,7 @@ class ControlException(BaseException):
 class ControlSession:
     "A session to a host"
     MRU_ROW_LIMIT = 256
+    _authpass = True
     server_errors = {
         ntp.control.CERR_UNSPEC: "UNSPEC",
         ntp.control.CERR_PERMISSION: "PERMISSION",
@@ -715,7 +725,7 @@ class ControlSession:
         # Packet version number we use
         self.pktversion = ntp.magic.NTP_OLDVERSION + 1
         self.always_auth = False  # Always send authenticated requests
-        self.keytype = "MD5"
+        self.keytype = "AES"
         self.keyid = None
         self.passwd = None
         self.auth = None
@@ -729,6 +739,11 @@ class ControlSession:
         self.ntpd_row_limit = ControlSession.MRU_ROW_LIMIT
         self.logfp = sys.stdout
         self.nonce_xmit = 0
+        self.slots = 0
+        self.flakey = None
+
+    def warndbg(self, text, threshold):
+        ntp.util.dolog(self.logfp, text, self.debug, threshold)
 
     def close(self):
         if self.sock:
@@ -743,8 +758,8 @@ class ControlSession:
         "Try different ways to interpret an address and family"
         if hname.startswith("["):
             hname = hname[1:-1]
-        # First try to resolve it as an ip address and if that fails,
-        # do a fullblown (DNS) lookup. That way we only use the dns
+        # First try to resolve it as an IP address and if that fails,
+        # do a fullblown (DNS) lookup. That way we only use the DNS
         # when it is needed and work around some implementations that
         # will return an "IPv4-mapped IPv6 address" address if you
         # give it an IPv4 address to lookup.
@@ -925,7 +940,7 @@ class ControlSession:
         # Pad out packet to a multiple of 8 octets to be sure
         # receiver can handle it. Note: these pad bytes should
         # *not* be counted in the header count field.
-        while ((ControlPacket.HEADER_LEN + len(pkt.extension)) & 7):
+        while ((ControlPacket.HEADER_LEN + len(pkt.extension)) & (MODE_SIX_ALIGNMENT - 1)):
             pkt.extension += b"\x00"
 
         # Do the MAC compuation.
@@ -1005,6 +1020,10 @@ class ControlSession:
                 # usually, errno 111: connection refused
                 raise ControlException(SERR_SOCKET)
 
+            if self.flakey and self.flakey >= random.random():
+                warndbg('Flaky: I deliberately dropped a packet.', 1)
+                rawdata = None
+
             warndbg("Received %d octets" % len(rawdata), 3)
             rpkt = ControlPacket(self)
             try:
@@ -1018,6 +1037,16 @@ class ControlSession:
                 continue
 
             # Someday, perhaps, check authentication here
+            if self._authpass and self.auth:
+                _pend = rpkt.count + MODE_SIX_HEADER_LENGTH
+                _pend += (-_pend % MODE_SIX_ALIGNMENT)
+                if len(rawdata) < (_pend + KEYID_LENGTH + MINIMUM_MAC_LENGTH):
+                    self.logfp.write('AUTH - packet too short for MAC %d < %d\n' %
+                                     (len(rawdata), (_pend + KEYID_LENGTH + MINIMUM_MAC_LENGTH)))
+                    self._authpass = False
+                elif not self.auth.verify_mac(rawdata, packet_end=_pend,
+                                              mac_begin=_pend):
+                    self._authpass = False
 
             # Clip off the MAC, if any
             rpkt.extension = rpkt.extension[:rpkt.count]
@@ -1082,7 +1111,7 @@ class ControlSession:
                                 % (f, len(fragments)), 1)
                         break
                 else:
-                    tempfraglist = [ntp.poly.polystr(f.extension) \
+                    tempfraglist = [ntp.poly.polystr(f.extension)
                                     for f in fragments]
                     self.response = ntp.poly.polybytes("".join(tempfraglist))
                     warndbg("Fragment collection ends. %d bytes "
@@ -1097,11 +1126,13 @@ class ControlSession:
                         warn("Response packet:\n%s\n" % repr(self.response))
                     elif self.debug >= 2:  # pragma: no cover
                         # FIXME: Garbage when retrieving assoc list (binary)
-                        eol = self.response.find("\n")
+                        eol = self.response.find(b"\n")
                         firstline = self.response[:eol]
                         warn("First line:\n%s\n" % repr(firstline))
                     return None
                 break
+        if not self._authpass:
+            warn('AUTH: Content untrusted due to authentication failure!\n')
 
     def __validate_packet(self, rpkt, rawdata, opcode, associd):
         # TODO: refactor to simplify while retaining semantic info
@@ -1300,70 +1331,108 @@ This combats source address spoofing
         self.logfp.write("## Nonce expected: %s" % resp)
         raise ControlException(SERR_BADNONCE)
 
+    def __mru_analyze(self, variables, span, direct):
+        """Extracts data from the key/value list into a more useful form"""
+        mru = None
+        nonce = None
+        items = list(variables.items())
+        fake_list = []
+        fake_dict = {}
+        if items:                   # See issue #642
+            items.sort(key=mru_kv_key)
+        for (tag, val) in items:
+            self.warndbg("tag=%s, val=%s" % (tag, val), 4)
+            if tag == "nonce":
+                nonce = "%s=%s" % (tag, val)
+            elif tag == "last.older":
+                continue
+            elif tag == "addr.older":
+                continue
+            if tag == "now":
+                # finished marker
+                span.now = ntp.ntpc.lfptofloat(val)
+                continue
+            elif tag == "last.newest":
+                # more finished
+                continue
+            for prefix in ("addr", "last", "first", "ct", "mv", "rs", "sc", "dr"):
+                if tag.startswith(prefix + "."):
+                    (member, idx) = tag.split(".")
+                    try:
+                        idx = int(idx)
+                    except ValueError:
+                        raise ControlException(SERR_BADTAG % tag)
+                    # Does not check missing/gappy entries
+                    if idx not in fake_list:
+                        fake_dict[str(idx)] = {}
+                        fake_list.append(idx)
+                    fake_dict[str(idx)][member] = val
+        fake_list.sort()
+        for idx in fake_list:
+            mru = MRUEntry()
+            self.slots += 1
+            # Always 6 in practice, in the tests not so much
+#            if len(fake_dict[str(idx)]) != 6:
+#                continue
+            for prefix in ("addr", "last", "first", "ct", "mv", "rs", "sc", "dr"):
+                if prefix in fake_dict[str(idx)]:  # dodgy test needs this line
+                    setattr(mru, prefix, fake_dict[str(idx)][prefix])
+            span.entries.append(mru)
+        if direct is not None:
+            direct(span.entries)
+        return nonce
+
+    def __mru_query_error(self, e, restarted_count, cap_frags, limit, frags):
+        if e.errorcode is None:
+            raise e
+        elif e.errorcode == ntp.control.CERR_UNKNOWNVAR:
+            # None of the supplied prior entries match, so
+            # toss them from our list and try again.
+            self.warndbg("no overlap between prior entries and "
+                         "server MRU list", 1)
+            restarted_count += 1
+            if restarted_count > 8:
+                raise ControlException(SERR_STALL)
+            self.warndbg("--->   Restarting from the beginning, "
+                         "retry #%u" % restarted_count, 1)
+        elif e.errorcode == ntp.control.CERR_BADVALUE:
+            if cap_frags:
+                cap_frags = False
+                self.warndbg("Reverted to row limit from "
+                             "fragments limit.", 1)
+            else:
+                # ntpd has lower cap on row limit
+                self.ntpd_row_limit -= 1
+                limit = min(limit, self.ntpd_row_limit)
+                self.warndbg("Row limit reduced to %d following "
+                             "CERR_BADVALUE." % limit, 1)
+        elif e.errorcode in (SERR_INCOMPLETE, SERR_TIMEOUT):
+            # Reduce the number of rows/frags requested by
+            # half to recover from lost response fragments.
+            if cap_frags:
+                frags = max(2, frags / 2)
+                self.warndbg("Frag limit reduced to %d following "
+                             "incomplete response." % frags, 1)
+            else:
+                limit = max(2, limit / 2)
+                self.warndbg("Row limit reduced to %d following "
+                             " incomplete response." % limit, 1)
+        elif e.errorcode:
+            raise e
+        return restarted_count, cap_frags, limit, frags
+
     def mrulist(self, variables=None, rawhook=None, direct=None):
         "Retrieve MRU list data"
         restarted_count = 0
         cap_frags = True
-        warndbg = (lambda txt, th: ntp.util.dolog(self.logfp, txt,
-                                                  self.debug, th))
         sorter = None
+        sortkey = None
         frags = MAXFRAGS
-        self.slots = 0
         if variables is None:
             variables = {}
 
         if variables:
-            if "sort" in variables:
-                sortkey = variables["sort"]
-                del variables["sort"]
-                # Slots are retrieved oldest first.
-                # Slots are printed in reverse so the normal/no-sort
-                #  case prints youngest first.
-                # That means sort functions are backwards.
-                # Note lstint is backwards again (aka normal/forward)
-                #  since we really want to sort on now-last rather than last.
-                sortdict = {
-                    # lstint ascending
-                    "lstint": lambda e: ntp.ntpc.lfptofloat(e.last),
-                    # lstint descending
-                    "-lstint": lambda e: -ntp.ntpc.lfptofloat(e.last),
-                    # avgint ascending
-                    "avgint": lambda e: -e.avgint(),
-                    # avgint descending
-                    "-avgint": lambda e: e.avgint(),
-                    # IPv4 asc. then IPv6 asc.
-                    "addr": lambda e: e.sortaddr(),
-                    # IPv6 desc. then IPv4 desc.
-                    "-addr": lambda e: e.sortaddr(),
-                    # hit count ascending
-                    "count": lambda e: -e.ct,
-                    # hit count descending
-                    "-count": lambda e: e.ct,
-                }
-                if sortkey == "lstint":
-                    sortkey = None   # normal/default case, no need to sort
-                if sortkey is not None:
-                    sorter = sortdict.get(sortkey)
-                    if sorter is None:
-                        raise ControlException(SERR_BADSORT % sortkey)
-            for k in list(variables.keys()):
-                if k in ("mincount", "resall", "resany", "kod", "limited",
-                         "maxlstint", "laddr", "recent", "sort",
-                         "frags", "limit"):
-                    continue
-                else:
-                    raise ControlException(SERR_BADPARAM % k)
-            if 'frags' in variables:
-                frags = int(variables.get('frags'))
-                del variables['frags']
-            if 'kod' in variables:
-                variables['resany'] = variables.get('resany', 0) \
-                    | ntp.magic.RES_KOD
-                del variables['kod']
-            if 'limited' in variables:
-                variables['resany'] = variables.get('resany', 0) \
-                    | ntp.magic.RES_LIMITED
-                del variables['limited']
+            sorter, sortkey, frags = parse_mru_variables(variables)
 
         nonce = self.fetch_nonce()
 
@@ -1377,12 +1446,8 @@ This combats source address spoofing
                     variables['resall'] = hex(variables['resall'])
                 if 'resany' in variables:
                     variables['resany'] = hex(variables['resany'])
-                parms = ", " + ",".join([("%s=%s" % it)
-                                         for it in list(variables.items())])
-            else:
-                parms = ""
-            req_buf += parms
-            first_time_only = "recent=%s" % variables.get("recent")
+            parms, firstParms = generate_mru_parms(variables)
+            req_buf += firstParms
 
             while True:
                 # Request additions to the MRU list
@@ -1392,42 +1457,9 @@ This combats source address spoofing
                     recoverable_read_errors = False
                 except ControlException as e:
                     recoverable_read_errors = True
-                    if e.errorcode is None:
-                        raise e
-                    elif e.errorcode == ntp.control.CERR_UNKNOWNVAR:
-                        # None of the supplied prior entries match, so
-                        # toss them from our list and try again.
-                        warndbg("no overlap between prior entries and "
-                                "server MRU list", 1)
-                        restarted_count += 1
-                        if restarted_count > 8:
-                            raise ControlException(SERR_STALL)
-                        warndbg("--->   Restarting from the beginning, "
-                                "retry #%u" % restarted_count, 1)
-                    elif e.errorcode == ntp.control.CERR_BADVALUE:
-                        if cap_frags:
-                            cap_frags = False
-                            warndbg("Reverted to row limit from "
-                                    "fragments limit.", 1)
-                        else:
-                            # ntpd has lower cap on row limit
-                            self.ntpd_row_limit -= 1
-                            limit = min(limit, self.ntpd_row_limit)
-                            warndbg("Row limit reduced to %d following "
-                                    "CERR_BADVALUE." % limit, 1)
-                    elif e.errorcode in (SERR_INCOMPLETE, SERR_TIMEOUT):
-                        # Reduce the number of rows/frags requested by
-                        # half to recover from lost response fragments.
-                        if cap_frags:
-                            frags = max(2, frags / 2)
-                            warndbg("Frag limit reduced to %d following "
-                                    "incomplete response." % frags, 1)
-                        else:
-                            limit = max(2, limit / 2)
-                            warndbg("Row limit reduced to %d following "
-                                    " incomplete response." % limit, 1)
-                    elif e.errorcode:
-                        raise e
+                    res = self.__mru_query_error(e, restarted_count,
+                                                 cap_frags, limit, frags)
+                    restarted_count, cap_frags, limit, frags = res
 
                 # Parse the response
                 variables = self.__parse_varlist()
@@ -1444,46 +1476,9 @@ This combats source address spoofing
                     rawhook(variables)
 
                 # Analyze the contents of this response into a span structure
-                curidx = -1
-                mru = None
-                for (tag, val) in variables.items():
-                    warndbg("tag=%s, val=%s" % (tag, val), 4)
-                    if tag == "nonce":
-                        nonce = "%s=%s" % (tag, val)
-                    elif tag == "last.older":
-                        continue
-                    elif tag == "addr.older":
-                        continue
-                    if tag == "now":
-                        # finished marker
-                        span.now = ntp.ntpc.lfptofloat(val)
-                        continue
-                    elif tag == "last.newest":
-                        # more finished
-                        continue
-                    for prefix in ("addr", "last", "first", "ct", "mv", "rs"):
-                        if tag.startswith(prefix + "."):
-                            (member, idx) = tag.split(".")
-                            try:
-                                idx = int(idx)
-                            except ValueError:
-                                raise ControlException(SERR_BADTAG % tag)
-                            if idx != curidx:
-                                # This makes duplicates
-                                curidx = idx
-
-                                if mru:
-                                    # Can't have partial slots on list
-                                    # or printing crashes after ^C
-                                    # Append full slot now
-                                    span.entries.append(mru)
-                                mru = MRUEntry()
-                                self.slots += 1
-                            setattr(mru, prefix, val)
-                if mru:
-                    span.entries.append(mru)
-                if direct is not None:
-                    direct(span.entries)
+                newNonce = self.__mru_analyze(variables, span, direct)
+                if newNonce:
+                    nonce = newNonce
 
                 # If we've seen the end sentinel on the span, break out
                 if span.is_complete():
@@ -1513,9 +1508,6 @@ This combats source address spoofing
                                     max(limit + 1,
                                         limit * 33 / 32))
 
-                # Only ship 'recent' on the first request
-                parms = parms.replace(first_time_only, "")
-
                 # Prepare next query with as many address and last-seen
                 # timestamps as will fit in a single packet.  A new nonce
                 # might be required.
@@ -1526,45 +1518,13 @@ This combats source address spoofing
                            "frags" if cap_frags else "limit",
                            frags if cap_frags else limit,
                            parms)
-                for i in range(len(span.entries)):
-                    e = span.entries[len(span.entries) - i - 1]
-                    incr = ", addr.%d=%s, last.%d=%s" % (i, e.addr, i, e.last)
-                    if (len(req_buf) + len(incr) >=
-                            ntp.control.CTL_MAX_DATA_LEN):
-                        break
-                    else:
-                        req_buf += incr
+                req_buf += generate_mru_lastseen(span, len(req_buf))
                 if direct is not None:
                     span.entries = []
         except KeyboardInterrupt:  # pragma: no cover
             pass        # We can test for interruption with is_complete()
 
-        # C ntpq's code for stitching together spans was absurdly
-        # overelaborate - all that dancing with last.older and
-        # addr.older was, as far as I can tell, just pointless.
-        # Much simpler to just run through the final list throwing
-        # out every entry with an IP address that is duplicated
-        # with a later most-recent-transmission time.
-        addrdict = {}
-        deletia = []
-        for (i, entry) in enumerate(span.entries):
-            if entry.addr not in addrdict:
-                addrdict[entry.addr] = []
-            addrdict[entry.addr].append((i, entry.last))
-        for addr in addrdict:
-            deletia += sorted(addrdict[addr], key=lambda x: x[1])[:-1]
-        deletia = [x[0] for x in deletia]
-        deletia.sort(reverse=True)
-        for i in deletia:
-            span.entries.pop(i)
-
-        # Sort for presentation
-        if sorter:
-            span.entries.sort(key=sorter)
-            if sortkey == "addr":
-                # I don't know how to feed a minus sign to text sort
-                span.entries.reverse()
-
+        stitch_mru(span, sorter, sortkey)
         return span
 
     def __ordlist(self, listtype):
@@ -1591,6 +1551,145 @@ This combats source address spoofing
         return self.__ordlist("ifstats")
 
 
+def parse_mru_variables(variables):
+    sorter = None
+    sortkey = None
+    frags = MAXFRAGS
+    if "sort" in variables:
+        sortkey = variables["sort"]
+        del variables["sort"]
+        # Slots are retrieved oldest first.
+        # Slots are printed in reverse so the normal/no-sort
+        #  case prints youngest first.
+        # That means sort functions are backwards.
+        # Note lstint is backwards again (aka normal/forward)
+        #  since we really want to sort on now-last rather than last.
+        sortdict = {
+            # lstint ascending
+            "lstint": lambda e: ntp.ntpc.lfptofloat(e.last),
+            # lstint descending
+            "-lstint": lambda e: -ntp.ntpc.lfptofloat(e.last),
+            # avgint ascending
+            "avgint": lambda e: -e.avgint(),
+            # avgint descending
+            "-avgint": lambda e: e.avgint(),
+            # IPv4 asc. then IPv6 asc.
+            "addr": lambda e: e.sortaddr(),
+            # IPv6 desc. then IPv4 desc.
+            "-addr": lambda e: e.sortaddr(),
+            # hit count ascending
+            "count": lambda e: -e.ct,
+            # hit count descending
+            "-count": lambda e: e.ct,
+            # score ascending
+            "score": lambda e: -e.sc,
+            # score descending
+            "-score": lambda e: e.sc,
+            # drop count ascending
+            "drop": lambda e: -e.dr,
+            # drop count descending
+            "-drop": lambda e: e.dr,
+        }
+        if sortkey == "lstint":
+            sortkey = None   # normal/default case, no need to sort
+        if sortkey is not None:
+            sorter = sortdict.get(sortkey)
+            if sorter is None:
+                raise ControlException(SERR_BADSORT % sortkey)
+    for k in list(variables.keys()):
+        if k in ("mincount", "mindrop", "minscore",
+                 "resall", "resany", "kod", "limited",
+                 "maxlstint", "minlstint", "laddr", "recent",
+                 "sort", "frags", "limit"):
+            continue
+        elif k.startswith('addr.') or k.startswith('last.'):
+            kn = k.split('.')
+            if len(kn) != 2 or kn[1] not in map(str, list(range(16))):
+                raise ControlException(SERR_BADPARAM % k)
+            continue
+        else:
+            raise ControlException(SERR_BADPARAM % k)
+    if 'frags' in variables:
+        frags = int(variables.get('frags'))
+        del variables['frags']
+    if 'kod' in variables:
+        variables['resany'] = variables.get('resany', 0) \
+                              | ntp.magic.RES_KOD
+        del variables['kod']
+    if 'limited' in variables:
+        variables['resany'] = variables.get('resany', 0) \
+                              | ntp.magic.RES_LIMITED
+        del variables['limited']
+    return sorter, sortkey, frags
+
+
+def stitch_mru(span, sorter, sortkey):
+    # C ntpq's code for stitching together spans was absurdly
+    # overelaborate - all that dancing with last.older and
+    # addr.older was, as far as I can tell, just pointless.
+    # Much simpler to just run through the final list throwing
+    # out every entry with an IP address that is duplicated
+    # with a later most-recent-transmission time.
+    addrdict = {}
+    deletia = []
+    for (i, entry) in enumerate(span.entries):
+        if entry.addr not in addrdict:
+            addrdict[entry.addr] = []
+        addrdict[entry.addr].append((i, entry.last))
+    for addr in addrdict:
+        deletia += sorted(addrdict[addr], key=lambda x: x[1])[:-1]
+    deletia = [x[0] for x in deletia]
+    deletia.sort(reverse=True)  # Delete from top down so indices don't change
+    for i in deletia:
+        span.entries.pop(i)
+
+    # Sort for presentation
+    if sorter:
+        span.entries.sort(key=sorter)
+        if sortkey == "addr":
+            # I don't know how to feed a minus sign to text sort
+            span.entries.reverse()
+
+
+def generate_mru_parms(variables):
+    if not variables:
+        return "", ""
+    # generate all sans recent
+    parmStrs = [("%s=%s" % it)
+                for it in list(variables.items()) if (it[0] != "recent")]
+    parms = ", " + ", ".join(parmStrs)
+    # Only ship 'recent' on the first request
+    if "recent" in variables:
+        firstParms = ", recent=%s" % variables["recent"]
+        firstParms += parms
+    else:
+        firstParms = parms
+    return parms, firstParms
+
+
+def generate_mru_lastseen(span, existingBufferSize):
+    buf = ""
+    for i in range(len(span.entries)):
+        e = span.entries[len(span.entries) - i - 1]
+        incr = ", addr.%d=%s, last.%d=%s" % (i, e.addr, i, e.last)
+        if (existingBufferSize + len(buf) + len(incr) >=
+                ntp.control.CTL_MAX_DATA_LEN):
+            break
+        else:
+            buf += incr
+    return buf
+
+
+def mru_kv_key(token):
+    bits = token[0].split('.')
+    if len(bits) == 1:
+        return -2
+    try:
+        return int(bits[1])
+    except ValueError:
+        return -1
+
+
 class Authenticator:
     "MAC authentication manager for NTP packets."
 
@@ -1605,16 +1704,24 @@ class Authenticator:
                 if not line:
                     continue
                 (keyid, keytype, passwd) = line.split()
+                if keytype.upper() in ['AES', 'AES128CMAC']:
+                    keytype = 'AES-128'
+                if len(passwd) > 20:
+                    # if len(passwd) > 64:
+                    #      print('AUTH: Truncating key %s to 256bits (32Bytes)' % keyid)
+                    passwd = ntp.util.hexstr2octets(passwd[:64])
                 self.passwords[int(keyid)] = (keytype, passwd)
 
     def __len__(self):
+        'return the number of keytype/passwd tuples stored'
         return len(self.passwords)
 
     def __getitem__(self, keyid):
+        'get a keytype/passwd tuple by keyid'
         return self.passwords.get(keyid)
 
     def control(self, keyid=None):
-        "Get a keyid/passwd pair that is trusted on localhost"
+        "Get the keytype/passwd tuple that controls localhost and its id"
         if keyid is not None:
             if keyid in self.passwords:
                 return (keyid,) + self.passwords[keyid]
@@ -1630,19 +1737,19 @@ class Authenticator:
                 if len(passwd) > 20:
                     passwd = ntp.util.hexstr2octets(passwd)
                 return (keyid, keytype, passwd)
-        else:
-            # No control lines found
-            raise ValueError
+        # No control lines found
+        raise ValueError
 
     @staticmethod
     def compute_mac(payload, keyid, keytype, passwd):
-        hasher = hashlib.new(keytype)
-        hasher.update(ntp.poly.polybytes(passwd))
-        hasher.update(payload)
-        if hasher.digest_size == 0:
-            return None
-        else:
-            return struct.pack("!I", keyid) + hasher.digest()
+        'Create the authentication payload to send'
+        if not ntp.ntpc.checkname(keytype):
+            return False
+        mac2 = ntp.ntpc.mac(ntp.poly.polybytes(payload),
+                            ntp.poly.polybytes(passwd), keytype)[:20]
+        if not mac2 or len(mac2) == 0:
+            return b''
+        return struct.pack("!I", keyid) + mac2
 
     @staticmethod
     def have_mac(packet):
@@ -1652,20 +1759,26 @@ class Authenticator:
         # On those you have to go in and look at the count.
         return len(packet) > ntp.magic.LEN_PKT_NOMAC
 
-    def verify_mac(self, packet):
+    def verify_mac(self, packet, packet_end=48, mac_begin=48):
         "Does the MAC on this packet verify according to credentials we have?"
-        # FIXME: Someday, figure out how to handle SHA1?
-        HASHLEN = 16    # Length of MD5 hash.
-        payload = packet[:-HASHLEN-4]
-        keyid = packet[-HASHLEN-4:-HASHLEN]
-        mac = packet[-HASHLEN:]
+        payload = packet[:packet_end]
+        keyid = packet[mac_begin:mac_begin+KEYID_LENGTH]
+        mac = packet[mac_begin+KEYID_LENGTH:]
         (keyid,) = struct.unpack("!I", keyid)
         if keyid not in self.passwords:
+            # print('AUTH: No key %08x...' % keyid)
             return False
         (keytype, passwd) = self.passwords[keyid]
-        hasher = hashlib.new(keytype)
-        hasher.update(passwd)
-        hasher.update(payload)
-        return ntp.poly.polybytes(hasher.digest()) == mac
+        if not ntp.ntpc.checkname(keytype):
+            return False
+        mac2 = ntp.ntpc.mac(ntp.poly.polybytes(payload),
+                            ntp.poly.polybytes(passwd), keytype)[:20]
+        if not mac2:
+            return False
+        # typically preferred to avoid timing attacks client-side (in theory)
+        try:
+            return hmac.compare_digest(mac, mac2)  # supported 2.7.7+ and 3.3+
+        except AttributeError:
+            return mac == mac2  # solves issue #666
 
 # end
