@@ -1,5 +1,8 @@
 # -*- coding: utf-8 -*-
 "Common utility functions"
+
+# Copyright the NTPsec project contributors
+#
 # SPDX-License-Identifier: BSD-2-Clause
 
 from __future__ import print_function, division
@@ -77,7 +80,7 @@ def deunicode_units():  # pragma: no cover
 S_VARS = ("tai", "poll")
 MS_VARS = ("rootdelay", "rootdisp", "rootdist", "offset", "sys_jitter",
            "clk_jitter", "leapsmearoffset", "authdelay", "koffset", "kmaxerr",
-           "kesterr", "kprecis", "kppsjitter", "fuzz", "clk_wander_threshold",
+           "kesterr", "kprecis", "kppsjitter", "clk_wander_threshold",
            "tick", "in", "out", "bias", "delay", "jitter", "dispersion",
            "fudgetime1", "fudgetime2")
 PPM_VARS = ("frequency", "clk_wander")
@@ -87,7 +90,7 @@ def dolog(logfp, text, debug, threshold):
     """debug is the current debug value
     threshold is the trigger for the current log"""
     if logfp is None:
-        return  # can turn off logging by supplying a None file descriptior
+        return  # can turn off logging by supplying a None file descriptor
     text = rfc3339(time.time()) + " " + text + "\n"
     if debug >= threshold:
         logfp.write(text)
@@ -112,6 +115,14 @@ def stdversion():
     return "ntpsec-%s" % "@NTPSEC_VERSION_EXTENDED@"
 
 
+def stdversioncheck(foreign):
+    "Print a warning to stderr if module and foreign versions do not match."
+    if stdversion() != foreign:
+        sys.stderr.write("Module/Binary version mismatch\n")
+        sys.stderr.write("Binary: %s\n" % foreign)
+        sys.stderr.write("Module: %s\n" % ntp.util.stdversion())
+
+
 def rfc3339(t):
     "RFC 3339 string from Unix time, including fractional second."
     rep = time.strftime("%Y-%m-%dT%H:%M:%S", time.gmtime(t))
@@ -266,7 +277,6 @@ def stringfiltcooker(data):
         if count > highestcount:
             mostcommon = key
             highestcount = count
-    newunit = UNITS_SEC[mostcommon + UNITS_SEC.index(UNIT_MS)]
     # Shift all values to the new unit
     cooked = []
     for part in parts:
@@ -497,7 +507,7 @@ def unitifyvar(value, varname, baseunit=None, width=8, unitSpace=False):
 
 
 def unitify(value, startingunit, baseunit=None, width=8, unitSpace=False):
-    "Formats a numberstring with relevant units. Attemps to fit in width."
+    "Formats a numberstring with relevant units. Attempts to fit in width."
     if baseunit is None:
         baseunit = getunitgroup(startingunit)[0]
     ooms = oomsbetweenunits(startingunit, baseunit)
@@ -639,7 +649,7 @@ def canonicalize_dns(inhost, family=socket.AF_UNSPEC):
     if resname is not None:
         return resname
     # Catch garbaged hostnames in corrupted Mode 6 responses
-    m = re.match("([:.[\]]|\w)*", inhost)
+    m = re.match(r"([:.[\]]|\w)*", inhost)
     if not m:
         raise TypeError
     (hostname, portsuffix) = portsplit(inhost)
@@ -713,7 +723,7 @@ class PeerStatusWord:
         else:
             self.conf = "no"
         # Reach
-        if statval & ntp.control.CTL_PST_BCAST:
+        if statval & ntp.control.CTL_PST_BCASTx:
             self.reach = "none"
         elif statval & ntp.control.CTL_PST_REACH:
             self.reach = "yes"
@@ -938,6 +948,9 @@ class PeerSummary:
             self.__header = self.__remote + \
                 "       local      ".ljust(self.refidwidth) + \
                 self.__common + "  disp"
+        elif self.displaymode == 'rpeers':
+            self.__header = ' st t when poll reach   delay   ' + \
+                            'offset   jitter refid           T remote'
         else:
             self.__header = self.__remote + \
                 "       refid      ".ljust(self.refidwidth) + \
@@ -1096,7 +1109,7 @@ class PeerSummary:
         if hmode == ntp.magic.MODE_BCLIENTX:
             # broadcastclient or multicastclient
             ptype = 'b'
-        elif hmode == ntp.magic.MODE_BROADCAST:
+        elif hmode == ntp.magic.MODE_BROADCASTx:
             # broadcast or multicast server
             if srcadr.startswith("224."):       # IANA multicast address prefix
                 ptype = 'M'
@@ -1114,9 +1127,9 @@ class PeerSummary:
                 ptype = chr(ntscookies + ord('0'))
             else:
                 ptype = 'u'     # unicast
-        elif hmode == ntp.magic.MODE_ACTIVE:
+        elif hmode == ntp.magic.MODE_ACTIVEx:
             ptype = 's'         # symmetric active
-        elif hmode == ntp.magic.MODE_PASSIVE:
+        elif hmode == ntp.magic.MODE_PASSIVEx:
             ptype = 'S'         # symmetric passive
 
         #
@@ -1140,7 +1153,9 @@ class PeerSummary:
                 and srcadr != "0.0.0.0" \
                 and not srcadr.startswith("127.127") \
                 and srcadr != "::":
-            if self.showhostnames:
+            if self.showhostnames & 2 and 'srchost' in locals() and srchost:
+                clock_name = srchost
+            elif self.showhostnames & 1:
                 try:
                     if self.debug:
                         self.logfp.write("DNS lookup begins...\n")
@@ -1158,12 +1173,13 @@ class PeerSummary:
                 clock_name = srcadr
             else:
                 clock_name = ""
-        if self.wideremote and len(clock_name) > self.namewidth:
-            line += ("%c%s\n" % (c, clock_name))
-            line += (" " * (self.namewidth + 2))
-        else:
-            line += ("%c%-*.*s " % (c, self.namewidth, self.namewidth,
-                                    clock_name[:self.namewidth]))
+        if self.displaymode != "rpeers":
+            if self.wideremote and len(clock_name) > self.namewidth:
+                line += ("%c%s\n" % (c, clock_name))
+                line += (" " * (self.namewidth + 2))
+            else:
+                line += ("%c%-*.*s " % (c, self.namewidth, self.namewidth,
+                                        clock_name[:self.namewidth]))
         # Destination address, assoc ID or refid.
         assocwidth = 7 if self.displaymode == "apeers" else 0
         if "." not in dstadr_refid and ":" not in dstadr_refid:
@@ -1172,45 +1188,33 @@ class PeerSummary:
             visible = "..."
         else:
             visible = dstadr_refid
-        line += self.high_truncate(visible, self.refidwidth)
-        if self.displaymode == "apeers":
-            line += (" " * (self.refidwidth - len(visible) - assocwidth + 1))
-            line += ("%-6d" % (associd))
-        else:
-            line += (" " * (self.refidwidth - len(visible)))
+        if self.displaymode != "rpeers":
+            line += self.high_truncate(visible, self.refidwidth)
+            if self.displaymode == "apeers":
+                line += (" " * (self.refidwidth - len(visible) - assocwidth + 1))
+                line += ("%-6d" % (associd))
+            else:
+                line += (" " * (self.refidwidth - len(visible)))
         # The rest of the story
         if last_sync is None:
             last_sync = now
         jd = estjitter if have_jitter else estdisp
+        if self.showunits:
+            fini = lambda x : unitify(x, UNIT_MS)
+        elif saw6:
+            fini = lambda x : f8dot4(x)
+        else:
+            fini = lambda x : f8dot3(x)
         line += (
-            " %2ld %c %4.4s %4.4s  %3lo"
+            " %2ld %c %4.4s %4.4s  %3lo %s %s %s"
             % (stratum, ptype,
                PeerSummary.prettyinterval(last_sync),
-               PeerSummary.prettyinterval(poll_sec), reach))
-        if saw6:
-            if self.showunits:
-                line += (
-                    " %s %s %s" %
-                    (unitify(estdelay, UNIT_MS),
-                     unitify(estoffset, UNIT_MS),
-                     unitify(jd, UNIT_MS)))
-            else:
-                line += (
-                    " %s %s %s" %
-                    (f8dot4(estdelay), f8dot4(estoffset), f8dot4(jd)))
-        else:
-            # old servers only have 3 digits of fraction
-            # don't print a fake 4th digit
-            if self.showunits:
-                line += (
-                    " %s %s %s" %
-                    (unitify(estdelay, UNIT_MS),
-                     unitify(estoffset, UNIT_MS),
-                     unitify(jd, UNIT_MS)))
-            else:
-                line += (
-                    " %s %s %s" %
-                    (f8dot3(estdelay), f8dot3(estoffset), f8dot3(jd)))
+               PeerSummary.prettyinterval(poll_sec), reach,
+               fini(estdelay), fini(estoffset), fini(jd)))
+        if self.displaymode == "rpeers":
+            line += " " + dstadr_refid
+            line += (" " * (self.refidwidth - len(dstadr_refid)))
+            line += (" %c %s" % (c, clock_name[:76 - len(line)]))
         line += "\n"
         # for debugging both case
         # if srcadr != None and srchost != None:
@@ -1232,34 +1236,34 @@ class MRUSummary:
         self.debug = debug
         self.logfp = logfp
         self.now = None
-        self.showhostnames = showhostnames      # If false, display numeric IPs
+        self.showhostnames = showhostnames  # if & 1, display names
         self.wideremote = wideremote
 
-    header = " lstint avgint rstr r m v  count rport remote address"
+    header = " lstint avgint rstr r m v  count    score   drop rport remote address"
 
     def summary(self, entry):
+        first = ntp.ntpc.lfptofloat(entry.first)
         last = ntp.ntpc.lfptofloat(entry.last)
+        active = float(last - first)
+        count = int(entry.ct)
         if self.now:
             lstint = int(self.now - last + 0.5)
             stats = "%7d" % lstint
+            if count == 1:
+                favgint = 0
+            else:
+                favgint = active / (count-1)
+            avgint = int(favgint + 0.5)
+            if 5.0 < favgint or 1 == count:
+                stats += " %6d" % avgint
+            elif 1.0 <= favgint:
+                stats += " %6.2f" % favgint
+            else:
+                stats += " %6.3f" % favgint
         else:
-            # direct mode doesn't have a reference time
             MJD_1970 = 40587     # MJD for 1 Jan 1970, Unix epoch
             days, lstint = divmod(int(last), 86400)
-            stats = "%5d %5d" % (days + MJD_1970, lstint)
-        first = ntp.ntpc.lfptofloat(entry.first)
-        active = float(last - first)
-        if entry.ct == 1:
-            favgint = 0
-        else:
-            favgint = active / (entry.ct-1)
-        avgint = int(favgint + 0.5)
-        if 5.0 < favgint or 1 == entry.ct:
-            stats += " %6d" % avgint
-        elif 1.0 <= favgint:
-            stats += " %6.2f" % favgint
-        else:
-            stats += " %6.3f" % favgint
+            stats = "%5d %5d %6d" % (days + MJD_1970, lstint, active)
         if entry.rs & ntp.magic.RES_KOD:
             rscode = 'K'
         elif entry.rs & ntp.magic.RES_LIMITED:
@@ -1268,7 +1272,7 @@ class MRUSummary:
             rscode = '.'
         (ip, port) = portsplit(entry.addr)
         try:
-            if not self.showhostnames:
+            if not self.showhostnames & 1:  # if not & 1 display numeric IPs
                 dns = ip
             else:
                 dns = canonicalize_dns(ip)
@@ -1290,11 +1294,25 @@ class MRUSummary:
             if not self.wideremote:
                 # truncate for narrow display
                 dns = dns[:40]
-            stats += " %4hx %c %d %d %6d %5s %s" % \
+            if entry.sc:
+                score = float(entry.sc)
+                if score > 100000.0:
+                  score = "%8.1f" % score
+                elif score > 10000.0:
+                  score = "%8.2f" % score
+                else:
+                  score = "%8.3f" % score
+            else:
+                score = "-"
+            if entry.dr!= None:     # 0 is valid
+                drop = "%4d" % entry.dr
+            else:
+                drop = "-"
+            stats += " %4hx %c %d %d %6d %8s %6s %5s %s" % \
                      (entry.rs, rscode,
                       ntp.magic.PKT_MODE(entry.mv),
                       ntp.magic.PKT_VERSION(entry.mv),
-                      entry.ct, port[1:], dns)
+                      entry.ct, score, drop, port[1:], dns)
             return stats
         except ValueError:
             # This can happen when ntpd ships a corrupt varlist
@@ -1392,7 +1410,7 @@ class IfstatsSummary:
         # FIXME, a brutal and slow way to check for invalid chars..
         # maybe just strip non-printing chars?
         for c in s:
-            if not c.isalnum() and c not in "/.:[] \%\n":
+            if not c.isalnum() and c not in "/.:[] %\n":
                 return ''
         return s
 
@@ -1431,4 +1449,43 @@ except ImportError:  # pragma: no cover
             for key in self.__keys:
                 yield key
 
+
+def packetize(packets, period, clipdigits=0, periodized=False):
+    """Given a number of packets and a duration (s) return a tuple.
+
+    return the packet quantity, and a two part rate in packets/seconds
+    or seconds/packet. On error the latter fields should be blank, the
+    first the number of packets if zero otherwise unhelpful text."""
+    if not isinstance(packets, int):
+        return ("???", "", "")
+    if packets == 0 or not isinstance(period, (int, float)):
+        return (packets, "", "")
+    if packets > period:
+        return (packets, round(packets / period, clipdigits), "p/s")
+    if periodized:
+        return (packets, periodize(period / packets, clipdigits)[1], "/p")
+    return (packets, round(period / packets, clipdigits), "s/p")
+
+
+def periodize(period, clipdigits=0):
+    """Given a number of seconds, return number and pretty string.
+
+    On error return None for the number and an unhelpful string."""
+    clip = clipdigits if isinstance(clipdigits, (int, float)) else 0
+    if not isinstance(period, (int, float)):
+        return (None, "???")
+    result = ""
+    _ = round(period, clip)
+    nperiod = int(_) if clip < 1 else _
+    if nperiod >= 86400:
+        result += "%dD " % (nperiod // 86400)
+    result += "%02d:%02d:%02d" % (
+        (nperiod % 86400) // 3600,
+        (nperiod % 3600) // 60,
+        nperiod % 60,
+    )
+    return (nperiod, result)
+
+
+uptime = lambda p: periodize(p)[1]
 # end
