@@ -1,19 +1,22 @@
-#!/usr/bin/env python
 # -*- coding: utf-8 -*-
 
+# Copyright the NTPsec project contributors
+#
+# SPDX-License-Identifier: BSD-2-Clause
+
 from __future__ import print_function, division
 
-import unittest
-import ntp.packet
-import ntp.control
-import ntp.util
-import ntp.magic
-import socket
-import select
-import sys
 import getpass
+import select
+import socket
+import sys
+import unittest
 import jigs
+import ntp.control
+import ntp.magic
+import ntp.packet
 import ntp.poly
+import ntp.util
 
 odict = ntp.util.OrderedDict
 
@@ -25,7 +28,7 @@ class SessionJig:
     def __init__(self):
         self.readvars_calls = 0
 
-    def readvars(self):
+    def readvar(self):
         self.readvars_calls += 1
         return {"foo": 23, "bar": 42}
 
@@ -124,6 +127,110 @@ class TestPacket(unittest.TestCase):
         # Test mode
         self.assertEqual(cls.mode(), 4)
 
+    def test_parse_mru_variables(self):
+        f = ntpp.parse_mru_variables
+        # Test working sort
+        data = {"sort": "count", "mincount": 50, "resall": 1, "resany": 5,
+                "kod": True, "limited": True, "maxlstint": 100,
+                "laddr": "foo.test", "recent": "foo", "frags": 20, "limit": 80}
+        sorter, sortkey, frags = f(data)
+        self.assertEqual(sorter != None, True)  # can't directly test lambda
+        self.assertEqual(sortkey, "count")
+        self.assertEqual(frags, 20)
+        self.assertEqual(data,
+                         {"mincount": 50, "resall": 1, "resany": 1061,
+                          "maxlstint": 100, "laddr": "foo.test",
+                          "recent": "foo", "limit": 80})
+        # Test no sort
+        data = {"mincount": 50, "resall": 1, "resany": 5, "kod": True,
+                "limited": True, "maxlstint": 100, "laddr": "foo.test",
+                "recent": "foo", "frags": 20, "limit": 80}
+        sorter, sortkey, frags = f(data)
+        self.assertEqual(sorter, None)  # can't directly test lambda
+        self.assertEqual(sortkey, None)
+        self.assertEqual(frags, 20)
+        self.assertEqual(data,
+                         {"mincount": 50, "resall": 1, "resany": 1061,
+                          "maxlstint": 100, "laddr": "foo.test",
+                          "recent": "foo", "limit": 80})
+        # Test bad sort
+        data = {"sort": "FAIL", "mincount": 50, "resall": 1, "resany": 5,
+                "kod": True, "limited": True, "maxlstint": 100,
+                "laddr": "foo.test", "recent": "foo", "frags": 20, "limit": 80}
+        try:
+            sorter, sortkey, frags = f(data)
+            errored = False
+        except ntpp.ControlException as e:
+            errored = e.message
+        self.assertEqual(errored, "***Sort order FAIL is not implemented")
+        # Test bad variables
+        data = {"FAIL": "FAIL", "mincount": 50, "resall": 1, "resany": 5,
+                "kod": True, "limited": True, "maxlstint": 100,
+                "laddr": "foo.test", "recent": "foo", "frags": 20, "limit": 80}
+        try:
+            sorter, sortkey, frags = f(data)
+            errored = False
+        except ntpp.ControlException as e:
+            errored = e.message
+        self.assertEqual(errored, "***Unknown parameter 'FAIL'")
+
+    def test_stitch_mru(self):
+        f = ntpp.stitch_mru
+        # Prepare data
+        span = ntpp.MRUList()
+        entry1 = ntpp.MRUEntry()
+        entry1.addr = "1.2.3.4:23"
+        entry1.last = "2019-12-31T00:15:00Z"
+        entry2 = ntpp.MRUEntry()
+        entry2.addr = "10.20.30.40:42"
+        entry2.last = "2019-12-31T03:00:00Z"
+        entry3 = ntpp.MRUEntry()
+        entry3.addr = "1.2.3.4:23"
+        entry3.last = "2019-12-31T03:30:00Z"
+        entry4 = ntpp.MRUEntry()
+        entry4.addr = "1.2.3.4:23"
+        entry4.last = "2019-12-31T04:45:00Z"
+        span.entries = [entry1, entry2, entry3, entry4]
+        sorter = (lambda e: e.sortaddr())  # Extracted from parse_mru_variables
+        sortkey = "addr"
+        f(span, sorter, sortkey)
+        self.assertEqual(span.entries, [entry2, entry4])
+
+    def test_generate_mru_parms(self):
+        f = ntpp.generate_mru_parms
+        # Data
+        vars = {"foo": 1, "bar": 2, "recent": 4}
+        # Run test
+        parms, firstParms = f(vars)
+        correct = parms == ", foo=1, bar=2" or parms == ", bar=2, foo=1"
+        self.assertEqual(correct, True)
+        one = firstParms == ", recent=4, foo=1, bar=2"
+        two = firstParms == ", recent=4, bar=2, foo=1"
+        self.assertEqual(one or two, True)
+
+    def test_generate_mru_lastseen(self):
+        f = ntpp.generate_mru_lastseen
+        # Data
+        span = ntpp.MRUList()
+        entry1 = ntpp.MRUEntry()
+        entry1.addr = "1.2.3.4:23"
+        entry1.last = "2019-12-31T00:15:00Z"
+        entry2 = ntpp.MRUEntry()
+        entry2.addr = "10.20.30.40:42"
+        entry2.last = "2019-12-31T03:00:00Z"
+        entry3 = ntpp.MRUEntry()
+        entry3.addr = "2.0.0.1:23"
+        entry3.last = "2019-12-31T03:30:00Z"
+        entry4 = ntpp.MRUEntry()
+        entry4.addr = "3.4.5.6:23"
+        entry4.last = "2019-12-31T04:45:00Z"
+        span.entries = [entry1, entry2, entry3, entry4]
+        # Run test
+        buf = f(span, 370)
+        expected = ", addr.0=3.4.5.6:23, last.0=2019-12-31T04:45:00Z, " \
+                              "addr.1=2.0.0.1:23, last.1=2019-12-31T03:30:00Z"
+        self.assertEqual(buf, expected)
+
 
 class TestSyncPacket(unittest.TestCase):
     target = ntpp.SyncPacket
@@ -225,7 +332,7 @@ class TestSyncPacket(unittest.TestCase):
             errored = False
         except ntpp.SyncException as e:
             errored = e.message
-        self.assertEqual(errored, "Unsupported DES authentication")
+        self.assertEqual(errored, "Packet is a runt")
         # Test with extension, runt 8
         data2 = data + ext + "\x11\x22\x33\x44\x55\x66\x77\x88"
         try:
@@ -495,6 +602,7 @@ class TestMisc(unittest.TestCase):
         cls.last = "0x00000200.00000000"
         cls.first = "0x00000100.00000000"
         cls.ct = 4
+        cls.sc = "0.12345"
         self.assertEqual(cls.avgint(), 64)
         if socket.has_ipv6:
             # Test sortaddr, ipv6
@@ -520,15 +628,17 @@ class TestMisc(unittest.TestCase):
                              "<MRUEntry: "
                              "'last': '0x00000200.00000000', "
                              "'addr': '11.22.33.44:23', 'rs': None, "
-                             "'mv': None, 'first': '0x00000100.00000000', "
-                             "'ct': 4>")
+                             "'mv': None, 'sc': '0.12345', "
+                             "'first': '0x00000100.00000000', "
+                             "'dr': None, 'ct': 4>")
         elif sys.version_info[1] >= 6:  # Already know it is 3.something
             # Python 3.6+, dicts enumerate in assignment order
             self.assertEqual(cls.__repr__(),
                              "<MRUEntry: 'addr': '11.22.33.44:23', "
                              "'last': '0x00000200.00000000', "
-                             "'first': '0x00000100.00000000', 'ct': 4, "
-                             "'mv': None, 'rs': None>")
+                             "'first': '0x00000100.00000000', "
+                             "'mv': None, 'rs': None, 'ct': 4, "
+                             "'sc': '0.12345', 'dr': None>")
         else:
             # Python 3.x < 3.6, dicts enumerate randomly
             # I can not test randomness of this type
@@ -672,7 +782,7 @@ class TestControlSession(unittest.TestCase):
         self.assertEqual(cls.secondary_timeout, ntpp.DEFSTIMEOUT)
         self.assertEqual(cls.pktversion, ntp.magic.NTP_OLDVERSION + 1)
         self.assertEqual(cls.always_auth, False)
-        self.assertEqual(cls.keytype, "MD5")
+        self.assertEqual(cls.keytype, "AES")
         self.assertEqual(cls.keyid, None)
         self.assertEqual(cls.passwd, None)
         self.assertEqual(cls.auth, None)
@@ -1038,16 +1148,16 @@ class TestControlSession(unittest.TestCase):
             ntpp.select = fakeselectmod
             # Test empty
             sockjig.return_data = [
-                "\x0E\x81\x00\x00\x00\x03\x00\x02\x00\x00\x00\x00"]
+                "\x16\x81\x00\x00\x00\x03\x00\x02\x00\x00\x00\x00"]
             cls.getresponse(1, 2, True)
             self.assertEqual(cls.response, ntp.poly.polybytes(""))
             # Test with data
             sockjig.return_data = [
-                "\x0E\xA1\x00\x01\x00\x02\x00\x03\x00\x00\x00\x09"
+                "\x16\xA1\x00\x01\x00\x02\x00\x03\x00\x00\x00\x09"
                 "foo=4223,\x00\x00\x00",
-                "\x0E\xA1\x00\x01\x00\x02\x00\x03\x00\x09\x00\x0E"
+                "\x16\xA1\x00\x01\x00\x02\x00\x03\x00\x09\x00\x0E"
                 "blah=248,x=23,\x00\x00",
-                "\x0E\x81\x00\x01\x00\x02\x00\x03\x00\x17\x00\x06"
+                "\x16\x81\x00\x01\x00\x02\x00\x03\x00\x17\x00\x06"
                 "quux=1\x00\x00"]
             cls.sequence = 1
             cls.getresponse(1, 3, True)
@@ -1055,13 +1165,13 @@ class TestControlSession(unittest.TestCase):
                              ntp.poly.polybytes("foo=4223,blah=248,x=23,quux=1"))
             # Test with data, duplicate packet
             sockjig.return_data = [
-                "\x0E\xA1\x00\x01\x00\x02\x00\x03\x00\x00\x00\x09"
+                "\x16\xA1\x00\x01\x00\x02\x00\x03\x00\x00\x00\x09"
                 "foo=4223,\x00\x00\x00",
-                "\x0E\xA1\x00\x01\x00\x02\x00\x03\x00\x09\x00\x0E"
+                "\x16\xA1\x00\x01\x00\x02\x00\x03\x00\x09\x00\x0E"
                 "blah=248,x=23,\x00\x00",
-                "\x0E\xA1\x00\x01\x00\x02\x00\x03\x00\x09\x00\x0E"
+                "\x16\xA1\x00\x01\x00\x02\x00\x03\x00\x09\x00\x0E"
                 "blah=248,x=23,\x00\x00",
-                "\x0E\x81\x00\x01\x00\x02\x00\x03\x00\x17\x00\x06"
+                "\x16\x81\x00\x01\x00\x02\x00\x03\x00\x17\x00\x06"
                 "quux=1\x00\x00"]
             cls.sequence = 1
             cls.getresponse(1, 3, True)
@@ -1071,7 +1181,7 @@ class TestControlSession(unittest.TestCase):
             maxtemp = ntpp.MAXFRAGS
             ntpp.MAXFRAGS = 1
             sockjig.return_data = [
-                "\x0E\xA1\x00\x01\x00\x02\x00\x03\x00\x00\x00\x09"
+                "\x16\xA1\x00\x01\x00\x02\x00\x03\x00\x00\x00\x09"
                 "foo=4223,\x00\x00\x00",
                 "\x0E\xA1\x00\x01\x00\x02\x00\x03\x00\x09\x00\x0E"
                 "blah=248,x=23,\x00\x00",
@@ -1103,11 +1213,11 @@ class TestControlSession(unittest.TestCase):
             self.assertEqual(errored, ntpp.SERR_TIMEOUT)
             # Test partial data and no timeout
             sockjig.return_data = [
-                "\x0E\xA1\x00\x01\x00\x02\x00\x03\x00\x00\x00\x09"
+                "\x16\xA1\x00\x01\x00\x02\x00\x03\x00\x00\x00\x09"
                 "foo=4223,\x00\x00\x00",
-                "\x0E\xA1\x00\x01\x00\x02\x00\x03\x00\x09\x00\x0E"
+                "\x16\xA1\x00\x01\x00\x02\x00\x03\x00\x09\x00\x0E"
                 "blah=248,x=23,\x00\x00",
-                "\x0E\x81\x00\x01\x00\x02\x00\x03\x00\x17\x00\x06"
+                "\x16\x81\x00\x01\x00\x02\x00\x03\x00\x17\x00\x06"
                 "quux=1\x00\x00"]
             fakeselectmod.do_return = [True, False]
             try:
@@ -1118,7 +1228,7 @@ class TestControlSession(unittest.TestCase):
             self.assertEqual(errored, ntpp.SERR_INCOMPLETE)
             # Test header parse fail
             sockjig.return_data = [
-                "\x0E\x81\x00\x00\x00\x03"]
+                "\x16\x81\x00\x00\x00\x03"]
             try:
                 cls.getresponse(1, 2, True)
                 errored = False
@@ -1140,7 +1250,7 @@ class TestControlSession(unittest.TestCase):
             ntp.util.time = faketimemod
             faketimemod.time_returns = [0] * 10
             # Test good packet, empty data
-            raw = "\x0E\x81\x00\x00\x00\x03\x00\x02\x00\x00\x00\x00"
+            raw = "\x16\x81\x00\x00\x00\x03\x00\x02\x00\x00\x00\x00"
             pkt = ntpp.ControlPacket(cls)
             pkt.analyze(raw)
             self.assertEqual(cls._ControlSession__validate_packet(pkt, raw,
@@ -1149,7 +1259,7 @@ class TestControlSession(unittest.TestCase):
             self.assertEqual(logjig.data, [])
             # Test good packet, with data
             logjig.data = []
-            raw = "\x0E\xA1\x00\x01\x00\x02\x00\x03\x00\x00\x00\x09" \
+            raw = "\x16\xA1\x00\x01\x00\x02\x00\x03\x00\x00\x00\x09" \
                   "foo=4223,\x00\x00\x00"
             pkt = ntpp.ControlPacket(cls)
             pkt.analyze(raw)
@@ -1173,7 +1283,7 @@ class TestControlSession(unittest.TestCase):
                               "received with version 0\n"])
             # Test bad packet, bad mode
             logjig.data = []
-            raw = "\x0D\x81\x00\x00\x00\x03\x00\x02\x00\x00\x00\x00"
+            raw = "\x15\x81\x00\x00\x00\x03\x00\x02\x00\x00\x00\x00"
             pkt = ntpp.ControlPacket(cls)
             pkt.analyze(raw)
             self.assertEqual(cls._ControlSession__validate_packet(pkt, raw,
@@ -1184,7 +1294,7 @@ class TestControlSession(unittest.TestCase):
                               "with mode 5\n"])
             # Test bad packet, bad response bit
             logjig.data = []
-            raw = "\x0E\x01\x00\x00\x00\x03\x00\x02\x00\x00\x00\x00"
+            raw = "\x16\x01\x00\x00\x00\x03\x00\x02\x00\x00\x00\x00"
             pkt = ntpp.ControlPacket(cls)
             pkt.analyze(raw)
             self.assertEqual(cls._ControlSession__validate_packet(pkt, raw,
@@ -1195,7 +1305,7 @@ class TestControlSession(unittest.TestCase):
                               "wanted response\n"])
             # Test bad packet, bad sequence
             logjig.data = []
-            raw = "\x0E\x81\x00\x01\x00\x03\x00\x02\x00\x00\x00\x00"
+            raw = "\x16\x81\x00\x01\x00\x03\x00\x02\x00\x00\x00\x00"
             pkt = ntpp.ControlPacket(cls)
             pkt.analyze(raw)
             self.assertEqual(cls._ControlSession__validate_packet(pkt, raw,
@@ -1206,7 +1316,7 @@ class TestControlSession(unittest.TestCase):
                               "number 1, wanted 0\n"])
             # Test bad packet, bad opcode
             logjig.data = []
-            raw = "\x0E\x80\x00\x00\x00\x03\x00\x02\x00\x00\x00\x00"
+            raw = "\x16\x80\x00\x00\x00\x03\x00\x02\x00\x00\x00\x00"
             pkt = ntpp.ControlPacket(cls)
             pkt.analyze(raw)
             self.assertEqual(cls._ControlSession__validate_packet(pkt, raw,
@@ -1217,7 +1327,7 @@ class TestControlSession(unittest.TestCase):
                               "wanted 1\n"])
             # Test error packet
             logjig.data = []
-            raw = "\x0E\xC1\x00\x00" + \
+            raw = "\x16\xC1\x00\x00" + \
                   chr(ntp.control.CERR_BADVALUE) + \
                   "\x03\x00\x02\x00\x00\x00\x00"
             pkt = ntpp.ControlPacket(cls)
@@ -1231,7 +1341,7 @@ class TestControlSession(unittest.TestCase):
             # Test error packet, with more bit
             logjig.data = []
             errcs = chr(ntp.control.CERR_BADVALUE)
-            raw = "\x0E\xE1\x00\x00" + errcs + "\x03\x00\x02\x00\x00\x00\x00"
+            raw = "\x16\xE1\x00\x00" + errcs + "\x03\x00\x02\x00\x00\x00\x00"
             pkt = ntpp.ControlPacket(cls)
             pkt.analyze(raw)
             try:
@@ -1244,7 +1354,7 @@ class TestControlSession(unittest.TestCase):
             self.assertEqual(logjig.data, [errstr])
             # Test ok-ish packet, bad associd
             logjig.data = []
-            raw = "\x0E\x81\x00\x00\x00\x03\x00\xFF\x00\x00\x00\x00"
+            raw = "\x16\x81\x00\x00\x00\x03\x00\xFF\x00\x00\x00\x00"
             pkt = ntpp.ControlPacket(cls)
             pkt.analyze(raw)
             self.assertEqual(cls._ControlSession__validate_packet(pkt, raw,
@@ -1254,7 +1364,7 @@ class TestControlSession(unittest.TestCase):
                              ["Association ID 255 doesn't match expected 2\n"])
             # Test bad data padding
             logjig.data = []
-            raw = "\x0E\x81\x00\x00\x00\x03\x00\x02\x00\x00\x00\x01@"
+            raw = "\x16\x81\x00\x00\x00\x03\x00\x02\x00\x00\x00\x01@"
             pkt = ntpp.ControlPacket(cls)
             pkt.analyze(raw)
             self.assertEqual(cls._ControlSession__validate_packet(pkt, raw,
@@ -1264,7 +1374,7 @@ class TestControlSession(unittest.TestCase):
                              ["Response fragment not padded, size = 13\n"])
             # Test too little data
             logjig.data = []
-            raw = "\x0E\x81\x00\x00\x00\x03\x00\x02\x00\x00\x00\x10foo\x00"
+            raw = "\x16\x81\x00\x00\x00\x03\x00\x02\x00\x00\x00\x10foo\x00"
             pkt = ntpp.ControlPacket(cls)
             pkt.analyze(raw)
             try:
@@ -1361,7 +1471,17 @@ class TestControlSession(unittest.TestCase):
     def test___parse_varlist(self):
         # Init
         cls = self.target()
-        cls.response = 'srcadr=0.0.0.0, srcport=0, srchost="0.ubuntu.pool.ntp.org",\r\ndstadr=0.0.0.0, dstport=0, leap=3, stratum=16, precision=-22,\r\nrootdelay=0.000, rootdisp=0.000, refid=POOL,\r\nreftime=0x00000000.00000000, rec=0x00000000.00000000, reach=0x0,\r\nunreach=0, hmode=3, pmode=0, hpoll=6, ppoll=10, headway=0, flash=0x1600,\r\nkeyid=0, offset=0.000, delay=0.000, dispersion=16000.000, jitter=0.000,\r\nfiltdelay= 0.00 0.00 0.00 0.00 0.00 0.00 0.00 0.00,\r\nfiltoffset= 0.00 0.00 0.00 0.00 0.00 0.00 0.00 0.00,\r\nfiltdisp= 16000.00 16000.00 16000.00 16000.00 16000.00 16000.00 16000.00 16000.00,\r\nnovalue, blankvalue=, quotedvalue="jabber"'
+        cls.response = (
+            'srcadr=0.0.0.0, srcport=0, srchost="0.ubuntu.pool.ntp.org",\r\n'
+            'dstadr=0.0.0.0, dstport=0, leap=3, stratum=16, precision=-22,\r\n'
+            'rootdelay=0.000, rootdisp=0.000, refid=POOL,\r\n'
+            'reftime=0x00000000.00000000, rec=0x00000000.00000000, reach=0x0,\r\n'
+            'unreach=0, hmode=3, pmode=0, hpoll=6, ppoll=10, headway=0, flash=0x1600,\r\n'
+            'keyid=0, offset=0.000, delay=0.000, dispersion=16000.000, jitter=0.000,\r\n'
+            'filtdelay= 0.00 0.00 0.00 0.00 0.00 0.00 0.00 0.00,\r\n'
+            'filtoffset= 0.00 0.00 0.00 0.00 0.00 0.00 0.00 0.00,\r\n'
+            'filtdisp= 16000.00 16000.00 16000.00 16000.00 16000.00 16000.00 16000.00 16000.00,\r\n'
+            'novalue, blankvalue=, quotedvalue="jabber"')
         # Test with basic packet
         self.assertEqual(cls._ControlSession__parse_varlist(),
                          odict((("srcadr", "0.0.0.0"), ("srcport", 0),
@@ -1523,6 +1643,41 @@ class TestControlSession(unittest.TestCase):
                           (ntp.control.CTL_OP_REQ_NONCE, 0, "", False),
                           (ntp.control.CTL_OP_REQ_NONCE, 0, "", False)])
 
+    def test___mru_analyze(self):
+        # data: nonce, last.older, addr.older, now, last.newest
+        #       addr, last, first, ct, mv, rs
+        vars = odict((("nonce", "noncevalue"), ("last.older", "FAIL0"),
+                ("addr.older", "FAIL1"), ("now", "0xcfba1ce0.80000000"),
+                ("last.newest", "FAIL2"), ("addr.1", "addrtest"),
+                ("last.2", "lasttest"), ("first.3", "firsttest"), ("ct.4", "cttest"),
+                ("mv.5", "mvtest"), ("rs.6", "rstest")))
+        cls = self.target()
+        span = ntpp.MRUList()
+        nonce = cls._ControlSession__mru_analyze(vars, span, None)
+        self.assertEqual(nonce, "nonce=noncevalue")
+        # expected entries, in vars order
+        m1 = ntpp.MRUEntry()
+        m1.addr = "addrtest"
+        m2 = ntpp.MRUEntry()
+        m2.last = "lasttest"
+        m3 = ntpp.MRUEntry()
+        m3.first = "firsttest"
+        m4 = ntpp.MRUEntry()
+        m4.ct = "cttest"
+        m5 = ntpp.MRUEntry()
+        m5.mv = "mvtest"
+        m6 = ntpp.MRUEntry()
+        m6.rs = "rstest"
+        expected = [m1, m2, m3, m4, m5, m6] # sort order
+        self.assertEqual(len(span.entries), len(expected))
+        for i in range(len(span.entries)):
+            self.assertEqual(span.entries[i].addr, expected[i].addr)
+            self.assertEqual(span.entries[i].last, expected[i].last)
+            self.assertEqual(span.entries[i].first, expected[i].first)
+            self.assertEqual(span.entries[i].ct, expected[i].ct)
+            self.assertEqual(span.entries[i].mv, expected[i].mv)
+            self.assertEqual(span.entries[i].rs, expected[i].rs)
+
     def test_mrulist(self):
 
         def setresponse(data):  # needed for doquery_jig
@@ -1940,17 +2095,18 @@ class TestAuthenticator(unittest.TestCase):
 
     def test_compute_mac(self):
         f = self.target.compute_mac
-        try:
-            temphash = ntpp.hashlib
-            fakehashlibmod = jigs.HashlibModuleJig()
-            ntpp.hashlib = fakehashlibmod
-            # Test no digest
-            self.assertEqual(f("", 0, None, ntp.poly.polybytes("")), None)
-            # Test with digest
-            self.assertEqual(f("foo", 0x42, "bar", "quux"),
-                             ntp.poly.polybytes("\x00\x00\x00\x42blahblahblahblah"))
-        finally:
-            ntpp.hashlib = temphash
+        pkt = ntp.util.hexstr2octets('240300e8000012ce0000091941138e89'
+                                     'e25b102e9fe94dc9e25b1175bd5a3000'
+                                     'e25b1175bd6cf48ee25b1175bd70e594')
+        mac1 = b'\x00\x00\x00\rL\x7f\xc1\xd1\xe9\xd3\xf8\xec\x91\xdf\xecS\x89e\xc5\xf3'
+        mac2 = b'\x00\x00\x00\x0e\'\xca\xf3\xa8G\xbe\x99\xaf\x96!<\x1a\x95\xd6\xc7i'
+        key1 = ntp.util.hexstr2octets('2f3badbb640bf975fec519df8a83e829')
+        key2 = ntp.util.hexstr2octets('fc0af35640142bb85ad03a892a81fa9d'
+                                      '891b108289c3f6725a01a10f697c1b4c')
+        key3 = ''
+        self.assertEqual(f(pkt, 0x0d, 'aes', key1), mac1)
+        self.assertEqual(f(pkt, 0x0e, 'aes-256', key2), mac2)
+        self.assertEqual(f(pkt, 0x0f, 'neun', key3), False)
 
     def test_have_mac(self):
         f = self.target.have_mac
@@ -1963,19 +2119,24 @@ class TestAuthenticator(unittest.TestCase):
 
     def test_verify_mac(self):
         cls = self.target()
-        cls.passwords[0x23] = ("a", "z")
-        good_pkt = "foobar\x00\x00\x00\x23blahblahblahblah"
-        bad_pkt = "foobar\xDE\xAD\xDE\xAFblahblahblah"
-        try:
-            temphash = ntpp.hashlib
-            fakehashlibmod = jigs.HashlibModuleJig()
-            ntpp.hashlib = fakehashlibmod
-            # Test good
-            self.assertEqual(cls.verify_mac(ntp.poly.polybytes(good_pkt)), True)
-            # Test bad
-            self.assertEqual(cls.verify_mac(ntp.poly.polybytes(bad_pkt)), False)
-        finally:
-            ntpp.hashlib = temphash
+        cls.passwords[0x0d] = (
+            'aes-128', ntp.util.hexstr2octets('2f3badbb640bf975fec519df8a83e829'))
+        good_pkt = ('240300e80000139a00000ae8cc0286a2'
+                    'e25c0c4dfff93ee2e25c0cca53f45000'
+                    'e25c0cca54048d79e25c0cca5408646b'
+                    '0000000d'
+                    'be93e3f1d530d9252147c298c00c85f9')
+        bad_pkt = ('240300e80000131f00000779cc0286a2'
+                   'e25c0d54ff6e4835e25c0dc2bea43000'
+                   'e25c0dc2beb78905e25c0dc2bebc0737'
+                   '0000000d'
+                   '4c2d64c447e701b74e3ad98ce65d13c3')
+        # Test good
+        self.assertEqual(cls.verify_mac(ntp.poly.polybytes(
+            ntp.util.hexstr2octets(good_pkt)), packet_end=48, mac_begin=48), True)
+        # Test bad
+        self.assertEqual(cls.verify_mac(ntp.poly.polybytes(
+            ntp.util.hexstr2octets(bad_pkt)), packet_end=48, mac_begin=48), False)
 
 
 if __name__ == "__main__":
