@@ -1,3 +1,7 @@
+# Copyright the NTPsec project contributors
+#
+# SPDX-License-Identifier: BSD-2-Clause
+
 def options_cmd(ctx, config):
     ctx.load("compiler_c")
     ctx.load("msvc")
@@ -10,8 +14,10 @@ def options_cmd(ctx, config):
     grp = ctx.add_option_group("NTP configure options")
     grp.add_option('--enable-debug', action='store_true',
                    default=False, help="Enable debugging code")
-    grp.add_option('--enable-debug-gdb', action='store_true',
-                   default=False, help="Enable GDB debugging symbols")
+    grp.add_option('--disable-debug-gdb', action='store_true',
+                   default=False, help="Disable GDB debugging symbols")
+    grp.add_option('--enable-attic', action='store_true',
+                   default=False, help="Enable building attic/*.")
     grp.add_option('--disable-nts', action='store_true',
                    default=False, help="Disable NTS.")
     grp.add_option('--disable-droproot', action='store_true',
@@ -31,6 +37,10 @@ def options_cmd(ctx, config):
     grp.add_option('--enable-debug-timing', action='store_true',
                    default=False,
                    help="Collect timing statistics for debugging.")
+    grp.add_option('--enable-pylib', action='store',
+                   default='ffi', choices=['ext', 'ffi', 'none'],
+                   help="""Choose which Python library to build.\n
+ext, ffi, or none. defaults to ffi.""", nargs=1)
 
     grp = ctx.add_option_group("NTP cross compile options")
     grp.add_option('--cross-compiler', type='string',
@@ -63,19 +73,17 @@ def options_cmd(ctx, config):
     grp.add_option('--cflags', type='string', action="callback",
                    callback=callback_flags,
                    help="Users should use CFLAGS in their environment.")
-    grp.add_option('--build-epoch', type='int', default=None,
-                   help="Force epoch, or use SOURCE_DATE_EPOCH in environment")
+    grp.add_option('--build-desc', type='string', default="",
+                   help="Arbitrary text string to append to extended version.")
     grp.add_option('--ldflags', type='string', action="callback",
                    callback=callback_flags,
                    help="Users should use LDFLAGS in their environment.")
     grp.add_option('--check', action='store_true', default=False,
                    help="Run tests")
-    grp.add_option('--enable-rtems-trace', action='store_true',
-                   default=False, help="Enable RTEMS Trace.")
     grp.add_option('--enable-warnings', action='store_true',
-                   default=False, help="Enable annoying CC warnings")
-    grp.add_option('--rtems-trace-path', type='string', default="",
-                   help="Path to rtems-tld.")
+                   default=False, help="Turn on a few more warnings")
+    grp.add_option('--enable-Werror', action='store_true',
+                   default=False, help="Turn on -Werror")
     grp.add_option(
         '--define', type='string', action="callback",
         callback=callback_flags,
@@ -85,4 +93,12 @@ def options_cmd(ctx, config):
                    help="Force undefinition of symbol.")
     grp.add_option('--disable-fuzz', action='store_true',
                    default=False,
-                   help="Disable fuzzing of low bits of time")
+                   help="Feature gone")
+
+    # This option group name needs to exactly match the name from
+    # waflib/Tools/python.py so this option groups with the other Python
+    # options.
+    grp = ctx.add_option_group("Python Options")
+    grp.add_option('--pyshebang', type='string',
+                   default='/usr/bin/env python',
+                   help="Shebang for Python scripts [Default: %(default)s}")
